@@ -30,19 +30,12 @@
 #include "sysdeps.h"
 #include "gstvaapicompat.h"
 #include "gstvaapicontext.h"
-#include "gstvaapicontext_overlay.h"
 #include "gstvaapidisplay_priv.h"
-#include "gstvaapiobject_priv.h"
-#include "gstvaapisurface.h"
 #include "gstvaapisurface_priv.h"
 #include "gstvaapisurfacepool.h"
 #include "gstvaapisurfaceproxy.h"
 #include "gstvaapivideopool_priv.h"
 #include "gstvaapiutils.h"
-#include "gstvaapiutils_core.h"
-
-#define DEBUG 1
-#include "gstvaapidebug.h"
 
 /* Define default VA surface chroma format to YUV 4:2:0 */
 #define DEFAULT_CHROMA_TYPE (GST_VAAPI_CHROMA_TYPE_YUV420)
@@ -50,54 +43,119 @@
 /* Number of scratch surfaces beyond those used as reference */
 #define SCRATCH_SURFACES_COUNT (4)
 
-static gboolean
-ensure_formats (GstVaapiContext * context)
-{
-  if (G_LIKELY (context->formats))
-    return TRUE;
-
-  context->formats =
-      gst_vaapi_get_surface_formats (GST_VAAPI_OBJECT_DISPLAY (context),
-      context->va_config);
-  return (context->formats != NULL);
-}
+/* Debug category for GstVaapiContext */
+GST_DEBUG_CATEGORY (gst_debug_vaapi_context);
+#define GST_CAT_DEFAULT gst_debug_vaapi_context
 
 static void
-unref_surface_cb (GstVaapiSurface * surface)
+_init_vaapi_context_debug (void)
 {
-  gst_vaapi_surface_set_parent_context (surface, NULL);
-  gst_vaapi_object_unref (surface);
+#ifndef GST_DISABLE_GST_DEBUG
+  static volatile gsize _init = 0;
+
+  if (g_once_init_enter (&_init)) {
+    GST_DEBUG_CATEGORY_INIT (gst_debug_vaapi_context, "vaapicontext", 0,
+        "VA-API context");
+    g_once_init_leave (&_init, 1);
+  }
+#endif
+}
+
+static inline gboolean
+_context_is_broken_jpeg_decoder (GstVaapiContext * context)
+{
+  GstVaapiDisplay *const display = GST_VAAPI_CONTEXT_DISPLAY (context);
+
+  return (context->info.profile == GST_VAAPI_PROFILE_JPEG_BASELINE
+      && context->info.entrypoint == GST_VAAPI_ENTRYPOINT_VLD
+      && gst_vaapi_display_has_driver_quirks (display,
+          GST_VAAPI_DRIVER_QUIRK_JPEG_DEC_BROKEN_FORMATS));
+}
+
+static gboolean
+ensure_attributes (GstVaapiContext * context)
+{
+  if (G_LIKELY (context->attribs))
+    return TRUE;
+
+  context->attribs =
+      gst_vaapi_config_surface_attributes_get (GST_VAAPI_CONTEXT_DISPLAY
+      (context), context->va_config);
+
+  if (!context->attribs)
+    return FALSE;
+
+  if (_context_is_broken_jpeg_decoder (context)) {
+    GstVideoFormat fmt = GST_VIDEO_FORMAT_NV12;
+    g_array_prepend_val (context->attribs->formats, fmt);
+
+    context->attribs->mem_types &= ~VA_SURFACE_ATTRIB_MEM_TYPE_DRM_PRIME;
+  }
+  return TRUE;
+}
+
+/* XXX(victor): verify the preferred video format concords with the
+ * chroma type; otherwise it is changed for the (very arbritrary)
+ * preferred format from the requested context chroma type, in the
+ * context attributes */
+static void
+ensure_preferred_format (GstVaapiContext * context)
+{
+  const GstVaapiContextInfo *const cip = &context->info;
+  GArray *formats;
+  guint i;
+
+  if (context->preferred_format != GST_VIDEO_FORMAT_UNKNOWN)
+    return;
+
+  if (_context_is_broken_jpeg_decoder (context))
+    return;
+
+  if (!ensure_attributes (context) || !context->attribs->formats)
+    return;
+
+  formats = context->attribs->formats;
+  for (i = 0; i < formats->len; i++) {
+    GstVideoFormat format = g_array_index (formats, GstVideoFormat, i);
+    if (format == gst_vaapi_video_format_from_chroma (cip->chroma_type)) {
+      context->preferred_format = format;
+      break;
+    }
+  }
+
+  return;
 }
 
 static inline gboolean
 context_get_attribute (GstVaapiContext * context, VAConfigAttribType type,
     guint * out_value_ptr)
 {
-  return gst_vaapi_get_config_attribute (GST_VAAPI_OBJECT_DISPLAY (context),
+  return gst_vaapi_get_config_attribute (GST_VAAPI_CONTEXT_DISPLAY (context),
       context->va_profile, context->va_entrypoint, type, out_value_ptr);
 }
 
 static void
 context_destroy_surfaces (GstVaapiContext * context)
 {
-  gst_vaapi_context_overlay_reset (context);
-
   if (context->surfaces) {
     g_ptr_array_unref (context->surfaces);
     context->surfaces = NULL;
   }
+
+  context->preferred_format = GST_VIDEO_FORMAT_UNKNOWN;
+
   gst_vaapi_video_pool_replace (&context->surfaces_pool, NULL);
 }
 
 static void
 context_destroy (GstVaapiContext * context)
 {
-  GstVaapiDisplay *const display = GST_VAAPI_OBJECT_DISPLAY (context);
+  GstVaapiDisplay *const display = GST_VAAPI_CONTEXT_DISPLAY (context);
   VAContextID context_id;
   VAStatus status;
 
-  context_id = GST_VAAPI_OBJECT_ID (context);
-  GST_DEBUG ("context 0x%08x", context_id);
+  context_id = GST_VAAPI_CONTEXT_ID (context);
+  GST_DEBUG ("context 0x%08x / config 0x%08x", context_id, context->va_config);
 
   if (context_id != VA_INVALID_ID) {
     GST_VAAPI_DISPLAY_LOCK (display);
@@ -106,7 +164,7 @@ context_destroy (GstVaapiContext * context)
     GST_VAAPI_DISPLAY_UNLOCK (display);
     if (!vaapi_check_status (status, "vaDestroyContext()"))
       GST_WARNING ("failed to destroy context 0x%08x", context_id);
-    GST_VAAPI_OBJECT_ID (context) = VA_INVALID_ID;
+    GST_VAAPI_CONTEXT_ID (context) = VA_INVALID_ID;
   }
 
   if (context->va_config != VA_INVALID_ID) {
@@ -119,35 +177,41 @@ context_destroy (GstVaapiContext * context)
     context->va_config = VA_INVALID_ID;
   }
 
-  if (context->formats) {
-    g_array_unref (context->formats);
-    context->formats = NULL;
+  if (context->attribs) {
+    gst_vaapi_config_surface_attributes_free (context->attribs);
+    context->attribs = NULL;
   }
 }
 
 static gboolean
 context_ensure_surfaces (GstVaapiContext * context)
 {
+  GstVaapiDisplay *display = GST_VAAPI_CONTEXT_DISPLAY (context);
   const GstVaapiContextInfo *const cip = &context->info;
   const guint num_surfaces = cip->ref_frames + SCRATCH_SURFACES_COUNT;
   GstVaapiSurface *surface;
-  guint i;
-
-  if (!ensure_formats (context))
-    return FALSE;
+  GstVideoFormat format;
+  guint i, capacity;
 
+  ensure_preferred_format (context);
+  format = context->preferred_format;
   for (i = context->surfaces->len; i < num_surfaces; i++) {
-    surface =
-        gst_vaapi_surface_new_from_formats (GST_VAAPI_OBJECT_DISPLAY (context),
-        cip->chroma_type, cip->width, cip->height, context->formats);
+    if (format != GST_VIDEO_FORMAT_UNKNOWN) {
+      surface = gst_vaapi_surface_new_with_format (display, format, cip->width,
+          cip->height, 0);
+    } else {
+      surface = gst_vaapi_surface_new (display, cip->chroma_type, cip->width,
+          cip->height);
+    }
     if (!surface)
       return FALSE;
-    gst_vaapi_surface_set_parent_context (surface, context);
     g_ptr_array_add (context->surfaces, surface);
     if (!gst_vaapi_video_pool_add_object (context->surfaces_pool, surface))
       return FALSE;
   }
-  gst_vaapi_video_pool_set_capacity (context->surfaces_pool, num_surfaces);
+
+  capacity = cip->usage == GST_VAAPI_CONTEXT_USAGE_DECODE ? 0 : num_surfaces;
+  gst_vaapi_video_pool_set_capacity (context->surfaces_pool, capacity);
   return TRUE;
 }
 
@@ -155,16 +219,13 @@ static gboolean
 context_create_surfaces (GstVaapiContext * context)
 {
   const GstVaapiContextInfo *const cip = &context->info;
-  GstVaapiDisplay *const display = GST_VAAPI_OBJECT_DISPLAY (context);
+  GstVaapiDisplay *const display = GST_VAAPI_CONTEXT_DISPLAY (context);
   guint num_surfaces;
 
-  if (!gst_vaapi_context_overlay_reset (context))
-    return FALSE;
-
   num_surfaces = cip->ref_frames + SCRATCH_SURFACES_COUNT;
   if (!context->surfaces) {
     context->surfaces = g_ptr_array_new_full (num_surfaces,
-        (GDestroyNotify) unref_surface_cb);
+        (GDestroyNotify) gst_mini_object_unref);
     if (!context->surfaces)
       return FALSE;
   }
@@ -172,7 +233,7 @@ context_create_surfaces (GstVaapiContext * context)
   if (!context->surfaces_pool) {
     context->surfaces_pool =
         gst_vaapi_surface_pool_new_with_chroma_type (display, cip->chroma_type,
-        cip->width, cip->height);
+        cip->width, cip->height, 0);
 
     if (!context->surfaces_pool)
       return FALSE;
@@ -184,13 +245,15 @@ static gboolean
 context_create (GstVaapiContext * context)
 {
   const GstVaapiContextInfo *const cip = &context->info;
-  GstVaapiDisplay *const display = GST_VAAPI_OBJECT_DISPLAY (context);
+  GstVaapiDisplay *const display = GST_VAAPI_CONTEXT_DISPLAY (context);
   VAContextID context_id;
   VASurfaceID surface_id;
+  VASurfaceID *surfaces_data = NULL;
   VAStatus status;
   GArray *surfaces = NULL;
   gboolean success = FALSE;
   guint i;
+  gint num_surfaces = 0;
 
   if (!context->surfaces && !context_create_surfaces (context))
     goto cleanup;
@@ -205,26 +268,35 @@ context_create (GstVaapiContext * context)
     GstVaapiSurface *const surface = g_ptr_array_index (context->surfaces, i);
     if (!surface)
       goto cleanup;
-    surface_id = GST_VAAPI_OBJECT_ID (surface);
+    surface_id = GST_VAAPI_SURFACE_ID (surface);
     g_array_append_val (surfaces, surface_id);
   }
+
   g_assert (surfaces->len == context->surfaces->len);
 
+  /* vaCreateContext() doesn't really need an array of VASurfaceIDs (see
+   * https://lists.01.org/pipermail/intel-vaapi-media/2017-July/000052.html and
+   * https://github.com/intel/libva/issues/251); pass a dummy list of valid
+   * (non-null) IDs until the signature gets updated. */
+  if (cip->usage != GST_VAAPI_CONTEXT_USAGE_DECODE) {
+    surfaces_data = (VASurfaceID *) surfaces->data;
+    num_surfaces = surfaces->len;
+  }
+
   GST_VAAPI_DISPLAY_LOCK (display);
   status = vaCreateContext (GST_VAAPI_DISPLAY_VADISPLAY (display),
       context->va_config, cip->width, cip->height, VA_PROGRESSIVE,
-      (VASurfaceID *) surfaces->data, surfaces->len, &context_id);
+      surfaces_data, num_surfaces, &context_id);
   GST_VAAPI_DISPLAY_UNLOCK (display);
   if (!vaapi_check_status (status, "vaCreateContext()"))
     goto cleanup;
 
-  GST_DEBUG ("context 0x%08x", context_id);
-  GST_VAAPI_OBJECT_ID (context) = context_id;
+  GST_VAAPI_CONTEXT_ID (context) = context_id;
   success = TRUE;
 
 cleanup:
   if (surfaces)
-    g_array_free (surfaces, TRUE);
+    g_array_unref (surfaces);
   return success;
 }
 
@@ -232,13 +304,14 @@ static gboolean
 config_create (GstVaapiContext * context)
 {
   const GstVaapiContextInfo *const cip = &context->info;
-  GstVaapiDisplay *const display = GST_VAAPI_OBJECT_DISPLAY (context);
+  GstVaapiDisplay *const display = GST_VAAPI_CONTEXT_DISPLAY (context);
   VAConfigAttrib attribs[7], *attrib;
   VAStatus status;
   guint value, va_chroma_format, attrib_index;
 
   /* Reset profile and entrypoint */
-  if (!cip->profile || !cip->entrypoint)
+  if (cip->profile == GST_VAAPI_PROFILE_UNKNOWN
+      || cip->entrypoint == GST_VAAPI_ENTRYPOINT_INVALID)
     goto cleanup;
   context->va_profile = gst_vaapi_profile_get_va_profile (cip->profile);
   context->va_entrypoint =
@@ -333,14 +406,6 @@ config_create (GstVaapiContext * context)
         attrib = &attribs[++attrib_index];
         g_assert (attrib_index < G_N_ELEMENTS (attribs));
       }
-#endif
-#if USE_H264_FEI_ENCODER
-      if (cip->entrypoint == GST_VAAPI_ENTRYPOINT_SLICE_ENCODE_FEI) {
-        attrib->type = (VAConfigAttribType) VAConfigAttribFEIFunctionType;
-        attrib = &attribs[++attrib_index];
-        g_assert (attrib_index < G_N_ELEMENTS (attribs));
-        /* FIXME: Query the read-only supported MV predictors */
-      }
 #endif
       break;
     }
@@ -405,21 +470,11 @@ gst_vaapi_context_init (GstVaapiContext * context,
 
   context->va_config = VA_INVALID_ID;
   context->reset_on_resize = TRUE;
-  gst_vaapi_context_overlay_init (context);
 
-  context->formats = NULL;
+  context->attribs = NULL;
+  context->preferred_format = GST_VIDEO_FORMAT_UNKNOWN;
 }
 
-static void
-gst_vaapi_context_finalize (GstVaapiContext * context)
-{
-  context_destroy (context);
-  context_destroy_surfaces (context);
-  gst_vaapi_context_overlay_finalize (context);
-}
-
-GST_VAAPI_OBJECT_DEFINE_CLASS (GstVaapiContext, gst_vaapi_context);
-
 /**
  * gst_vaapi_context_new:
  * @display: a #GstVaapiDisplay
@@ -437,13 +492,24 @@ gst_vaapi_context_new (GstVaapiDisplay * display,
 {
   GstVaapiContext *context;
 
-  g_return_val_if_fail (cip->profile, NULL);
-  g_return_val_if_fail (cip->entrypoint, NULL);
+  g_return_val_if_fail (display, NULL);
 
-  context = gst_vaapi_object_new (gst_vaapi_context_class (), display);
+  _init_vaapi_context_debug ();
+
+  if (cip->profile == GST_VAAPI_PROFILE_UNKNOWN
+      || cip->entrypoint == GST_VAAPI_ENTRYPOINT_INVALID)
+    return NULL;
+
+  context = g_slice_new (GstVaapiContext);
   if (!context)
     return NULL;
 
+  GST_VAAPI_CONTEXT_DISPLAY (context) = gst_object_ref (display);
+  GST_VAAPI_CONTEXT_ID (context) = VA_INVALID_ID;
+  g_atomic_int_set (&context->ref_count, 1);
+  context->surfaces = NULL;
+  context->surfaces_pool = NULL;
+
   gst_vaapi_context_init (context, cip);
 
   if (!config_create (context))
@@ -461,12 +527,14 @@ gst_vaapi_context_new (GstVaapiDisplay * display,
     goto error;
 
 done:
+  GST_DEBUG ("context 0x%08" G_GSIZE_MODIFIER "x / config 0x%08x",
+      GST_VAAPI_CONTEXT_ID (context), context->va_config);
   return context;
 
   /* ERRORS */
 error:
   {
-    gst_vaapi_object_unref (context);
+    gst_vaapi_context_unref (context);
     return NULL;
   }
 }
@@ -491,6 +559,10 @@ gst_vaapi_context_reset (GstVaapiContext * context,
   gboolean grow_surfaces = FALSE;
   GstVaapiChromaType chroma_type;
 
+  if (new_cip->profile == GST_VAAPI_PROFILE_UNKNOWN
+      || new_cip->entrypoint == GST_VAAPI_ENTRYPOINT_INVALID)
+    return FALSE;
+
   chroma_type = new_cip->chroma_type ? new_cip->chroma_type :
       DEFAULT_CHROMA_TYPE;
   if (cip->chroma_type != chroma_type) {
@@ -555,7 +627,7 @@ gst_vaapi_context_get_id (GstVaapiContext * context)
 {
   g_return_val_if_fail (context != NULL, VA_INVALID_ID);
 
-  return GST_VAAPI_OBJECT_ID (context);
+  return GST_VAAPI_CONTEXT_ID (context);
 }
 
 /**
@@ -596,6 +668,8 @@ gst_vaapi_context_get_surface_count (GstVaapiContext * context)
 {
   g_return_val_if_fail (context != NULL, 0);
 
+  if (gst_vaapi_video_pool_get_capacity (context->surfaces_pool) == 0)
+    return G_MAXUINT;
   return gst_vaapi_video_pool_get_size (context->surfaces_pool);
 }
 
@@ -633,7 +707,82 @@ gst_vaapi_context_get_surface_formats (GstVaapiContext * context)
 {
   g_return_val_if_fail (context, NULL);
 
-  if (!ensure_formats (context))
+  if (!ensure_attributes (context))
     return NULL;
-  return g_array_ref (context->formats);
+
+  if (context->attribs->formats)
+    return g_array_ref (context->attribs->formats);
+  return NULL;
+}
+
+/**
+ * gst_vaapi_context_get_surface_attributes:
+ * @context: a #GstVaapiContext
+ * @out_attribs: an allocated #GstVaapiConfigSurfaceAttributes
+ *
+ * Copy context's surface restrictions to @out_attribs, EXCEPT the
+ * color formats. Use gst_vaapi_context_get_surface_formats() to get
+ * them.
+ *
+ * Returns: %TRUE if the attributes were extracted and copied; %FALSE,
+ * otherwise
+ **/
+gboolean
+gst_vaapi_context_get_surface_attributes (GstVaapiContext * context,
+    GstVaapiConfigSurfaceAttributes * out_attribs)
+{
+  g_return_val_if_fail (context, FALSE);
+
+  if (!ensure_attributes (context))
+    return FALSE;
+
+  if (out_attribs) {
+    out_attribs->min_width = context->attribs->min_width;
+    out_attribs->min_height = context->attribs->min_height;
+    out_attribs->max_width = context->attribs->max_width;
+    out_attribs->max_height = context->attribs->max_height;
+    out_attribs->mem_types = context->attribs->mem_types;
+    out_attribs->formats = NULL;
+  }
+
+  return TRUE;
+}
+
+/**
+ * gst_vaapi_context_ref:
+ * @context: a #GstVaapiContext
+ *
+ * Atomically increases the reference count of the given @context by one.
+ *
+ * Returns: The same @context argument
+ */
+GstVaapiContext *
+gst_vaapi_context_ref (GstVaapiContext * context)
+{
+  g_return_val_if_fail (context != NULL, NULL);
+
+  g_atomic_int_inc (&context->ref_count);
+
+  return context;
+}
+
+/**
+ * gst_vaapi_context_unref:
+ * @context: a #GstVaapiContext
+ *
+ * Atomically decreases the reference count of the @context by one. If
+ * the reference count reaches zero, the object will be free'd.
+ */
+void
+gst_vaapi_context_unref (GstVaapiContext * context)
+{
+  g_return_if_fail (context != NULL);
+  g_return_if_fail (context->ref_count > 0);
+
+  if (g_atomic_int_dec_and_test (&context->ref_count)) {
+    context_destroy (context);
+    context_destroy_surfaces (context);
+    gst_vaapi_display_replace (&context->display, NULL);
+    g_slice_free (GstVaapiContext, context);
+  }
 }
