@@ -25,11 +25,10 @@
 #ifndef GST_VAAPI_CONTEXT_H
 #define GST_VAAPI_CONTEXT_H
 
-#include "gstvaapiobject.h"
-#include "gstvaapiobject_priv.h"
 #include "gstvaapiprofile.h"
 #include "gstvaapidisplay.h"
 #include "gstvaapisurface.h"
+#include "gstvaapiutils_core.h"
 #include "gstvaapivideopool.h"
 
 G_BEGIN_DECLS
@@ -40,7 +39,6 @@ G_BEGIN_DECLS
 typedef struct _GstVaapiConfigInfoEncoder GstVaapiConfigInfoEncoder;
 typedef struct _GstVaapiContextInfo GstVaapiContextInfo;
 typedef struct _GstVaapiContext GstVaapiContext;
-typedef struct _GstVaapiContextClass GstVaapiContextClass;
 
 /**
  * GstVaapiContextUsage:
@@ -62,7 +60,6 @@ typedef enum {
  * @packed_headers: notify encoder that packed headers are submitted (mask).
  * @roi_capability: if encoder supports regions-of-interest.
  * @roi_num_supported: The number of regions-of-interest supported.
- * @fei_function: The functional mode for FEI Entrypoint (VA_FEI_FUNCTION_*).
  *
  * Extra configuration for encoding.
  */
@@ -72,7 +69,6 @@ struct _GstVaapiConfigInfoEncoder
   guint packed_headers;
   gboolean roi_capability;
   guint roi_num_supported;
-  guint fei_function;
 };
 
 /**
@@ -104,30 +100,24 @@ struct _GstVaapiContextInfo
 struct _GstVaapiContext
 {
   /*< private >*/
-  GstVaapiObject parent_instance;
+  volatile gint ref_count;
+  GstVaapiDisplay *display;
+  GstVaapiID object_id;
 
+  /*< public >*/
   GstVaapiContextInfo info;
   VAProfile va_profile;
   VAEntrypoint va_entrypoint;
   VAConfigID va_config;
   GPtrArray *surfaces;
   GstVaapiVideoPool *surfaces_pool;
-  GPtrArray *overlays[2];
-  guint overlay_id;
   gboolean reset_on_resize;
-  GArray *formats;
+  GstVaapiConfigSurfaceAttributes *attribs;
+  GstVideoFormat preferred_format;
 };
 
-/**
- * GstVaapiContextClass:
- *
- * A VA context wrapper class.
- */
-struct _GstVaapiContextClass
-{
-  /*< private >*/
-  GstVaapiObjectClass parent_class;
-};
+#define GST_VAAPI_CONTEXT_ID(context)        (((GstVaapiContext *)(context))->object_id)
+#define GST_VAAPI_CONTEXT_DISPLAY(context)   (((GstVaapiContext *)(context))->display)
 
 G_GNUC_INTERNAL
 GstVaapiContext *
@@ -160,6 +150,21 @@ G_GNUC_INTERNAL
 GArray *
 gst_vaapi_context_get_surface_formats (GstVaapiContext * context);
 
+G_GNUC_INTERNAL
+gboolean
+gst_vaapi_context_get_surface_attributes (GstVaapiContext * context,
+    GstVaapiConfigSurfaceAttributes * out_attribs);
+
+G_GNUC_INTERNAL
+GstVaapiContext *
+gst_vaapi_context_ref (GstVaapiContext * context);
+
+G_GNUC_INTERNAL
+void
+gst_vaapi_context_unref (GstVaapiContext * context);
+
+G_DEFINE_AUTOPTR_CLEANUP_FUNC(GstVaapiContext, gst_vaapi_context_unref)
+
 G_END_DECLS
 
 #endif /* GST_VAAPI_CONTEXT_H */
