@@ -51,11 +51,11 @@ GST_DEBUG_CATEGORY (gst_debug_vaapi_display);
 G_DEFINE_TYPE_WITH_CODE (GstVaapiDisplay, gst_vaapi_display, GST_TYPE_OBJECT,
     _do_init);
 
-typedef struct _GstVaapiConfig GstVaapiConfig;
-struct _GstVaapiConfig
+typedef struct _GstVaapiProfileConfig GstVaapiProfileConfig;
+struct _GstVaapiProfileConfig
 {
   GstVaapiProfile profile;
-  GstVaapiEntrypoint entrypoint;
+  guint32 entrypoints;          /* bits map of GstVaapiEntrypoint */
 };
 
 typedef struct _GstVaapiProperty GstVaapiProperty;
@@ -76,6 +76,8 @@ struct _GstVaapiFormatInfo
 #define DEFAULT_RENDER_MODE     GST_VAAPI_RENDER_MODE_TEXTURE
 #define DEFAULT_ROTATION        GST_VAAPI_ROTATION_0
 
+#define ENTRY_POINT_FLAG(entry) (1U << G_PASTE(GST_VAAPI_ENTRYPOINT_, entry))
+
 enum
 {
   PROP_RENDER_MODE = 1,
@@ -260,18 +262,19 @@ compare_rgb_formats (gconstpointer a, gconstpointer b)
 
 /* Check if configs array contains profile at entrypoint */
 static inline gboolean
-find_config (GArray * configs,
-    GstVaapiProfile profile, GstVaapiEntrypoint entrypoint)
+find_config (GPtrArray * configs, GstVaapiProfile profile,
+    GstVaapiEntrypoint entrypoint)
 {
-  GstVaapiConfig *config;
+  GstVaapiProfileConfig *config;
   guint i;
 
   if (!configs)
     return FALSE;
 
   for (i = 0; i < configs->len; i++) {
-    config = &g_array_index (configs, GstVaapiConfig, i);
-    if (config->profile == profile && config->entrypoint == entrypoint)
+    config = g_ptr_array_index (configs, i);
+    if (config->profile == profile
+        && (config->entrypoints & (1U << entrypoint)))
       return TRUE;
   }
   return FALSE;
@@ -279,21 +282,21 @@ find_config (GArray * configs,
 
 /* HACK: append H.263 Baseline profile if MPEG-4:2 Simple profile is supported */
 static void
-append_h263_config (GArray * configs)
+append_h263_config (GArray * configs, GPtrArray * decoders)
 {
-  GstVaapiConfig *config, tmp_config;
-  GstVaapiConfig *mpeg4_simple_config = NULL;
-  GstVaapiConfig *h263_baseline_config = NULL;
+  GstVaapiProfileConfig *config, tmp_config;
+  GstVaapiProfileConfig *mpeg4_simple_config = NULL;
+  GstVaapiProfileConfig *h263_baseline_config = NULL;
   guint i;
 
   if (!WORKAROUND_H263_BASELINE_DECODE_PROFILE)
     return;
 
-  if (!configs)
+  if (!decoders)
     return;
 
-  for (i = 0; i < configs->len; i++) {
-    config = &g_array_index (configs, GstVaapiConfig, i);
+  for (i = 0; i < decoders->len; i++) {
+    config = g_ptr_array_index (decoders, i);
     if (config->profile == GST_VAAPI_PROFILE_MPEG4_SIMPLE)
       mpeg4_simple_config = config;
     else if (config->profile == GST_VAAPI_PROFILE_H263_BASELINE)
@@ -303,7 +306,10 @@ append_h263_config (GArray * configs)
   if (mpeg4_simple_config && !h263_baseline_config) {
     tmp_config = *mpeg4_simple_config;
     tmp_config.profile = GST_VAAPI_PROFILE_H263_BASELINE;
+    tmp_config.entrypoints = ENTRY_POINT_FLAG (VLD);
     g_array_append_val (configs, tmp_config);
+    g_ptr_array_add (decoders, &g_array_index (configs,
+            GstVaapiProfileConfig, configs->len - 1));
   }
 }
 
@@ -311,20 +317,18 @@ append_h263_config (GArray * configs)
 static gint
 compare_profiles (gconstpointer a, gconstpointer b)
 {
-  const GstVaapiConfig *const config1 = (GstVaapiConfig *) a;
-  const GstVaapiConfig *const config2 = (GstVaapiConfig *) b;
-
-  if (config1->profile == config2->profile)
-    return config1->entrypoint - config2->entrypoint;
+  const GstVaapiProfileConfig *const config1 = (GstVaapiProfileConfig *) a;
+  const GstVaapiProfileConfig *const config2 = (GstVaapiProfileConfig *) b;
 
+  g_assert (config1->profile != config2->profile);
   return config1->profile - config2->profile;
 }
 
 /* Convert configs array to profiles as GstCaps */
 static GArray *
-get_profiles (GArray * configs)
+get_profiles (GPtrArray * configs, GstVaapiCodec codec)
 {
-  GstVaapiConfig *config;
+  GstVaapiProfileConfig *config;
   GArray *out_profiles;
   guint i;
 
@@ -336,8 +340,9 @@ get_profiles (GArray * configs)
     return NULL;
 
   for (i = 0; i < configs->len; i++) {
-    config = &g_array_index (configs, GstVaapiConfig, i);
-    g_array_append_val (out_profiles, config->profile);
+    config = g_ptr_array_index (configs, i);
+    if (!codec || (codec == gst_vaapi_profile_get_codec (config->profile)))
+      g_array_append_val (out_profiles, config->profile);
   }
   return out_profiles;
 }
@@ -470,10 +475,14 @@ ensure_profiles (GstVaapiDisplay * display)
     return TRUE;
   }
 
-  priv->decoders = g_array_new (FALSE, FALSE, sizeof (GstVaapiConfig));
+  priv->codecs = g_array_new (FALSE, FALSE, sizeof (GstVaapiProfileConfig));
+  if (!priv->codecs)
+    goto cleanup;
+
+  priv->decoders = g_ptr_array_new ();
   if (!priv->decoders)
     goto cleanup;
-  priv->encoders = g_array_new (FALSE, FALSE, sizeof (GstVaapiConfig));
+  priv->encoders = g_ptr_array_new ();
   if (!priv->encoders)
     goto cleanup;
   priv->has_profiles = TRUE;
@@ -493,16 +502,13 @@ ensure_profiles (GstVaapiDisplay * display)
 
   GST_DEBUG ("%d profiles", n);
   for (i = 0; i < n; i++) {
-#if VA_CHECK_VERSION(0,34,0)
-    /* Introduced in VA/VPP API */
     if (profiles[i] == VAProfileNone)
       continue;
-#endif
     GST_DEBUG ("  %s", string_of_VAProfile (profiles[i]));
   }
 
   for (i = 0; i < n; i++) {
-    GstVaapiConfig config;
+    GstVaapiProfileConfig config = { 0, };
 
     config.profile = gst_vaapi_profile (profiles[i]);
     if (!config.profile)
@@ -513,27 +519,31 @@ ensure_profiles (GstVaapiDisplay * display)
     if (!vaapi_check_status (status, "vaQueryConfigEntrypoints()"))
       continue;
 
-    for (j = 0; j < num_entrypoints; j++) {
-      config.entrypoint = gst_vaapi_entrypoint (entrypoints[j]);
-      switch (config.entrypoint) {
-        case GST_VAAPI_ENTRYPOINT_VLD:
-        case GST_VAAPI_ENTRYPOINT_IDCT:
-        case GST_VAAPI_ENTRYPOINT_MOCO:
-          g_array_append_val (priv->decoders, config);
-          break;
-        case GST_VAAPI_ENTRYPOINT_SLICE_ENCODE:
-        case GST_VAAPI_ENTRYPOINT_PICTURE_ENCODE:
-        case GST_VAAPI_ENTRYPOINT_SLICE_ENCODE_LP:
-        case GST_VAAPI_ENTRYPOINT_SLICE_ENCODE_FEI:
-          g_array_append_val (priv->encoders, config);
-          break;
-      }
-    }
-  }
-  append_h263_config (priv->decoders);
+    for (j = 0; j < num_entrypoints; j++)
+      config.entrypoints |= (1U << gst_vaapi_entrypoint (entrypoints[j]));
 
-  g_array_sort (priv->decoders, compare_profiles);
-  g_array_sort (priv->encoders, compare_profiles);
+    priv->codecs = g_array_append_val (priv->codecs, config);
+  }
+
+  for (i = 0; i < priv->codecs->len; i++) {
+    GstVaapiProfileConfig *cfg;
+
+    cfg = &g_array_index (priv->codecs, GstVaapiProfileConfig, i);
+
+    if ((cfg->entrypoints & ENTRY_POINT_FLAG (VLD))
+        || (cfg->entrypoints & ENTRY_POINT_FLAG (IDCT))
+        || (cfg->entrypoints & ENTRY_POINT_FLAG (MOCO)))
+      g_ptr_array_add (priv->decoders, cfg);
+    if ((cfg->entrypoints & ENTRY_POINT_FLAG (SLICE_ENCODE))
+        || (cfg->entrypoints & ENTRY_POINT_FLAG (PICTURE_ENCODE))
+        || (cfg->entrypoints & ENTRY_POINT_FLAG (SLICE_ENCODE_LP)))
+      g_ptr_array_add (priv->encoders, cfg);
+  }
+
+  append_h263_config (priv->codecs, priv->decoders);
+
+  g_ptr_array_sort (priv->decoders, compare_profiles);
+  g_ptr_array_sort (priv->encoders, compare_profiles);
 
   /* Video processing API */
   status = vaQueryConfigEntrypoints (priv->display, VAProfileNone,
@@ -590,11 +600,6 @@ ensure_properties (GstVaapiDisplay * display)
     GST_DEBUG ("  %s", string_of_VADisplayAttributeType (attr->type));
 
     switch (attr->type) {
-#if !VA_CHECK_VERSION(0,34,0)
-      case VADisplayAttribDirectSurface:
-        prop.name = GST_VAAPI_DISPLAY_PROP_RENDER_MODE;
-        break;
-#endif
       case VADisplayAttribRenderMode:
         prop.name = GST_VAAPI_DISPLAY_PROP_RENDER_MODE;
         break;
@@ -648,7 +653,7 @@ ensure_image_formats (GstVaapiDisplay * display)
   GstVaapiDisplayPrivate *const priv = GST_VAAPI_DISPLAY_GET_PRIVATE (display);
   VAImageFormat *formats = NULL;
   VAStatus status;
-  gint i, n;
+  gint i, n, max_images;
   gboolean success = FALSE;
 
   GST_VAAPI_DISPLAY_LOCK (display);
@@ -662,7 +667,8 @@ ensure_image_formats (GstVaapiDisplay * display)
     goto cleanup;
 
   /* VA image formats */
-  formats = g_new (VAImageFormat, vaMaxNumImageFormats (priv->display));
+  max_images = vaMaxNumImageFormats (priv->display);
+  formats = g_new (VAImageFormat, max_images);
   if (!formats)
     goto cleanup;
 
@@ -671,10 +677,38 @@ ensure_image_formats (GstVaapiDisplay * display)
   if (!vaapi_check_status (status, "vaQueryImageFormats()"))
     goto cleanup;
 
+  /* XXX(victor): Force RGBA in i965 display formats.
+   *
+   * This is required for GLTextureUploadMeta since it only negotiates
+   * RGBA, nevertheless i965 driver only reports RGBx breaking back
+   * compatibility.
+   *
+   * Side effects are not expected since it worked before commit
+   * 32bf6f1e */
+  if (gst_vaapi_display_has_driver_quirks (display,
+          GST_VAAPI_DRIVER_QUIRK_MISSING_RGBA_IMAGE_FORMAT)) {
+    formats = g_renew (VAImageFormat, formats, max_images + 1);
+
+    formats[n].fourcc = VA_FOURCC_RGBA;
+    formats[n].byte_order = VA_LSB_FIRST;
+    formats[n].bits_per_pixel = 32;
+    formats[n].depth = 32;
+    formats[n].red_mask = 0x000000ff;
+    formats[n].green_mask = 0x0000ff00;
+    formats[n].blue_mask = 0x00ff0000;
+    formats[n].alpha_mask = 0xff000000;
+    n++;
+  }
+
   GST_DEBUG ("%d image formats", n);
   for (i = 0; i < n; i++)
     GST_DEBUG ("  %" GST_FOURCC_FORMAT, GST_FOURCC_ARGS (formats[i].fourcc));
 
+  if (!gst_vaapi_video_format_create_map (formats, n)) {
+    GST_ERROR ("fail to create map between gst video format and vaImageFormat");
+    goto cleanup;
+  }
+
   append_formats (priv->image_formats, formats, NULL, n);
   g_array_sort (priv->image_formats, compare_yuv_formats);
   success = TRUE;
@@ -738,6 +772,62 @@ cleanup:
   return success;
 }
 
+/* Ensures the VA driver vendor string was copied */
+static gboolean
+ensure_vendor_string (GstVaapiDisplay * display)
+{
+  GstVaapiDisplayPrivate *const priv = GST_VAAPI_DISPLAY_GET_PRIVATE (display);
+  const gchar *vendor_string;
+
+  GST_VAAPI_DISPLAY_LOCK (display);
+  if (!priv->vendor_string) {
+    vendor_string = vaQueryVendorString (priv->display);
+    if (vendor_string)
+      priv->vendor_string = g_strdup (vendor_string);
+    GST_INFO_OBJECT (display, "vendor: %s", priv->vendor_string);
+  }
+  GST_VAAPI_DISPLAY_UNLOCK (display);
+  return priv->vendor_string != NULL;
+}
+
+static void
+set_driver_quirks (GstVaapiDisplay * display)
+{
+  GstVaapiDisplayPrivate *const priv = GST_VAAPI_DISPLAY_GET_PRIVATE (display);
+  guint i;
+
+  /* *INDENT-OFF* */
+  static const struct
+  {
+    const char *match_string;
+    guint quirks;
+  } vaapi_driver_quirks_table[] = {
+    /* @XXX(victor): is this string enough to identify it */
+    { "AMD", GST_VAAPI_DRIVER_QUIRK_NO_CHECK_SURFACE_PUT_IMAGE },
+    { "i965", GST_VAAPI_DRIVER_QUIRK_NO_CHECK_VPP_COLOR_STD },
+    { "iHD", GST_VAAPI_DRIVER_QUIRK_NO_RGBYUV_VPP_COLOR_PRIMARY },
+    { "i965", GST_VAAPI_DRIVER_QUIRK_MISSING_RGBA_IMAGE_FORMAT },
+    { "iHD", GST_VAAPI_DRIVER_QUIRK_JPEG_ENC_SHIFT_VALUE_BY_50 },
+    { "iHD", GST_VAAPI_DRIVER_QUIRK_HEVC_ENC_SLICE_NOT_SPAN_TILE },
+    { "i965", GST_VAAPI_DRIVER_QUIRK_JPEG_DEC_BROKEN_FORMATS },
+  };
+  /* *INDENT-ON* */
+
+  if (!ensure_vendor_string (display))
+    return;
+
+  for (i = 0; i < G_N_ELEMENTS (vaapi_driver_quirks_table); i++) {
+    const char *match_str = vaapi_driver_quirks_table[i].match_string;
+    if (g_strstr_len (priv->vendor_string, strlen (priv->vendor_string),
+            match_str) != NULL) {
+      priv->driver_quirks |= vaapi_driver_quirks_table[i].quirks;
+    }
+  }
+
+  GST_INFO_OBJECT (display, "Matched driver string \"%s\", setting quirks "
+      "(%#x)", priv->vendor_string, priv->driver_quirks);
+}
+
 static void
 gst_vaapi_display_calculate_pixel_aspect_ratio (GstVaapiDisplay * display)
 {
@@ -810,31 +900,14 @@ static void
 gst_vaapi_display_destroy (GstVaapiDisplay * display)
 {
   GstVaapiDisplayPrivate *const priv = GST_VAAPI_DISPLAY_GET_PRIVATE (display);
+  GstVaapiDisplayClass *klass = GST_VAAPI_DISPLAY_GET_CLASS (display);
 
-  if (priv->decoders) {
-    g_array_free (priv->decoders, TRUE);
-    priv->decoders = NULL;
-  }
-
-  if (priv->encoders) {
-    g_array_free (priv->encoders, TRUE);
-    priv->encoders = NULL;
-  }
-
-  if (priv->image_formats) {
-    g_array_free (priv->image_formats, TRUE);
-    priv->image_formats = NULL;
-  }
-
-  if (priv->subpicture_formats) {
-    g_array_free (priv->subpicture_formats, TRUE);
-    priv->subpicture_formats = NULL;
-  }
-
-  if (priv->properties) {
-    g_array_free (priv->properties, TRUE);
-    priv->properties = NULL;
-  }
+  g_clear_pointer (&priv->decoders, g_ptr_array_unref);
+  g_clear_pointer (&priv->encoders, g_ptr_array_unref);
+  g_clear_pointer (&priv->codecs, g_array_unref);
+  g_clear_pointer (&priv->image_formats, g_array_unref);
+  g_clear_pointer (&priv->subpicture_formats, g_array_unref);
+  g_clear_pointer (&priv->properties, g_array_unref);
 
   if (priv->display) {
     if (!priv->parent)
@@ -842,11 +915,8 @@ gst_vaapi_display_destroy (GstVaapiDisplay * display)
     priv->display = NULL;
   }
 
-  if (!priv->use_foreign_display) {
-    GstVaapiDisplayClass *klass = GST_VAAPI_DISPLAY_GET_CLASS (display);
-    if (klass->close_display)
-      klass->close_display (display);
-  }
+  if (klass->close_display)
+    klass->close_display (display);
 
   g_clear_pointer (&priv->display_name, g_free);
   g_clear_pointer (&priv->vendor_string, g_free);
@@ -906,6 +976,14 @@ gst_vaapi_display_create (GstVaapiDisplay * display,
   GST_INFO_OBJECT (display, "new display addr=%p", display);
   g_free (priv->display_name);
   priv->display_name = g_strdup (info.display_name);
+
+  set_driver_quirks (display);
+
+  if (!ensure_image_formats (display)) {
+    gst_vaapi_display_destroy (display);
+    return FALSE;
+  }
+
   return TRUE;
 }
 
@@ -1148,6 +1226,7 @@ gst_vaapi_display_class_init (GstVaapiDisplayClass * klass)
       "The display contrast value", 0.0, 2.0, 1.0, G_PARAM_READWRITE);
 
   g_object_class_install_properties (object_class, N_PROPERTIES, g_properties);
+  gst_type_mark_as_plugin_api (gst_vaapi_display_type_get_type (), 0);
 }
 
 /**
@@ -1480,7 +1559,7 @@ gst_vaapi_display_has_video_processing (GstVaapiDisplay * display)
  * reference to the resulting array of #GstVaapiProfile elements, so
  * it shall be released with g_array_unref() after usage.
  *
- * Return value: a newly allocated #GArray, or %NULL or error or if
+ * Return value: a newly allocated #GArray, or %NULL if error or if
  *   decoding is not supported at all
  */
 GArray *
@@ -1490,7 +1569,7 @@ gst_vaapi_display_get_decode_profiles (GstVaapiDisplay * display)
 
   if (!ensure_profiles (display))
     return NULL;
-  return get_profiles (GST_VAAPI_DISPLAY_GET_PRIVATE (display)->decoders);
+  return get_profiles (GST_VAAPI_DISPLAY_GET_PRIVATE (display)->decoders, 0);
 }
 
 /**
@@ -1524,7 +1603,7 @@ gst_vaapi_display_has_decoder (GstVaapiDisplay * display,
  * reference to the resulting array of #GstVaapiProfile elements, so
  * it shall be released with g_array_unref() after usage.
  *
- * Return value: a newly allocated #GArray, or %NULL or error or if
+ * Return value: a newly allocated #GArray, or %NULL if error or if
  *   encoding is not supported at all
  */
 GArray *
@@ -1534,7 +1613,32 @@ gst_vaapi_display_get_encode_profiles (GstVaapiDisplay * display)
 
   if (!ensure_profiles (display))
     return NULL;
-  return get_profiles (GST_VAAPI_DISPLAY_GET_PRIVATE (display)->encoders);
+  return get_profiles (GST_VAAPI_DISPLAY_GET_PRIVATE (display)->encoders, 0);
+}
+
+/**
+ * gst_vaapi_display_get_encode_profiles_by_codec:
+ * @display: a #GstVaapiDisplay
+ * @codec: a #GstVaapiCodec
+ *
+ * Gets the supported profiles which belongs to @codec for encoding.
+ * The caller owns an extra reference to the resulting array of
+ * #GstVaapiProfile elements, so it shall be released with g_array_unref()
+ * after usage.
+ *
+ * Return value: a newly allocated #GArray, or %NULL if error or if
+ *   no encoding profile is found specified by the @codec.
+ */
+GArray *
+gst_vaapi_display_get_encode_profiles_by_codec (GstVaapiDisplay * display,
+    GstVaapiCodec codec)
+{
+  g_return_val_if_fail (display != NULL, NULL);
+
+  if (!ensure_profiles (display))
+    return NULL;
+  return get_profiles (GST_VAAPI_DISPLAY_GET_PRIVATE (display)->encoders,
+      codec);
 }
 
 /**
@@ -1767,26 +1871,6 @@ get_render_mode_VADisplayAttribRenderMode (GstVaapiDisplay * display,
   return TRUE;
 }
 
-static gboolean
-get_render_mode_VADisplayAttribDirectSurface (GstVaapiDisplay * display,
-    GstVaapiRenderMode * pmode)
-{
-#if VA_CHECK_VERSION(0,34,0)
-  /* VADisplayAttribDirectsurface was removed in VA-API >= 0.34.0 */
-  return FALSE;
-#else
-  gint direct_surface;
-
-  if (!get_attribute (display, VADisplayAttribDirectSurface, &direct_surface))
-    return FALSE;
-  if (direct_surface)
-    *pmode = GST_VAAPI_RENDER_MODE_OVERLAY;
-  else
-    *pmode = GST_VAAPI_RENDER_MODE_TEXTURE;
-  return TRUE;
-#endif
-}
-
 static gboolean
 get_render_mode_default (GstVaapiDisplay * display, GstVaapiRenderMode * pmode)
 {
@@ -1830,10 +1914,6 @@ gst_vaapi_display_get_render_mode (GstVaapiDisplay * display,
   if (get_render_mode_VADisplayAttribRenderMode (display, pmode))
     return TRUE;
 
-  /* Try with direct-surface attribute */
-  if (get_render_mode_VADisplayAttribDirectSurface (display, pmode))
-    return TRUE;
-
   /* Default: determine from the display type */
   return get_render_mode_default (display, pmode);
 }
@@ -2005,23 +2085,6 @@ set_color_balance (GstVaapiDisplay * display, guint prop_id, gfloat v)
   return TRUE;
 }
 
-/* Ensures the VA driver vendor string was copied */
-static gboolean
-ensure_vendor_string (GstVaapiDisplay * display)
-{
-  GstVaapiDisplayPrivate *const priv = GST_VAAPI_DISPLAY_GET_PRIVATE (display);
-  const gchar *vendor_string;
-
-  GST_VAAPI_DISPLAY_LOCK (display);
-  if (!priv->vendor_string) {
-    vendor_string = vaQueryVendorString (priv->display);
-    if (vendor_string)
-      priv->vendor_string = g_strdup (vendor_string);
-  }
-  GST_VAAPI_DISPLAY_UNLOCK (display);
-  return priv->vendor_string != NULL;
-}
-
 /**
  * gst_vaapi_display_get_vendor_string:
  * @display: a #GstVaapiDisplay
@@ -2091,3 +2154,18 @@ gst_vaapi_display_reset_texture_map (GstVaapiDisplay * display)
   if ((map = klass->get_texture_map (display)))
     gst_vaapi_texture_map_reset (map);
 }
+
+/**
+ * gst_vaapi_display_get_driver_quirks:
+ * @display: a #GstVaapiDisplay
+ * @quirks: the #GstVaapiDriverQuirks bitwise to check
+ *
+ * Returns: %TRUE if @quirks are set in @display's driver
+ **/
+gboolean
+gst_vaapi_display_has_driver_quirks (GstVaapiDisplay * display, guint quirks)
+{
+  g_return_val_if_fail (display != NULL, FALSE);
+
+  return (GST_VAAPI_DISPLAY_GET_PRIVATE (display)->driver_quirks & quirks);
+}
