@@ -27,10 +27,6 @@
 #include <gst/vaapi/gstvaapidecoder_objects.h>
 #include <gst/vaapi/gstvaapiencoder.h>
 
-#if USE_H264_FEI_ENCODER
-#include <gst/vaapi/gstvaapifei_objects.h>
-#endif
-
 G_BEGIN_DECLS
 
 typedef struct _GstVaapiEncPicture GstVaapiEncPicture;
@@ -270,14 +266,6 @@ struct _GstVaapiEncPicture
   guint frame_num;
   guint poc;
   guint temporal_id;
-#if USE_H264_FEI_ENCODER
-  GstVaapiEncFeiMbControl *mbcntrl;
-  GstVaapiEncFeiMvPredictor *mvpred;
-  GstVaapiEncFeiQp *qp;
-  GstVaapiEncFeiMbCode *mbcode;
-  GstVaapiEncFeiMv *mv;
-  GstVaapiEncFeiDistortion *dist;
-#endif
 };
 
 G_GNUC_INTERNAL
@@ -342,6 +330,12 @@ gst_vaapi_enc_picture_encode (GstVaapiEncPicture * picture);
       VAEncMiscParameterTypeQualityLevel,                               \
       sizeof (VAEncMiscParameterBufferQualityLevel))
 
+/* GstVaapiEncQuantizationMiscParam */
+#define GST_VAAPI_ENC_QUANTIZATION_MISC_PARAM_NEW(encoder)              \
+  gst_vaapi_enc_misc_param_new (GST_VAAPI_ENCODER_CAST (encoder),       \
+      VAEncMiscParameterTypeQuantization,                               \
+      sizeof (VAEncMiscParameterQuantization))
+
 /* GstVaapiEncPicture  */
 #define GST_VAAPI_ENC_PICTURE_NEW(codec, encoder, frame)                \
   gst_vaapi_enc_picture_new (GST_VAAPI_ENCODER_CAST (encoder),          \
