@@ -27,34 +27,23 @@
 
 G_BEGIN_DECLS
 
+#define GST_TYPE_VAAPI_ENCODER_VP9 \
+    (gst_vaapi_encoder_vp9_get_type ())
 #define GST_VAAPI_ENCODER_VP9(encoder) \
-  ((GstVaapiEncoderVP9 *) (encoder))
+    (G_TYPE_CHECK_INSTANCE_CAST ((encoder), GST_TYPE_VAAPI_ENCODER_VP9, GstVaapiEncoderVP9))
+#define GST_IS_VAAPI_ENCODER_VP9(encoder) \
+    (G_TYPE_CHECK_INSTANCE_TYPE ((encoder), GST_TYPE_VAAPI_ENCODER_VP9))
 
 typedef struct _GstVaapiEncoderVP9 GstVaapiEncoderVP9;
+typedef struct _GstVaapiEncoderVP9Class GstVaapiEncoderVP9Class;
 
-/**
- * GstVaapiEncoderVP9Prop:
- * @GST_VAAPI_ENCODER_VP9_PROP_LOOP_FILTER_LEVEL: Loop Filter Level(uint).
- * @GST_VAAPI_ENCODER_VP9_PROP_LOOP_SHARPNESS_LEVEL: Sharpness Level(uint).
- * @GST_VAAPI_ENCODER_VP9_PROP_YAC_Q_INDEX: Quantization table index for luma AC
- * @GST_VAAPI_ENCODER_VP9_PROP_REF_PIC_MODE: Reference picute selection modes
- * @GST_VAAPI_ENCODER_VP9_PROP_CPB_LENGTH:Length of CPB buffer in milliseconds
- *
- * The set of VP9 encoder specific configurable properties.
- */
-typedef enum {
-  GST_VAAPI_ENCODER_VP9_PROP_LOOP_FILTER_LEVEL = -1,
-  GST_VAAPI_ENCODER_VP9_PROP_SHARPNESS_LEVEL = -2,
-  GST_VAAPI_ENCODER_VP9_PROP_YAC_Q_INDEX = -3,
-  GST_VAAPI_ENCODER_VP9_PROP_REF_PIC_MODE = -4,
-  GST_VAAPI_ENCODER_VP9_PROP_CPB_LENGTH = -5
-} GstVaapiEncoderVP9Prop;
+GType
+gst_vaapi_encoder_vp9_get_type (void) G_GNUC_CONST;
 
 GstVaapiEncoder *
 gst_vaapi_encoder_vp9_new (GstVaapiDisplay * display);
 
-GPtrArray *
-gst_vaapi_encoder_vp9_get_default_properties (void);
+G_DEFINE_AUTOPTR_CLEANUP_FUNC(GstVaapiEncoderVP9, gst_object_unref)
 
 G_END_DECLS
 #endif /*GST_VAAPI_ENCODER_VP9_H */
