@@ -61,13 +61,23 @@ struct _GstVaapiFilter
   GPtrArray *operations;
   GstVideoFormat format;
   GstVaapiScaleMethod scale_method;
-  GArray *formats;
+  GstVideoOrientationMethod video_direction;
+  GstVaapiConfigSurfaceAttributes *attribs;
   GArray *forward_references;
   GArray *backward_references;
   GstVaapiRectangle crop_rect;
   GstVaapiRectangle target_rect;
   guint use_crop_rect:1;
   guint use_target_rect:1;
+  guint32 mirror_flags;
+  guint32 rotation_flags;
+
+  GstVideoColorimetry input_colorimetry;
+  GstVideoColorimetry output_colorimetry;
+
+#if VA_CHECK_VERSION(1,4,0)
+  VAHdrMetaDataHDR10 hdr_meta;
+#endif
 };
 
 typedef struct _GstVaapiFilterClass GstVaapiFilterClass;
@@ -111,6 +121,8 @@ gst_vaapi_scale_method_get_type (void)
     const GType type =
         g_enum_register_static ("GstVaapiScaleMethod", enum_values);
     g_once_init_leave (&g_type, type);
+
+    gst_type_mark_as_plugin_api (type, 0);
   }
   return g_type;
 }
@@ -137,6 +149,7 @@ gst_vaapi_deinterlace_method_get_type (void)
   if (g_once_init_enter (&g_type)) {
     const GType type =
         g_enum_register_static ("GstVaapiDeinterlaceMethod", enum_values);
+    gst_type_mark_as_plugin_api (type, 0);
     g_once_init_leave (&g_type, type);
   }
   return g_type;
@@ -160,6 +173,7 @@ gst_vaapi_deinterlace_flags_get_type (void)
   if (g_once_init_enter (&g_type)) {
     const GType type =
         g_enum_register_static ("GstVaapiDeinterlaceFlags", enum_values);
+    gst_type_mark_as_plugin_api (type, 0);
     g_once_init_leave (&g_type, type);
   }
   return g_type;
@@ -271,12 +285,48 @@ vpp_get_filter_caps (GstVaapiFilter * filter, VAProcFilterType type,
   return caps;
 }
 
+static void
+vpp_get_pipeline_caps_unlocked (GstVaapiFilter * filter)
+{
+#if VA_CHECK_VERSION(1,1,0)
+  VAProcPipelineCaps pipeline_caps = { 0, };
+
+  VAStatus va_status = vaQueryVideoProcPipelineCaps (filter->va_display,
+      filter->va_context, NULL, 0, &pipeline_caps);
+
+  if (vaapi_check_status (va_status, "vaQueryVideoProcPipelineCaps()")) {
+    filter->mirror_flags = pipeline_caps.mirror_flags;
+    filter->rotation_flags = pipeline_caps.rotation_flags;
+    return;
+  }
+#endif
+
+  filter->mirror_flags = 0;
+  filter->rotation_flags = 0;
+}
+
+static void
+vpp_get_pipeline_caps (GstVaapiFilter * filter)
+{
+  GST_VAAPI_DISPLAY_LOCK (filter->display);
+  vpp_get_pipeline_caps_unlocked (filter);
+  GST_VAAPI_DISPLAY_UNLOCK (filter->display);
+}
+
 /* ------------------------------------------------------------------------- */
 /* --- VPP Operations                                                   --- */
 /* ------------------------------------------------------------------------- */
 
 #define DEFAULT_FORMAT  GST_VIDEO_FORMAT_UNKNOWN
-#define DEFAULT_SCALING GST_VAAPI_SCALE_METHOD_DEFAULT
+
+#define OP_DATA_DEFAULT_VALUE(type, op_data) \
+    g_value_get_##type (g_param_spec_get_default_value (op_data->pspec))
+
+#define OP_RET_DEFAULT_VALUE(type, filter, op) \
+    do { \
+      g_return_val_if_fail (filter != NULL, FALSE); \
+      return OP_DATA_DEFAULT_VALUE (type, find_operation (filter, op)); \
+    } while (0)
 
 enum
 {
@@ -297,7 +347,12 @@ enum
   PROP_CONTRAST = GST_VAAPI_FILTER_OP_CONTRAST,
   PROP_DEINTERLACING = GST_VAAPI_FILTER_OP_DEINTERLACING,
   PROP_SCALING = GST_VAAPI_FILTER_OP_SCALING,
+  PROP_VIDEO_DIRECTION = GST_VAAPI_FILTER_OP_VIDEO_DIRECTION,
+  PROP_HDR_TONE_MAP = GST_VAAPI_FILTER_OP_HDR_TONE_MAP,
+#ifndef GST_REMOVE_DEPRECATED
   PROP_SKINTONE = GST_VAAPI_FILTER_OP_SKINTONE,
+#endif
+  PROP_SKINTONE_LEVEL = GST_VAAPI_FILTER_OP_SKINTONE_LEVEL,
 
   N_PROPERTIES
 };
@@ -418,8 +473,33 @@ init_properties (void)
       "Scaling Method",
       "Scaling method to use",
       GST_VAAPI_TYPE_SCALE_METHOD,
-      DEFAULT_SCALING, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+      GST_VAAPI_SCALE_METHOD_DEFAULT,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
 
+  /**
+   * GstVaapiFilter:video-direction:
+   *
+   * The video-direction to use, expressed as an enum value. See
+   * #GstVideoOrientationMethod.
+   */
+  g_properties[PROP_VIDEO_DIRECTION] = g_param_spec_enum ("video-direction",
+      "Video Direction",
+      "Video direction: rotation and flipping",
+      GST_TYPE_VIDEO_ORIENTATION_METHOD,
+      GST_VIDEO_ORIENTATION_IDENTITY,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+
+  /**
+   * GstVaapiFilter:tone-map:
+   *
+   * Apply HDR tone mapping
+   **/
+  g_properties[PROP_HDR_TONE_MAP] = g_param_spec_boolean ("hdr-tone-map",
+      "HDR Tone Mapping",
+      "Apply HDR tone mapping",
+      FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+
+#ifndef GST_REMOVE_DEPRECATED
   /**
    * GstVaapiFilter:skin-tone-enhancement:
    *
@@ -429,6 +509,18 @@ init_properties (void)
       "Skin tone enhancement",
       "Apply the skin tone enhancement algorithm",
       FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+#endif
+
+  /**
+   * GstVaapiFilter:skin-tone-enhancement-level:
+   *
+   * Apply the skin tone enhancement algorithm with specified value.
+   */
+  g_properties[PROP_SKINTONE_LEVEL] =
+      g_param_spec_uint ("skin-tone-enhancement-level",
+      "Skin tone enhancement level",
+      "Apply the skin tone enhancement algorithm with specified level", 0, 9, 3,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
 }
 
 static void
@@ -462,9 +554,22 @@ op_data_new (GstVaapiFilterOp op, GParamSpec * pspec)
   op_data->va_buffer = VA_INVALID_ID;
 
   switch (op) {
+    case GST_VAAPI_FILTER_OP_HDR_TONE_MAP:
+#if VA_CHECK_VERSION(1,4,0)
+      /* Only HDR10 tone mapping is supported */
+      op_data->va_type = VAProcFilterHighDynamicRangeToneMapping;
+      op_data->va_subtype = VAProcHighDynamicRangeMetadataHDR10;
+      op_data->va_cap_size = sizeof (VAProcFilterCapHighDynamicRange);
+      op_data->va_buffer_size =
+          sizeof (VAProcFilterParameterBufferHDRToneMapping);
+      break;
+#else
+      /* fall-through */
+#endif
     case GST_VAAPI_FILTER_OP_FORMAT:
     case GST_VAAPI_FILTER_OP_CROP:
     case GST_VAAPI_FILTER_OP_SCALING:
+    case GST_VAAPI_FILTER_OP_VIDEO_DIRECTION:
       op_data->va_type = VAProcFilterNone;
       break;
     case GST_VAAPI_FILTER_OP_DENOISE:
@@ -477,7 +582,10 @@ op_data_new (GstVaapiFilterOp op, GParamSpec * pspec)
       op_data->va_cap_size = sizeof (VAProcFilterCap);
       op_data->va_buffer_size = sizeof (VAProcFilterParameterBuffer);
       break;
+#ifndef GST_REMOVE_DEPRECATED
     case GST_VAAPI_FILTER_OP_SKINTONE:
+#endif
+    case GST_VAAPI_FILTER_OP_SKINTONE_LEVEL:
       op_data->va_type = VAProcFilterSkinToneEnhancement;
       op_data->va_buffer_size = sizeof (VAProcFilterParameterBuffer);
       break;
@@ -688,6 +796,8 @@ get_operations_ordered (GstVaapiFilter * filter, GPtrArray * default_ops)
     filter_caps = NULL;
   }
 
+  vpp_get_pipeline_caps (filter);
+
   if (filter->operations)
     g_ptr_array_unref (filter->operations);
   filter->operations = g_ptr_array_ref (ops);
@@ -764,13 +874,20 @@ find_operation (GstVaapiFilter * filter, GstVaapiFilterOp op)
 
 /* Ensure the operation's VA buffer is allocated */
 static inline gboolean
-op_ensure_buffer (GstVaapiFilter * filter, GstVaapiFilterOpData * op_data)
+op_ensure_n_elements_buffer (GstVaapiFilter * filter,
+    GstVaapiFilterOpData * op_data, gint op_num)
 {
   if (G_LIKELY (op_data->va_buffer != VA_INVALID_ID))
     return TRUE;
-  return vaapi_create_buffer (filter->va_display, filter->va_context,
+  return vaapi_create_n_elements_buffer (filter->va_display, filter->va_context,
       VAProcFilterParameterBufferType, op_data->va_buffer_size, NULL,
-      &op_data->va_buffer, NULL);
+      &op_data->va_buffer, NULL, op_num);
+}
+
+static inline gboolean
+op_ensure_buffer (GstVaapiFilter * filter, GstVaapiFilterOpData * op_data)
+{
+  return op_ensure_n_elements_buffer (filter, op_data, 1);
 }
 
 /* Update a generic filter (float value) */
@@ -785,8 +902,7 @@ op_set_generic_unlocked (GstVaapiFilter * filter,
   if (!op_data || !op_ensure_buffer (filter, op_data))
     return FALSE;
 
-  op_data->is_enabled =
-      (value != G_PARAM_SPEC_FLOAT (op_data->pspec)->default_value);
+  op_data->is_enabled = (value != OP_DATA_DEFAULT_VALUE (float, op_data));
   if (!op_data->is_enabled)
     return TRUE;
 
@@ -817,6 +933,9 @@ op_set_generic (GstVaapiFilter * filter, GstVaapiFilterOpData * op_data,
 }
 
 /* Update the color balance filter */
+#define COLOR_BALANCE_NUM \
+    GST_VAAPI_FILTER_OP_CONTRAST - GST_VAAPI_FILTER_OP_HUE + 1
+
 static gboolean
 op_set_color_balance_unlocked (GstVaapiFilter * filter,
     GstVaapiFilterOpData * op_data, gfloat value)
@@ -824,28 +943,86 @@ op_set_color_balance_unlocked (GstVaapiFilter * filter,
   VAProcFilterParameterBufferColorBalance *buf;
   VAProcFilterCapColorBalance *filter_cap;
   gfloat va_value;
+  gint i;
+  GstVaapiFilterOpData *color_data[COLOR_BALANCE_NUM];
+  GstVaapiFilterOpData *enabled_data = NULL;
+  gboolean ret = TRUE;
 
-  if (!op_data || !op_ensure_buffer (filter, op_data))
+  if (!op_data)
     return FALSE;
 
-  op_data->is_enabled =
-      (value != G_PARAM_SPEC_FLOAT (op_data->pspec)->default_value);
-  if (!op_data->is_enabled)
-    return TRUE;
+  /* collect all the Color Balance operators and find the first
+   * enabled one */
+  for (i = 0; i < COLOR_BALANCE_NUM; i++) {
+    color_data[i] = find_operation (filter, GST_VAAPI_FILTER_OP_HUE + i);
+    if (!color_data[i])
+      return FALSE;
 
-  filter_cap = op_data->va_caps;
-  if (!op_data_get_value_float (op_data, &filter_cap->range, value, &va_value))
-    return FALSE;
+    if (!enabled_data && color_data[i]->is_enabled)
+      enabled_data = color_data[i];
+  }
 
-  buf = vaapi_map_buffer (filter->va_display, op_data->va_buffer);
-  if (!buf)
-    return FALSE;
+  /* If there's no enabled operators let's enable this one.
+   *
+   * HACK: This operator will be the only one with an allocated buffer
+   * which will store all the color balance operators.
+   */
+  if (!enabled_data) {
+    /* *INDENT-OFF* */
+    if (value == OP_DATA_DEFAULT_VALUE (float, op_data))
+      return TRUE;
+    /* *INDENT-ON* */
 
-  buf->type = op_data->va_type;
-  buf->attrib = op_data->va_subtype;
-  buf->value = va_value;
-  vaapi_unmap_buffer (filter->va_display, op_data->va_buffer, NULL);
-  return TRUE;
+    if (!op_ensure_n_elements_buffer (filter, op_data, COLOR_BALANCE_NUM))
+      return FALSE;
+
+    enabled_data = op_data;
+
+    buf = vaapi_map_buffer (filter->va_display, enabled_data->va_buffer);
+    if (!buf)
+      return FALSE;
+
+    /* Write all the color balance operator values in the buffer. --
+     * Use the default value for all the operators except the set
+     * one. */
+    for (i = 0; i < COLOR_BALANCE_NUM; i++) {
+      buf[i].type = color_data[i]->va_type;
+      buf[i].attrib = color_data[i]->va_subtype;
+
+      va_value = OP_DATA_DEFAULT_VALUE (float, color_data[i]);
+      if (color_data[i]->op == op_data->op) {
+        filter_cap = color_data[i]->va_caps;
+        /* fail but ignore current value and set default one */
+        if (!op_data_get_value_float (color_data[i], &filter_cap->range, value,
+                &va_value))
+          ret = FALSE;
+      }
+
+      buf[i].value = va_value;
+    }
+
+    enabled_data->is_enabled = 1;
+  } else {
+    /* There's already one operator enabled, *in theory* with a
+     * buffer associated. */
+    if (G_UNLIKELY (enabled_data->va_buffer == VA_INVALID_ID))
+      return FALSE;
+
+    filter_cap = op_data->va_caps;
+    if (!op_data_get_value_float (op_data, &filter_cap->range, value,
+            &va_value))
+      return FALSE;
+
+    buf = vaapi_map_buffer (filter->va_display, enabled_data->va_buffer);
+    if (!buf)
+      return FALSE;
+
+    buf[op_data->op - GST_VAAPI_FILTER_OP_HUE].value = va_value;
+  }
+
+  vaapi_unmap_buffer (filter->va_display, enabled_data->va_buffer, NULL);
+
+  return ret;
 }
 
 static inline gboolean
@@ -909,29 +1086,56 @@ op_set_deinterlace (GstVaapiFilter * filter, GstVaapiFilterOpData * op_data,
   return success;
 }
 
-/* Update skin tone enhancement */
+/* Update skin tone enhancement level */
 static gboolean
-op_set_skintone_unlocked (GstVaapiFilter * filter,
-    GstVaapiFilterOpData * op_data, gboolean value)
+op_set_skintone_level_unlocked (GstVaapiFilter * filter,
+    GstVaapiFilterOpData * op_data, guint value)
 {
   VAProcFilterParameterBuffer *buf;
 
   if (!op_data || !op_ensure_buffer (filter, op_data))
     return FALSE;
 
-  op_data->is_enabled = value;
-  if (!op_data->is_enabled)
-    return TRUE;
+  op_data->is_enabled = 1;
 
   buf = vaapi_map_buffer (filter->va_display, op_data->va_buffer);
   if (!buf)
     return FALSE;
   buf->type = op_data->va_type;
-  buf->value = 0;
+  buf->value = value;
   vaapi_unmap_buffer (filter->va_display, op_data->va_buffer, NULL);
   return TRUE;
 }
 
+static inline gboolean
+op_set_skintone_level (GstVaapiFilter * filter,
+    GstVaapiFilterOpData * op_data, guint value)
+{
+  gboolean success = FALSE;
+
+  GST_VAAPI_DISPLAY_LOCK (filter->display);
+  success = op_set_skintone_level_unlocked (filter, op_data, value);
+  GST_VAAPI_DISPLAY_UNLOCK (filter->display);
+  return success;
+}
+
+#ifndef GST_REMOVE_DEPRECATED
+/* Update skin tone enhancement */
+static gboolean
+op_set_skintone_unlocked (GstVaapiFilter * filter,
+    GstVaapiFilterOpData * op_data, gboolean value)
+{
+  if (!op_data)
+    return FALSE;
+
+  if (!value) {
+    op_data->is_enabled = 0;
+    return TRUE;
+  }
+
+  return op_set_skintone_level_unlocked (filter, op_data, 3);
+}
+
 static inline gboolean
 op_set_skintone (GstVaapiFilter * filter, GstVaapiFilterOpData * op_data,
     gboolean enhance)
@@ -943,7 +1147,54 @@ op_set_skintone (GstVaapiFilter * filter, GstVaapiFilterOpData * op_data,
   GST_VAAPI_DISPLAY_UNLOCK (filter->display);
   return success;
 }
+#endif
 
+static gboolean
+op_set_hdr_tone_map_unlocked (GstVaapiFilter * filter,
+    GstVaapiFilterOpData * op_data, gboolean value)
+{
+#if VA_CHECK_VERSION(1,4,0)
+  const VAProcFilterCapHighDynamicRange *filter_caps;
+  guint i;
+
+  if (!op_data)
+    return !value;
+
+  if (!value) {
+    op_data->is_enabled = 0;
+    return TRUE;
+  }
+
+  if (!op_ensure_buffer (filter, op_data))
+    return FALSE;
+
+  for (i = 0, filter_caps = op_data->va_caps; i < op_data->va_num_caps; i++) {
+    if (filter_caps[i].metadata_type == op_data->va_subtype &&
+        (filter_caps[i].caps_flag & VA_TONE_MAPPING_HDR_TO_SDR))
+      break;
+  }
+  if (i == op_data->va_num_caps)
+    return FALSE;
+
+  op_data->is_enabled = 1;
+
+  return TRUE;
+#else
+  return !value;
+#endif
+}
+
+static inline gboolean
+op_set_hdr_tone_map (GstVaapiFilter * filter, GstVaapiFilterOpData * op_data,
+    gboolean value)
+{
+  gboolean success = FALSE;
+  GST_VAAPI_DISPLAY_LOCK (filter->display);
+  success = op_set_hdr_tone_map_unlocked (filter, op_data, value);
+  GST_VAAPI_DISPLAY_UNLOCK (filter->display);
+
+  return success;
+}
 
 static gboolean
 deint_refs_set (GArray * refs, GstVaapiSurface ** surfaces, guint num_surfaces)
@@ -954,7 +1205,7 @@ deint_refs_set (GArray * refs, GstVaapiSurface ** surfaces, guint num_surfaces)
     return FALSE;
 
   for (i = 0; i < num_surfaces; i++)
-    g_array_append_val (refs, GST_VAAPI_OBJECT_ID (surfaces[i]));
+    g_array_append_val (refs, GST_VAAPI_SURFACE_ID (surfaces[i]));
   return TRUE;
 }
 
@@ -973,18 +1224,18 @@ deint_refs_clear_all (GstVaapiFilter * filter)
 }
 
 /* ------------------------------------------------------------------------- */
-/* --- Surface Formats                                                   --- */
+/* --- Surface Attribs                                                   --- */
 /* ------------------------------------------------------------------------- */
 
 static gboolean
-ensure_formats (GstVaapiFilter * filter)
+ensure_attributes (GstVaapiFilter * filter)
 {
-  if (G_LIKELY (filter->formats))
+  if (G_LIKELY (filter->attribs))
     return TRUE;
 
-  filter->formats = gst_vaapi_get_surface_formats (filter->display,
+  filter->attribs = gst_vaapi_config_surface_attributes_get (filter->display,
       filter->va_config);
-  return (filter->formats != NULL);
+  return (filter->attribs != NULL);
 }
 
 static inline gboolean
@@ -998,12 +1249,14 @@ static gboolean
 find_format (GstVaapiFilter * filter, GstVideoFormat format)
 {
   guint i;
+  GArray *formats;
 
-  if (is_special_format (format) || !filter->formats)
+  formats = filter->attribs->formats;
+  if (is_special_format (format) || !formats)
     return FALSE;
 
-  for (i = 0; i < filter->formats->len; i++) {
-    if (g_array_index (filter->formats, GstVideoFormat, i) == format)
+  for (i = 0; i < formats->len; i++) {
+    if (g_array_index (formats, GstVideoFormat, i) == format)
       return TRUE;
   }
   return FALSE;
@@ -1044,6 +1297,10 @@ gst_vaapi_filter_initialize (GstVaapiFilter * filter)
       NULL, 0, &filter->va_context);
   if (!vaapi_check_status (va_status, "vaCreateContext() [VPP]"))
     return FALSE;
+
+  gst_video_colorimetry_from_string (&filter->input_colorimetry, NULL);
+  gst_video_colorimetry_from_string (&filter->output_colorimetry, NULL);
+
   return TRUE;
 }
 
@@ -1053,6 +1310,9 @@ gst_vaapi_filter_finalize (GObject * object)
   GstVaapiFilter *const filter = GST_VAAPI_FILTER (object);
   guint i;
 
+  if (!filter->display)
+    goto bail;
+
   GST_VAAPI_DISPLAY_LOCK (filter->display);
   if (filter->operations) {
     for (i = 0; i < filter->operations->len; i++) {
@@ -1076,6 +1336,7 @@ gst_vaapi_filter_finalize (GObject * object)
   GST_VAAPI_DISPLAY_UNLOCK (filter->display);
   gst_vaapi_display_replace (&filter->display, NULL);
 
+bail:
   if (filter->forward_references) {
     g_array_unref (filter->forward_references);
     filter->forward_references = NULL;
@@ -1086,9 +1347,9 @@ gst_vaapi_filter_finalize (GObject * object)
     filter->backward_references = NULL;
   }
 
-  if (filter->formats) {
-    g_array_unref (filter->formats);
-    filter->formats = NULL;
+  if (filter->attribs) {
+    gst_vaapi_config_surface_attributes_free (filter->attribs);
+    filter->attribs = NULL;
   }
 
   G_OBJECT_CLASS (gst_vaapi_filter_parent_class)->finalize (object);
@@ -1310,32 +1571,99 @@ gst_vaapi_filter_set_operation (GstVaapiFilter * filter, GstVaapiFilterOp op,
     case GST_VAAPI_FILTER_OP_SHARPEN:
       return op_set_generic (filter, op_data,
           (value ? g_value_get_float (value) :
-              G_PARAM_SPEC_FLOAT (op_data->pspec)->default_value));
+              OP_DATA_DEFAULT_VALUE (float, op_data)));
     case GST_VAAPI_FILTER_OP_HUE:
     case GST_VAAPI_FILTER_OP_SATURATION:
     case GST_VAAPI_FILTER_OP_BRIGHTNESS:
     case GST_VAAPI_FILTER_OP_CONTRAST:
       return op_set_color_balance (filter, op_data,
           (value ? g_value_get_float (value) :
-              G_PARAM_SPEC_FLOAT (op_data->pspec)->default_value));
+              OP_DATA_DEFAULT_VALUE (float, op_data)));
     case GST_VAAPI_FILTER_OP_DEINTERLACING:
       return op_set_deinterlace (filter, op_data,
           (value ? g_value_get_enum (value) :
-              G_PARAM_SPEC_ENUM (op_data->pspec)->default_value), 0);
+              OP_DATA_DEFAULT_VALUE (enum, op_data)), 0);
       break;
     case GST_VAAPI_FILTER_OP_SCALING:
-      return gst_vaapi_filter_set_scaling (filter, value ?
-          g_value_get_enum (value) : DEFAULT_SCALING);
+      return gst_vaapi_filter_set_scaling (filter,
+          (value ? g_value_get_enum (value) :
+              OP_DATA_DEFAULT_VALUE (enum, op_data)));
+#ifndef GST_REMOVE_DEPRECATED
     case GST_VAAPI_FILTER_OP_SKINTONE:
       return op_set_skintone (filter, op_data,
           (value ? g_value_get_boolean (value) :
-              G_PARAM_SPEC_BOOLEAN (op_data->pspec)->default_value));
+              OP_DATA_DEFAULT_VALUE (boolean, op_data)));
+#endif
+    case GST_VAAPI_FILTER_OP_SKINTONE_LEVEL:
+      return op_set_skintone_level (filter, op_data,
+          (value ? g_value_get_uint (value) :
+              OP_DATA_DEFAULT_VALUE (uint, op_data)));
+    case GST_VAAPI_FILTER_OP_VIDEO_DIRECTION:
+      return gst_vaapi_filter_set_video_direction (filter,
+          (value ? g_value_get_enum (value) :
+              OP_DATA_DEFAULT_VALUE (enum, op_data)));
+    case GST_VAAPI_FILTER_OP_HDR_TONE_MAP:
+      return op_set_hdr_tone_map (filter, op_data,
+          (value ? g_value_get_boolean (value) :
+              OP_DATA_DEFAULT_VALUE (boolean, op_data)));
     default:
       break;
   }
   return FALSE;
 }
 
+#if VA_CHECK_VERSION(1,2,0)
+static void
+fill_color_standard (GstVideoColorimetry * colorimetry,
+    VAProcColorStandardType * type, VAProcColorProperties * properties)
+{
+  *type = from_GstVideoColorimetry (colorimetry);
+
+  properties->colour_primaries =
+      gst_video_color_primaries_to_iso (colorimetry->primaries);
+  properties->transfer_characteristics =
+      gst_video_transfer_function_to_iso (colorimetry->transfer);
+  properties->matrix_coefficients =
+      gst_video_color_matrix_to_iso (colorimetry->matrix);
+
+  properties->color_range = from_GstVideoColorRange (colorimetry->range);
+}
+#endif
+
+static void
+gst_vaapi_filter_fill_color_standards (GstVaapiFilter * filter,
+    VAProcPipelineParameterBuffer * pipeline_param)
+{
+#if VA_CHECK_VERSION(1,2,0)
+  fill_color_standard (&filter->input_colorimetry,
+      &pipeline_param->surface_color_standard,
+      &pipeline_param->input_color_properties);
+
+  fill_color_standard (&filter->output_colorimetry,
+      &pipeline_param->output_color_standard,
+      &pipeline_param->output_color_properties);
+
+  /* Handle RGB <-> YUV color primary driver quirk */
+  if (gst_vaapi_display_has_driver_quirks (filter->display,
+          GST_VAAPI_DRIVER_QUIRK_NO_RGBYUV_VPP_COLOR_PRIMARY)) {
+    gboolean src_is_rgb = gst_video_colorimetry_matches
+        (&filter->input_colorimetry, GST_VIDEO_COLORIMETRY_SRGB);
+    gboolean dst_is_rgb = gst_video_colorimetry_matches
+        (&filter->output_colorimetry, GST_VIDEO_COLORIMETRY_SRGB);
+
+    if ((!src_is_rgb && dst_is_rgb) || (src_is_rgb && !dst_is_rgb)) {
+      pipeline_param->output_color_standard = VAProcColorStandardExplicit;
+      pipeline_param->output_color_properties.colour_primaries =
+          gst_video_color_primaries_to_iso (filter->
+          input_colorimetry.primaries);
+    }
+  }
+#else
+  pipeline_param->surface_color_standard = VAProcColorStandardNone;
+  pipeline_param->output_color_standard = VAProcColorStandardNone;
+#endif
+}
+
 /**
  * gst_vaapi_filter_process:
  * @filter: a #GstVaapiFilter
@@ -1362,6 +1690,7 @@ gst_vaapi_filter_process_unlocked (GstVaapiFilter * filter,
   guint i, num_filters = 0;
   VAStatus va_status;
   VARectangle src_rect, dst_rect;
+  guint va_mirror = 0, va_rotation = 0;
 
   if (!ensure_operations (filter))
     return GST_VAAPI_FILTER_STATUS_ERROR_ALLOCATION_FAILED;
@@ -1433,17 +1762,26 @@ gst_vaapi_filter_process_unlocked (GstVaapiFilter * filter,
     goto error;
 
   memset (pipeline_param, 0, sizeof (*pipeline_param));
-  pipeline_param->surface = GST_VAAPI_OBJECT_ID (src_surface);
+  pipeline_param->surface = GST_VAAPI_SURFACE_ID (src_surface);
   pipeline_param->surface_region = &src_rect;
-  pipeline_param->surface_color_standard = VAProcColorStandardNone;
+
+  gst_vaapi_filter_fill_color_standards (filter, pipeline_param);
+
   pipeline_param->output_region = &dst_rect;
-  pipeline_param->output_color_standard = VAProcColorStandardNone;
   pipeline_param->output_background_color = 0xff000000;
   pipeline_param->filter_flags = from_GstVaapiSurfaceRenderFlags (flags) |
       from_GstVaapiScaleMethod (filter->scale_method);
   pipeline_param->filters = filters;
   pipeline_param->num_filters = num_filters;
 
+  from_GstVideoOrientationMethod (filter->video_direction, &va_mirror,
+      &va_rotation);
+
+#if VA_CHECK_VERSION(1,1,0)
+  pipeline_param->mirror_state = va_mirror;
+  pipeline_param->rotation_state = va_rotation;
+#endif
+
   // Reference frames for advanced deinterlacing
   if (filter->forward_references->len > 0) {
     pipeline_param->forward_references = (VASurfaceID *)
@@ -1470,7 +1808,7 @@ gst_vaapi_filter_process_unlocked (GstVaapiFilter * filter,
   vaapi_unmap_buffer (filter->va_display, pipeline_param_buf_id, NULL);
 
   va_status = vaBeginPicture (filter->va_display, filter->va_context,
-      GST_VAAPI_OBJECT_ID (dst_surface));
+      GST_VAAPI_SURFACE_ID (dst_surface));
   if (!vaapi_check_status (va_status, "vaBeginPicture()"))
     goto error;
 
@@ -1532,9 +1870,11 @@ gst_vaapi_filter_get_formats (GstVaapiFilter * filter)
 {
   g_return_val_if_fail (filter != NULL, NULL);
 
-  if (!ensure_formats (filter))
+  if (!ensure_attributes (filter))
     return NULL;
-  return g_array_ref (filter->formats);
+  if (filter->attribs->formats)
+    return g_array_ref (filter->attribs->formats);
+  return NULL;
 }
 
 /**
@@ -1560,7 +1900,7 @@ gst_vaapi_filter_set_format (GstVaapiFilter * filter, GstVideoFormat format)
 {
   g_return_val_if_fail (filter != NULL, FALSE);
 
-  if (!ensure_formats (filter))
+  if (!ensure_attributes (filter))
     return FALSE;
 
   if (!is_special_format (format) && !find_format (filter, format))
@@ -1570,6 +1910,61 @@ gst_vaapi_filter_set_format (GstVaapiFilter * filter, GstVideoFormat format)
   return TRUE;
 }
 
+/**
+ * gst_vaapi_filter_append_caps:
+ * @filter: a #GstVaapiFilter
+ * @structure: a #GstStructure from #GstCaps
+ *
+ * Extracts the config's surface attributes, from @filter's context,
+ * and transforms it into a caps formats and appended them into
+ * @structure.
+ *
+ * Returns: %TRUE if the capabilities could be extracted and appended
+ * into @structure; otherwise %FALSE
+ **/
+gboolean
+gst_vaapi_filter_append_caps (GstVaapiFilter * filter, GstStructure * structure)
+{
+  GstVaapiConfigSurfaceAttributes *attribs;
+
+  g_return_val_if_fail (filter != NULL, FALSE);
+  g_return_val_if_fail (structure != NULL, FALSE);
+
+  if (!ensure_attributes (filter))
+    return FALSE;
+
+  attribs = filter->attribs;
+
+  if (attribs->min_width >= attribs->max_width ||
+      attribs->min_height >= attribs->max_height)
+    return FALSE;
+
+  gst_structure_set (structure, "width", GST_TYPE_INT_RANGE, attribs->min_width,
+      attribs->max_width, "height", GST_TYPE_INT_RANGE, attribs->min_height,
+      attribs->max_height, NULL);
+
+  return TRUE;
+
+}
+
+/**
+ * gst_vaapi_filter_get_memory_types:
+ * @filter: a #GstVaapiFilter
+ *
+ * Gets the surface's memory types available in @filter's context.
+ *
+ * Returns: surface's memory types available in @filter context.
+ **/
+guint
+gst_vaapi_filter_get_memory_types (GstVaapiFilter * filter)
+{
+  g_return_val_if_fail (filter != NULL, FALSE);
+
+  if (!ensure_attributes (filter))
+    return 0;
+  return filter->attribs->mem_types;
+}
+
 /**
  * gst_vaapi_filter_set_cropping_rectangle:
  * @filter: a #GstVaapiFilter
@@ -1819,6 +2214,7 @@ gst_vaapi_filter_set_scaling (GstVaapiFilter * filter,
   return TRUE;
 }
 
+#ifndef GST_REMOVE_DEPRECATED
 /**
  * gst_vaapi_filter_set_skintone:
  * @filter: a #GstVaapiFilter
@@ -1837,81 +2233,311 @@ gst_vaapi_filter_set_skintone (GstVaapiFilter * filter, gboolean enhance)
   return op_set_skintone (filter,
       find_operation (filter, GST_VAAPI_FILTER_OP_SKINTONE), enhance);
 }
+#endif
 
-static inline gfloat
-op_get_float_default_value (GstVaapiFilter * filter,
-    GstVaapiFilterOpData * op_data)
+/**
+ * gst_vaapi_filter_set_skintone_level:
+ * @filter: a #GstVaapiFilter
+ * @value: the value if enable the skin tone enhancement algorithm
+ *
+ * Applies the skin tone enhancement algorithm with specifled value.
+ *
+ * Return value: %TRUE if the operation is supported, %FALSE
+ * otherwise.
+  **/
+gboolean
+gst_vaapi_filter_set_skintone_level (GstVaapiFilter * filter, guint value)
 {
-  GParamSpecFloat *const pspec = G_PARAM_SPEC_FLOAT (op_data->pspec);
-  return pspec->default_value;
+  g_return_val_if_fail (filter != NULL, FALSE);
+
+  return op_set_skintone_level (filter,
+      find_operation (filter, GST_VAAPI_FILTER_OP_SKINTONE_LEVEL), value);
+}
+
+/**
+ * gst_vaapi_filter_set_video_direction:
+ * @filter: a #GstVaapiFilter
+ * @method: the video direction (see #GstVideoOrientationMethod)
+ *
+ * Applies mirror/rotation to the video processing pipeline.
+ *
+ * Return value: %TRUE if the operation is supported, %FALSE otherwise.
+ */
+gboolean
+gst_vaapi_filter_set_video_direction (GstVaapiFilter * filter,
+    GstVideoOrientationMethod method)
+{
+  g_return_val_if_fail (filter != NULL, FALSE);
+
+#if VA_CHECK_VERSION(1,1,0)
+  {
+    guint32 va_mirror = VA_MIRROR_NONE;
+    guint32 va_rotation = VA_ROTATION_NONE;
+
+    from_GstVideoOrientationMethod (method, &va_mirror, &va_rotation);
+
+    if (va_mirror != VA_MIRROR_NONE && !(filter->mirror_flags & va_mirror))
+      return FALSE;
+
+    if (va_rotation != VA_ROTATION_NONE
+        && !(filter->rotation_flags & (1 << va_rotation)))
+      return FALSE;
+  }
+#else
+  return FALSE;
+#endif
+
+  filter->video_direction = method;
+  return TRUE;
+}
+
+/**
+ * gst_vaapi_filter_get_video_direction:
+ * @filter: a #GstVaapiFilter
+ *
+ * Return value: the currently applied video direction (see #GstVideoOrientationMethod)
+ */
+GstVideoOrientationMethod
+gst_vaapi_filter_get_video_direction (GstVaapiFilter * filter)
+{
+  g_return_val_if_fail (filter != NULL, GST_VIDEO_ORIENTATION_IDENTITY);
+  return filter->video_direction;
 }
 
 gfloat
 gst_vaapi_filter_get_denoising_level_default (GstVaapiFilter * filter)
 {
-  g_return_val_if_fail (filter != NULL, FALSE);
-
-  return op_get_float_default_value (filter,
-      find_operation (filter, GST_VAAPI_FILTER_OP_DENOISE));
+  OP_RET_DEFAULT_VALUE (float, filter, GST_VAAPI_FILTER_OP_DENOISE);
 }
 
 gfloat
 gst_vaapi_filter_get_sharpening_level_default (GstVaapiFilter * filter)
 {
-  g_return_val_if_fail (filter != NULL, FALSE);
-
-  return op_get_float_default_value (filter,
-      find_operation (filter, GST_VAAPI_FILTER_OP_SHARPEN));
+  OP_RET_DEFAULT_VALUE (float, filter, GST_VAAPI_FILTER_OP_SHARPEN);
 }
 
 gfloat
 gst_vaapi_filter_get_hue_default (GstVaapiFilter * filter)
 {
-  g_return_val_if_fail (filter != NULL, FALSE);
-
-  return op_get_float_default_value (filter,
-      find_operation (filter, GST_VAAPI_FILTER_OP_HUE));
+  OP_RET_DEFAULT_VALUE (float, filter, GST_VAAPI_FILTER_OP_HUE);
 }
 
 gfloat
 gst_vaapi_filter_get_saturation_default (GstVaapiFilter * filter)
 {
-  g_return_val_if_fail (filter != NULL, FALSE);
-
-  return op_get_float_default_value (filter,
-      find_operation (filter, GST_VAAPI_FILTER_OP_SATURATION));
+  OP_RET_DEFAULT_VALUE (float, filter, GST_VAAPI_FILTER_OP_SATURATION);
 }
 
 gfloat
 gst_vaapi_filter_get_brightness_default (GstVaapiFilter * filter)
 {
-  g_return_val_if_fail (filter != NULL, FALSE);
-
-  return op_get_float_default_value (filter,
-      find_operation (filter, GST_VAAPI_FILTER_OP_BRIGHTNESS));
+  OP_RET_DEFAULT_VALUE (float, filter, GST_VAAPI_FILTER_OP_BRIGHTNESS);
 }
 
 gfloat
 gst_vaapi_filter_get_contrast_default (GstVaapiFilter * filter)
 {
-  g_return_val_if_fail (filter != NULL, FALSE);
-
-  return op_get_float_default_value (filter,
-      find_operation (filter, GST_VAAPI_FILTER_OP_CONTRAST));
+  OP_RET_DEFAULT_VALUE (float, filter, GST_VAAPI_FILTER_OP_CONTRAST);
 }
 
 GstVaapiScaleMethod
 gst_vaapi_filter_get_scaling_default (GstVaapiFilter * filter)
 {
-  g_return_val_if_fail (filter != NULL, FALSE);
-
-  return DEFAULT_SCALING;
+  OP_RET_DEFAULT_VALUE (enum, filter, GST_VAAPI_FILTER_OP_SCALING);
 }
 
+#ifndef GST_REMOVE_DEPRECATED
 gboolean
 gst_vaapi_filter_get_skintone_default (GstVaapiFilter * filter)
 {
+  OP_RET_DEFAULT_VALUE (boolean, filter, GST_VAAPI_FILTER_OP_SKINTONE);
+}
+#endif
+
+guint
+gst_vaapi_filter_get_skintone_level_default (GstVaapiFilter * filter)
+{
+  OP_RET_DEFAULT_VALUE (uint, filter, GST_VAAPI_FILTER_OP_SKINTONE_LEVEL);
+}
+
+GstVideoOrientationMethod
+gst_vaapi_filter_get_video_direction_default (GstVaapiFilter * filter)
+{
+  OP_RET_DEFAULT_VALUE (enum, filter, GST_VAAPI_FILTER_OP_VIDEO_DIRECTION);
+}
+
+static gboolean
+gst_vaapi_filter_set_colorimetry_unlocked (GstVaapiFilter * filter,
+    GstVideoColorimetry * input, GstVideoColorimetry * output)
+{
+  gchar *in_color, *out_color;
+
+  if (input)
+    filter->input_colorimetry = *input;
+  else
+    gst_video_colorimetry_from_string (&filter->input_colorimetry, NULL);
+
+  if (output)
+    filter->output_colorimetry = *output;
+  else
+    gst_video_colorimetry_from_string (&filter->output_colorimetry, NULL);
+
+  in_color = gst_video_colorimetry_to_string (&filter->input_colorimetry);
+  GST_DEBUG_OBJECT (filter, " input colorimetry '%s'", in_color);
+
+  out_color = gst_video_colorimetry_to_string (&filter->output_colorimetry);
+  GST_DEBUG_OBJECT (filter, "output colorimetry '%s'", out_color);
+
+  if (!gst_vaapi_display_has_driver_quirks (filter->display,
+          GST_VAAPI_DRIVER_QUIRK_NO_CHECK_VPP_COLOR_STD)) {
+    VAProcPipelineCaps pipeline_caps = { 0, };
+    VAProcColorStandardType type;
+    guint32 i;
+
+    VAStatus va_status = vaQueryVideoProcPipelineCaps (filter->va_display,
+        filter->va_context, NULL, 0, &pipeline_caps);
+
+    if (!vaapi_check_status (va_status, "vaQueryVideoProcPipelineCaps()"))
+      return FALSE;
+
+    type = from_GstVideoColorimetry (&filter->input_colorimetry);
+    for (i = 0; i < pipeline_caps.num_input_color_standards; i++)
+      if (type == pipeline_caps.input_color_standards[i])
+        break;
+    if ((i == pipeline_caps.num_input_color_standards)
+        && (type != VAProcColorStandardNone))
+      GST_WARNING_OBJECT (filter,
+          "driver does not support '%s' input colorimetry."
+          " vpp may fail or produce unexpected results.", in_color);
+
+    type = from_GstVideoColorimetry (&filter->output_colorimetry);
+    for (i = 0; i < pipeline_caps.num_output_color_standards; i++)
+      if (type == pipeline_caps.output_color_standards[i])
+        break;
+    if ((i == pipeline_caps.num_output_color_standards)
+        && (type != VAProcColorStandardNone))
+      GST_WARNING_OBJECT (filter,
+          "driver does not support '%s' output colorimetry."
+          " vpp may fail or produce unexpected results.", out_color);
+  } else {
+    GST_WARNING_OBJECT (filter,
+        "driver does not report the supported input/output colorimetry."
+        " vpp may fail or produce unexpected results.");
+  }
+
+  g_free (in_color);
+  g_free (out_color);
+
+  return TRUE;
+}
+
+gboolean
+gst_vaapi_filter_set_colorimetry (GstVaapiFilter * filter,
+    GstVideoColorimetry * input, GstVideoColorimetry * output)
+{
+  gboolean result;
+
   g_return_val_if_fail (filter != NULL, FALSE);
 
+  GST_VAAPI_DISPLAY_LOCK (filter->display);
+  result = gst_vaapi_filter_set_colorimetry_unlocked (filter, input, output);
+  GST_VAAPI_DISPLAY_UNLOCK (filter->display);
+
+  return result;
+}
+
+/**
+ * gst_vaapi_filter_set_hdr_tone_map:
+ * @filter: a #GstVaapiFilter
+ * @value: %TRUE to enable hdr tone map algorithm
+ *
+ * Applies HDR tone mapping algorithm.
+ *
+ * Return value: %TRUE if the operation is supported, %FALSE otherwise.
+ */
+gboolean
+gst_vaapi_filter_set_hdr_tone_map (GstVaapiFilter * filter, gboolean value)
+{
+  g_return_val_if_fail (filter != NULL, FALSE);
+
+  return op_set_hdr_tone_map (filter,
+      find_operation (filter, GST_VAAPI_FILTER_OP_HDR_TONE_MAP), value);
+}
+
+static gboolean
+gst_vaapi_filter_set_hdr_tone_map_meta_unlocked (GstVaapiFilter * filter,
+    GstVideoMasteringDisplayInfo * minfo, GstVideoContentLightLevel * linfo)
+{
+#if VA_CHECK_VERSION(1,4,0)
+  GstVaapiFilterOpData *op_data;
+  VAProcFilterParameterBufferHDRToneMapping *buf;
+  VAHdrMetaDataHDR10 *meta = &filter->hdr_meta;
+
+  op_data = find_operation (filter, GST_VAAPI_FILTER_OP_HDR_TONE_MAP);
+
+  if (!op_data)
+    return FALSE;
+
+  meta->display_primaries_x[0] = minfo->display_primaries[1].x;
+  meta->display_primaries_x[1] = minfo->display_primaries[2].x;
+  meta->display_primaries_x[2] = minfo->display_primaries[0].x;
+
+  meta->display_primaries_y[0] = minfo->display_primaries[1].y;
+  meta->display_primaries_y[1] = minfo->display_primaries[2].y;
+  meta->display_primaries_y[2] = minfo->display_primaries[0].y;
+
+  meta->white_point_x = minfo->white_point.x;
+  meta->white_point_y = minfo->white_point.y;
+
+  meta->max_display_mastering_luminance =
+      minfo->max_display_mastering_luminance;
+  meta->min_display_mastering_luminance =
+      minfo->min_display_mastering_luminance;
+
+  meta->max_content_light_level = linfo->max_content_light_level;
+  meta->max_pic_average_light_level = linfo->max_frame_average_light_level;
+
+  buf = vaapi_map_buffer (filter->va_display, op_data->va_buffer);
+  if (!buf)
+    return FALSE;
+
+  buf->type = op_data->va_type;
+  buf->data.metadata_type = op_data->va_subtype;
+  buf->data.metadata = meta;
+  buf->data.metadata_size = sizeof (meta);
+
+  vaapi_unmap_buffer (filter->va_display, op_data->va_buffer, NULL);
+
+  return TRUE;
+#else
   return FALSE;
+#endif
+}
+
+/**
+ * gst_vaapi_filter_set_hdr_tone_map_meta:
+ * @filter: a #GstVaapiFilter
+ * @minfo: a #GstVideoMasteringDisplayInfo
+ * @linfo: a #GstVideoContentLightLevel
+ *
+ * Sets the input HDR meta data used for tone mapping.
+ *
+ * Return value: %TRUE if the operation is supported, %FALSE otherwise.
+ */
+gboolean
+gst_vaapi_filter_set_hdr_tone_map_meta (GstVaapiFilter * filter,
+    GstVideoMasteringDisplayInfo * minfo, GstVideoContentLightLevel * linfo)
+{
+  gboolean status = FALSE;
+
+  g_return_val_if_fail (filter != NULL, FALSE);
+  g_return_val_if_fail (minfo != NULL, FALSE);
+  g_return_val_if_fail (linfo != NULL, FALSE);
+
+  GST_VAAPI_DISPLAY_LOCK (filter->display);
+  status =
+      gst_vaapi_filter_set_hdr_tone_map_meta_unlocked (filter, minfo, linfo);
+  GST_VAAPI_DISPLAY_UNLOCK (filter->display);
+
+  return status;
 }
