@@ -25,6 +25,7 @@
 #ifndef GST_VAAPI_PROFILE_H
 #define GST_VAAPI_PROFILE_H
 
+#include <va/va.h>
 #include <gst/gstvalue.h>
 
 G_BEGIN_DECLS
@@ -180,6 +181,15 @@ typedef enum {
     GST_VAAPI_PROFILE_H265_MAIN_422_10        = GST_VAAPI_MAKE_PROFILE(H265,4),
     GST_VAAPI_PROFILE_H265_MAIN_444           = GST_VAAPI_MAKE_PROFILE(H265,5),
     GST_VAAPI_PROFILE_H265_MAIN_444_10        = GST_VAAPI_MAKE_PROFILE(H265,6),
+    GST_VAAPI_PROFILE_H265_SCREEN_EXTENDED_MAIN =
+                                                GST_VAAPI_MAKE_PROFILE(H265,7),
+    GST_VAAPI_PROFILE_H265_SCREEN_EXTENDED_MAIN_10 =
+                                                GST_VAAPI_MAKE_PROFILE(H265,8),
+    GST_VAAPI_PROFILE_H265_SCREEN_EXTENDED_MAIN_444 =
+                                                GST_VAAPI_MAKE_PROFILE(H265,9),
+    GST_VAAPI_PROFILE_H265_SCREEN_EXTENDED_MAIN_444_10 =
+                                                GST_VAAPI_MAKE_PROFILE(H265,10),
+    GST_VAAPI_PROFILE_H265_MAIN12             = GST_VAAPI_MAKE_PROFILE(H265,11),
     GST_VAAPI_PROFILE_VP9_0                   = GST_VAAPI_MAKE_PROFILE(VP9,1),
     GST_VAAPI_PROFILE_VP9_1                   = GST_VAAPI_MAKE_PROFILE(VP9,2),
     GST_VAAPI_PROFILE_VP9_2                   = GST_VAAPI_MAKE_PROFILE(VP9,3),
@@ -188,6 +198,7 @@ typedef enum {
 
 /**
  * GstVaapiEntrypoint:
+ * @GST_VAAPI_ENTRYPOINT_INVALID: Invalid entrypoint
  * @GST_VAAPI_ENTRYPOINT_VLD: Variable Length Decoding
  * @GST_VAAPI_ENTRYPOINT_IDCT: Inverse Decrete Cosine Transform
  * @GST_VAAPI_ENTRYPOINT_MOCO: Motion Compensation
@@ -195,18 +206,17 @@ typedef enum {
  * @GST_VAAPI_ENTRYPOINT_PICTURE_ENCODE: Encode Picture
  * @GST_VAAPI_ENTRYPOINT_SLICE_ENCODE_LP: Encode Slice low power/
  *                                        high performace varient
- * @GST_VAAPI_ENTRYPOINT_SLICE_ENCODE_FEI: FEI Encode
  *
  * The set of all entrypoints for #GstVaapiEntrypoint
  */
 typedef enum {
-    GST_VAAPI_ENTRYPOINT_VLD = 1,
+    GST_VAAPI_ENTRYPOINT_INVALID,
+    GST_VAAPI_ENTRYPOINT_VLD,
     GST_VAAPI_ENTRYPOINT_IDCT,
     GST_VAAPI_ENTRYPOINT_MOCO,
     GST_VAAPI_ENTRYPOINT_SLICE_ENCODE,
     GST_VAAPI_ENTRYPOINT_PICTURE_ENCODE,
     GST_VAAPI_ENTRYPOINT_SLICE_ENCODE_LP,
-    GST_VAAPI_ENTRYPOINT_SLICE_ENCODE_FEI
 } GstVaapiEntrypoint;
 
 const gchar *
@@ -218,9 +228,15 @@ gst_vaapi_profile(VAProfile profile);
 GstVaapiProfile
 gst_vaapi_profile_from_caps(const GstCaps *caps);
 
+GstVaapiCodec
+gst_vaapi_get_codec_from_caps (const GstCaps *caps);
+
 const gchar *
 gst_vaapi_profile_get_name(GstVaapiProfile profile);
 
+const gchar *
+gst_vaapi_profile_get_va_name(GstVaapiProfile profile);
+
 const gchar *
 gst_vaapi_profile_get_media_type_name(GstVaapiProfile profile);
 
