@@ -40,6 +40,10 @@ static const struct map gst_vaapi_h265_profile_map[] = {
   { GST_VAAPI_PROFILE_H265_MAIN,                 "main"                 },
   { GST_VAAPI_PROFILE_H265_MAIN10,               "main-10"              },
   { GST_VAAPI_PROFILE_H265_MAIN_STILL_PICTURE,   "main-still-picture"   },
+  { GST_VAAPI_PROFILE_H265_MAIN_444,             "main-444"             },
+  { GST_VAAPI_PROFILE_H265_MAIN_444_10,          "main-444-10"          },
+  { GST_VAAPI_PROFILE_H265_MAIN_422_10,          "main-422-10"          },
+  { GST_VAAPI_PROFILE_H265_MAIN12,               "main-12"              },
   { 0, NULL }
 /* *INDENT-ON* */
 };
@@ -137,59 +141,64 @@ gst_vaapi_utils_h265_get_profile_score (GstVaapiProfile profile)
 GstVaapiProfile
 gst_vaapi_utils_h265_get_profile (GstH265SPS * sps)
 {
-  GstVaapiProfile profile;
+  GstVaapiProfile vaapi_profile;
+  GstH265Profile profile;
 
-  switch (sps->profile_tier_level.profile_idc) {
-    case GST_H265_PROFILE_IDC_MAIN:
-      profile = GST_VAAPI_PROFILE_H265_MAIN;
+  g_return_val_if_fail (sps != NULL, GST_VAAPI_PROFILE_UNKNOWN);
+
+  profile = gst_h265_profile_tier_level_get_profile (&sps->profile_tier_level);
+  switch (profile) {
+    case GST_H265_PROFILE_MAIN:
+      /* Main Intra, recognize it as MAIN */
+    case GST_H265_PROFILE_MAIN_INTRA:
+      vaapi_profile = GST_VAAPI_PROFILE_H265_MAIN;
       break;
-    case GST_H265_PROFILE_IDC_MAIN_10:
-      profile = GST_VAAPI_PROFILE_H265_MAIN10;
+    case GST_H265_PROFILE_MAIN_10:
+      /* Main 10 Intra, recognize it as MAIN10 */
+    case GST_H265_PROFILE_MAIN_10_INTRA:
+      vaapi_profile = GST_VAAPI_PROFILE_H265_MAIN10;
       break;
-    case GST_H265_PROFILE_IDC_MAIN_STILL_PICTURE:
-      profile = GST_VAAPI_PROFILE_H265_MAIN_STILL_PICTURE;
+    case GST_H265_PROFILE_MAIN_12:
+      /* Main 12 Intra, recognize it as MAIN_12 */
+    case GST_H265_PROFILE_MAIN_12_INTRA:
+      vaapi_profile = GST_VAAPI_PROFILE_H265_MAIN12;
+      break;
+    case GST_H265_PROFILE_MAIN_STILL_PICTURE:
+      vaapi_profile = GST_VAAPI_PROFILE_H265_MAIN_STILL_PICTURE;
+      break;
+    case GST_H265_PROFILE_MAIN_422_10:
+      /* Main 422_10 Intra, recognize it as MAIN_422_10 */
+    case GST_H265_PROFILE_MAIN_422_10_INTRA:
+      vaapi_profile = GST_VAAPI_PROFILE_H265_MAIN_422_10;
+      break;
+    case GST_H265_PROFILE_MAIN_444:
+      /* Main 444 Intra, recognize it as MAIN_444 */
+    case GST_H265_PROFILE_MAIN_444_INTRA:
+      vaapi_profile = GST_VAAPI_PROFILE_H265_MAIN_444;
+      break;
+    case GST_H265_PROFILE_MAIN_444_10:
+      /* Main 444_10 Intra, recognize it as MAIN_444_10 */
+    case GST_H265_PROFILE_MAIN_444_10_INTRA:
+      vaapi_profile = GST_VAAPI_PROFILE_H265_MAIN_444_10;
+      break;
+    case GST_H265_PROFILE_SCREEN_EXTENDED_MAIN:
+      vaapi_profile = GST_VAAPI_PROFILE_H265_SCREEN_EXTENDED_MAIN;
+      break;
+    case GST_H265_PROFILE_SCREEN_EXTENDED_MAIN_10:
+      vaapi_profile = GST_VAAPI_PROFILE_H265_SCREEN_EXTENDED_MAIN_10;
+      break;
+    case GST_H265_PROFILE_SCREEN_EXTENDED_MAIN_444:
+      vaapi_profile = GST_VAAPI_PROFILE_H265_SCREEN_EXTENDED_MAIN_444;
+      break;
+    case GST_H265_PROFILE_SCREEN_EXTENDED_MAIN_444_10:
+      vaapi_profile = GST_VAAPI_PROFILE_H265_SCREEN_EXTENDED_MAIN_444_10;
       break;
-    case GST_H265_PROFILE_IDC_FORMAT_RANGE_EXTENSION:
-      if (sps->profile_tier_level.max_12bit_constraint_flag == 1
-          && sps->profile_tier_level.max_10bit_constraint_flag == 1
-          && sps->profile_tier_level.max_8bit_constraint_flag == 0
-          && sps->profile_tier_level.max_422chroma_constraint_flag == 1
-          && sps->profile_tier_level.max_420chroma_constraint_flag == 0
-          && sps->profile_tier_level.max_monochrome_constraint_flag == 0
-          && sps->profile_tier_level.intra_constraint_flag == 0
-          && sps->profile_tier_level.one_picture_only_constraint_flag == 0
-          && sps->profile_tier_level.lower_bit_rate_constraint_flag == 1) {
-        profile = GST_VAAPI_PROFILE_H265_MAIN_422_10;
-        break;
-      } else if (sps->profile_tier_level.max_12bit_constraint_flag == 1
-          && sps->profile_tier_level.max_10bit_constraint_flag == 1
-          && sps->profile_tier_level.max_8bit_constraint_flag == 1
-          && sps->profile_tier_level.max_422chroma_constraint_flag == 0
-          && sps->profile_tier_level.max_420chroma_constraint_flag == 0
-          && sps->profile_tier_level.max_monochrome_constraint_flag == 0
-          && sps->profile_tier_level.intra_constraint_flag == 0
-          && sps->profile_tier_level.one_picture_only_constraint_flag == 0
-          && sps->profile_tier_level.lower_bit_rate_constraint_flag == 1) {
-        profile = GST_VAAPI_PROFILE_H265_MAIN_444;
-        break;
-      } else if (sps->profile_tier_level.max_12bit_constraint_flag == 1
-          && sps->profile_tier_level.max_10bit_constraint_flag == 1
-          && sps->profile_tier_level.max_8bit_constraint_flag == 0
-          && sps->profile_tier_level.max_422chroma_constraint_flag == 0
-          && sps->profile_tier_level.max_420chroma_constraint_flag == 0
-          && sps->profile_tier_level.max_monochrome_constraint_flag == 0
-          && sps->profile_tier_level.intra_constraint_flag == 0
-          && sps->profile_tier_level.one_picture_only_constraint_flag == 0
-          && sps->profile_tier_level.lower_bit_rate_constraint_flag == 1) {
-        profile = GST_VAAPI_PROFILE_H265_MAIN_444_10;
-        break;
-      }
     default:
       GST_DEBUG ("unsupported profile_idc value");
-      profile = GST_VAAPI_PROFILE_UNKNOWN;
+      vaapi_profile = GST_VAAPI_PROFILE_UNKNOWN;
       break;
   }
-  return profile;
+  return vaapi_profile;
 }
 
 /** Returns H.265 profile_idc value from GstVaapiProfile */
@@ -200,19 +209,22 @@ gst_vaapi_utils_h265_get_profile_idc (GstVaapiProfile profile)
 
   switch (profile) {
     case GST_VAAPI_PROFILE_H265_MAIN:
-      profile_idc = GST_H265_PROFILE_MAIN;
+      profile_idc = GST_H265_PROFILE_IDC_MAIN;
       break;
     case GST_VAAPI_PROFILE_H265_MAIN10:
-      profile_idc = GST_H265_PROFILE_MAIN_10;
+      profile_idc = GST_H265_PROFILE_IDC_MAIN_10;
       break;
     case GST_VAAPI_PROFILE_H265_MAIN_STILL_PICTURE:
-      profile_idc = GST_H265_PROFILE_MAIN_STILL_PICTURE;
+      profile_idc = GST_H265_PROFILE_IDC_MAIN_STILL_PICTURE;
       break;
     case GST_VAAPI_PROFILE_H265_MAIN_422_10:
-      profile_idc = GST_H265_PROFILE_MAIN_422_10;
-      break;
+      /* Fall through */
     case GST_VAAPI_PROFILE_H265_MAIN_444:
-      profile_idc = GST_H265_PROFILE_MAIN_444;
+      /* Fall through */
+    case GST_VAAPI_PROFILE_H265_MAIN_444_10:
+      /* Fall through */
+    case GST_VAAPI_PROFILE_H265_MAIN12:
+      profile_idc = GST_H265_PROFILE_IDC_FORMAT_RANGE_EXTENSION;
       break;
     default:
       GST_DEBUG ("unsupported GstVaapiProfile value");
@@ -327,37 +339,50 @@ gst_vaapi_utils_h265_get_level_limits_table (guint * out_length_ptr)
 /** Returns GstVaapiChromaType from H.265 chroma_format_idc value */
 GstVaapiChromaType
 gst_vaapi_utils_h265_get_chroma_type (guint chroma_format_idc,
-    guint luma_bit_depth)
+    guint luma_bit_depth, guint chroma_bit_depth)
 {
   GstVaapiChromaType chroma_type = (GstVaapiChromaType) 0;
+  guint depth = 0;
+
+  if (luma_bit_depth < 8 || chroma_bit_depth < 8 ||
+      luma_bit_depth > 16 || chroma_bit_depth > 16) {
+    GST_WARNING ("invalid luma_bit_depth or chroma_bit_depth value");
+    return chroma_type;
+  }
+
+  depth = MAX (luma_bit_depth, chroma_bit_depth);
 
   switch (chroma_format_idc) {
     case 0:
       chroma_type = GST_VAAPI_CHROMA_TYPE_YUV400;
       break;
     case 1:
-      if (luma_bit_depth == 8)
+      if (depth == 8)
         chroma_type = GST_VAAPI_CHROMA_TYPE_YUV420;
-      else if (luma_bit_depth > 8)
+      else if (depth > 8 && depth <= 10)
         chroma_type = GST_VAAPI_CHROMA_TYPE_YUV420_10BPP;
+      else if (depth > 10 && depth <= 12)
+        chroma_type = GST_VAAPI_CHROMA_TYPE_YUV420_12BPP;
       break;
     case 2:
-      if (luma_bit_depth == 8)
+      if (depth == 8)
         chroma_type = GST_VAAPI_CHROMA_TYPE_YUV422;
-      else if (luma_bit_depth > 8)
+      else if (depth > 8 && depth <= 10)
         chroma_type = GST_VAAPI_CHROMA_TYPE_YUV422_10BPP;
       break;
     case 3:
-      if (luma_bit_depth == 8)
+      if (depth == 8)
         chroma_type = GST_VAAPI_CHROMA_TYPE_YUV444;
-      else if (luma_bit_depth > 8)
+      else if (depth > 8 && depth <= 10)
         chroma_type = GST_VAAPI_CHROMA_TYPE_YUV444_10BPP;
       break;
     default:
-      GST_DEBUG ("unsupported chroma_format_idc value");
-      chroma_type = (GstVaapiChromaType) 0;
       break;
   }
+
+  if (chroma_type == (GstVaapiChromaType) 0)
+    GST_DEBUG ("unsupported chroma_format_idc value");
+
   return chroma_type;
 }
 
@@ -373,12 +398,15 @@ gst_vaapi_utils_h265_get_chroma_format_idc (GstVaapiChromaType chroma_type)
       break;
     case GST_VAAPI_CHROMA_TYPE_YUV420:
     case GST_VAAPI_CHROMA_TYPE_YUV420_10BPP:
+    case GST_VAAPI_CHROMA_TYPE_YUV420_12BPP:
       chroma_format_idc = 1;
       break;
     case GST_VAAPI_CHROMA_TYPE_YUV422:
+    case GST_VAAPI_CHROMA_TYPE_YUV422_10BPP:
       chroma_format_idc = 2;
       break;
     case GST_VAAPI_CHROMA_TYPE_YUV444:
+    case GST_VAAPI_CHROMA_TYPE_YUV444_10BPP:
       chroma_format_idc = 3;
       break;
     default:
