@@ -28,7 +28,7 @@
 
 #include "sysdeps.h"
 #include "gstvaapicompat.h"
-#include "gstvaapiobject_priv.h"
+#include "gstvaapisurface_priv.h"
 #include "gstvaapiwindow_wayland.h"
 #include "gstvaapiwindow_priv.h"
 #include "gstvaapidisplay_wayland.h"
@@ -37,7 +37,10 @@
 #include "gstvaapifilter.h"
 #include "gstvaapisurfacepool.h"
 
+#include <unistd.h>
+
 GST_DEBUG_CATEGORY_EXTERN (gst_debug_vaapi_window);
+GST_DEBUG_CATEGORY_EXTERN (gst_debug_vaapi);
 #define GST_CAT_DEFAULT gst_debug_vaapi_window
 
 #define GST_VAAPI_WINDOW_WAYLAND_CAST(obj) \
@@ -58,6 +61,7 @@ struct _FrameState
   GstVaapiWindow *window;
   GstVaapiSurface *surface;
   GstVaapiVideoPool *surface_pool;
+  struct wl_buffer *buffer;
   struct wl_callback *callback;
   gboolean done;
 };
@@ -79,31 +83,15 @@ frame_state_new (GstVaapiWindow * window)
   return frame;
 }
 
-static void
-frame_state_free (FrameState * frame)
-{
-  if (!frame)
-    return;
-
-  if (frame->surface) {
-    if (frame->surface_pool)
-      gst_vaapi_video_pool_put_object (frame->surface_pool, frame->surface);
-    frame->surface = NULL;
-  }
-  gst_vaapi_video_pool_replace (&frame->surface_pool, NULL);
-
-  g_clear_pointer (&frame->callback, wl_callback_destroy);
-  g_slice_free (FrameState, frame);
-}
-
 struct _GstVaapiWindowWaylandPrivate
 {
   struct xdg_surface *xdg_surface;
   struct xdg_toplevel *xdg_toplevel;
   struct wl_shell_surface *wl_shell_surface;
   struct wl_surface *surface;
-  struct wl_region *opaque_region;
+  struct wl_subsurface *video_subsurface;
   struct wl_event_queue *event_queue;
+  GList *frames;
   FrameState *last_frame;
   GstPoll *poll;
   GstPollFD pollfd;
@@ -111,7 +99,11 @@ struct _GstVaapiWindowWaylandPrivate
   guint fullscreen_on_show:1;
   guint sync_failed:1;
   volatile guint num_frames_pending;
+  gint configure_pending;
   gboolean need_vpp;
+  gboolean dmabuf_broken;
+  GMutex opaque_mutex;
+  gint opaque_width, opaque_height;
 };
 
 /**
@@ -128,7 +120,7 @@ struct _GstVaapiWindowWayland
 /**
  * GstVaapiWindowWaylandClass:
  *
- * An Wayland #Window wrapper class.
+ * An Wayland Window wrapper class.
  */
 struct _GstVaapiWindowWaylandClass
 {
@@ -148,6 +140,29 @@ enum
 
 static guint signals[N_SIGNALS];
 
+static void
+frame_state_free (FrameState * frame)
+{
+  GstVaapiWindowWaylandPrivate *priv;
+
+  if (!frame)
+    return;
+
+  priv = GST_VAAPI_WINDOW_WAYLAND_GET_PRIVATE (frame->window);
+  priv->frames = g_list_remove (priv->frames, frame);
+
+  if (frame->surface) {
+    if (frame->surface_pool)
+      gst_vaapi_video_pool_put_object (frame->surface_pool, frame->surface);
+    frame->surface = NULL;
+  }
+  gst_vaapi_video_pool_replace (&frame->surface_pool, NULL);
+
+  g_clear_pointer (&frame->callback, wl_callback_destroy);
+  wl_buffer_destroy (frame->buffer);
+  g_slice_free (FrameState, frame);
+}
+
 static void
 handle_xdg_toplevel_configure (void *data, struct xdg_toplevel *xdg_toplevel,
     int32_t width, int32_t height, struct wl_array *states)
@@ -184,6 +199,8 @@ static const struct xdg_toplevel_listener xdg_toplevel_listener = {
   handle_xdg_toplevel_close,
 };
 
+static gboolean gst_vaapi_window_wayland_sync (GstVaapiWindow * window);
+
 static gboolean
 gst_vaapi_window_wayland_show (GstVaapiWindow * window)
 {
@@ -200,6 +217,8 @@ gst_vaapi_window_wayland_show (GstVaapiWindow * window)
     return TRUE;
   }
 
+  g_atomic_int_set (&priv->configure_pending, 1);
+  g_atomic_int_inc (&priv->num_frames_pending);
   /* Create a toplevel window out of it */
   priv->xdg_toplevel = xdg_surface_get_toplevel (priv->xdg_surface);
   g_return_val_if_fail (priv->xdg_toplevel, FALSE);
@@ -213,7 +232,7 @@ gst_vaapi_window_wayland_show (GstVaapiWindow * window)
   /* Commit the xdg_surface state as top-level window */
   wl_surface_commit (priv->surface);
 
-  return TRUE;
+  return gst_vaapi_window_wayland_sync (window);
 }
 
 static gboolean
@@ -322,7 +341,13 @@ static void
 handle_xdg_surface_configure (void *data, struct xdg_surface *xdg_surface,
     uint32_t serial)
 {
+  GstVaapiWindow *window = GST_VAAPI_WINDOW (data);
+  GstVaapiWindowWaylandPrivate *priv =
+      GST_VAAPI_WINDOW_WAYLAND_GET_PRIVATE (window);
+
   xdg_surface_ack_configure (xdg_surface, serial);
+  if (g_atomic_int_compare_and_exchange (&priv->configure_pending, 1, 0))
+    g_atomic_int_dec_and_test (&priv->num_frames_pending);
 }
 
 static const struct xdg_surface_listener xdg_surface_listener = {
@@ -336,6 +361,9 @@ gst_vaapi_window_wayland_set_fullscreen (GstVaapiWindow * window,
   GstVaapiWindowWaylandPrivate *const priv =
       GST_VAAPI_WINDOW_WAYLAND_GET_PRIVATE (window);
 
+  if (window->use_foreign_window)
+    return TRUE;
+
   if (!priv->is_shown) {
     priv->fullscreen_on_show = fullscreen;
     return TRUE;
@@ -389,8 +417,34 @@ gst_vaapi_window_wayland_create (GstVaapiWindow * window,
     return FALSE;
   wl_proxy_set_queue ((struct wl_proxy *) priv->surface, priv->event_queue);
 
-  /* Prefer XDG-shell over deprecated wl_shell (if available) */
-  if (priv_display->xdg_wm_base) {
+  if (window->use_foreign_window) {
+    struct wl_surface *wl_surface;
+
+    if (priv_display->subcompositor) {
+      if (GST_VAAPI_SURFACE_ID (window) == VA_INVALID_ID) {
+        GST_ERROR ("Invalid window");
+        return FALSE;
+      }
+
+      wl_surface = (struct wl_surface *) GST_VAAPI_WINDOW_ID (window);
+      GST_VAAPI_WINDOW_LOCK_DISPLAY (window);
+      priv->video_subsurface =
+          wl_subcompositor_get_subsurface (priv_display->subcompositor,
+          priv->surface, wl_surface);
+      GST_VAAPI_WINDOW_UNLOCK_DISPLAY (window);
+      if (!priv->video_subsurface)
+        return FALSE;
+
+      wl_proxy_set_queue ((struct wl_proxy *) priv->video_subsurface,
+          priv->event_queue);
+
+      wl_subsurface_set_desync (priv->video_subsurface);
+    } else {
+      GST_ERROR ("Wayland server does not support subsurfaces");
+      window->use_foreign_window = FALSE;
+    }
+    /* Prefer XDG-shell over deprecated wl_shell (if available) */
+  } else if (priv_display->xdg_wm_base) {
     /* Create the XDG surface. We make the toplevel on VaapiWindow::show() */
     GST_VAAPI_WINDOW_LOCK_DISPLAY (window);
     priv->xdg_surface = xdg_wm_base_get_xdg_surface (priv_display->xdg_wm_base,
@@ -420,6 +474,8 @@ gst_vaapi_window_wayland_create (GstVaapiWindow * window,
   priv->poll = gst_poll_new (TRUE);
   gst_poll_fd_init (&priv->pollfd);
 
+  g_mutex_init (&priv->opaque_mutex);
+
   if (priv->fullscreen_on_show)
     gst_vaapi_window_wayland_set_fullscreen (window, TRUE);
 
@@ -451,8 +507,12 @@ gst_vaapi_window_wayland_finalize (GObject * object)
   if (priv->event_queue)
     wl_display_roundtrip_queue (wl_display, priv->event_queue);
 
+  while (priv->frames)
+    frame_state_free ((FrameState *) priv->frames->data);
+
   g_clear_pointer (&priv->xdg_surface, xdg_surface_destroy);
   g_clear_pointer (&priv->wl_shell_surface, wl_shell_surface_destroy);
+  g_clear_pointer (&priv->video_subsurface, wl_subsurface_destroy);
   g_clear_pointer (&priv->surface, wl_surface_destroy);
   g_clear_pointer (&priv->event_queue, wl_event_queue_destroy);
 
@@ -461,27 +521,46 @@ gst_vaapi_window_wayland_finalize (GObject * object)
   G_OBJECT_CLASS (gst_vaapi_window_wayland_parent_class)->finalize (object);
 }
 
-static gboolean
-gst_vaapi_window_wayland_resize (GstVaapiWindow * window,
+static void
+gst_vaapi_window_wayland_update_opaque_region (GstVaapiWindow * window,
     guint width, guint height)
 {
   GstVaapiWindowWaylandPrivate *const priv =
       GST_VAAPI_WINDOW_WAYLAND_GET_PRIVATE (window);
-  GstVaapiDisplayWaylandPrivate *const priv_display =
-      GST_VAAPI_DISPLAY_WAYLAND_GET_PRIVATE (GST_VAAPI_WINDOW_DISPLAY (window));
+
+  g_mutex_lock (&priv->opaque_mutex);
+  priv->opaque_width = width;
+  priv->opaque_height = height;
+  g_mutex_unlock (&priv->opaque_mutex);
+}
+
+static gboolean
+gst_vaapi_window_wayland_resize (GstVaapiWindow * window,
+    guint width, guint height)
+{
+  if (window->use_foreign_window)
+    return TRUE;
 
   GST_DEBUG ("resize window, new size %ux%u", width, height);
 
-  if (priv->opaque_region)
-    wl_region_destroy (priv->opaque_region);
-  GST_VAAPI_WINDOW_LOCK_DISPLAY (window);
-  priv->opaque_region = wl_compositor_create_region (priv_display->compositor);
-  GST_VAAPI_WINDOW_UNLOCK_DISPLAY (window);
-  wl_region_add (priv->opaque_region, 0, 0, width, height);
+  gst_vaapi_window_wayland_update_opaque_region (window, width, height);
 
   return TRUE;
 }
 
+void
+gst_vaapi_window_wayland_set_render_rect (GstVaapiWindow * window, gint x,
+    gint y, gint width, gint height)
+{
+  GstVaapiWindowWaylandPrivate *const priv =
+      GST_VAAPI_WINDOW_WAYLAND_GET_PRIVATE (window);
+
+  if (priv->video_subsurface)
+    wl_subsurface_set_position (priv->video_subsurface, x, y);
+
+  gst_vaapi_window_wayland_update_opaque_region (window, width, height);
+}
+
 static inline gboolean
 frame_done (FrameState * frame)
 {
@@ -513,7 +592,6 @@ frame_release_callback (void *data, struct wl_buffer *wl_buffer)
   if (!frame->done)
     if (!frame_done (frame))
       GST_INFO ("cannot remove last frame because it didn't match or empty");
-  wl_buffer_destroy (wl_buffer);
   frame_state_free (frame);
 }
 
@@ -521,6 +599,297 @@ static const struct wl_buffer_listener frame_buffer_listener = {
   frame_release_callback
 };
 
+typedef enum
+{
+  GST_VAAPI_DMABUF_SUCCESS,
+  GST_VAAPI_DMABUF_BAD_FLAGS,
+  GST_VAAPI_DMABUF_BAD_FORMAT,
+  GST_VAAPI_DMABUF_BAD_MODIFIER,
+  GST_VAAPI_DMABUF_NOT_SUPPORTED,
+  GST_VAAPI_DMABUF_FLUSH,
+
+} GstVaapiDmabufStatus;
+
+#define DRM_FORMAT_MOD_INVALID 0xffffffffffffff
+
+static GstVaapiDmabufStatus
+dmabuf_format_supported (GstVaapiDisplayWaylandPrivate * const priv_display,
+    guint format, guint64 modifier)
+{
+  GArray *formats = priv_display->dmabuf_formats;
+  gboolean linear = FALSE;
+  gint i;
+
+  for (i = 0; i < formats->len; i++) {
+    GstDRMFormat fmt = g_array_index (formats, GstDRMFormat, i);
+    if (fmt.format == format && (fmt.modifier == modifier ||
+            (fmt.modifier == DRM_FORMAT_MOD_INVALID && modifier == 0)))
+      return GST_VAAPI_DMABUF_SUCCESS;
+    if (fmt.format == format && (fmt.modifier == 0 ||
+            fmt.modifier == DRM_FORMAT_MOD_INVALID))
+      linear = TRUE;
+  }
+  if (linear)
+    return GST_VAAPI_DMABUF_BAD_MODIFIER;
+  else
+    return GST_VAAPI_DMABUF_BAD_FORMAT;
+}
+
+GstVideoFormat
+check_format (GstVaapiDisplay * const display, gint index,
+    GstVideoFormat expect)
+{
+  GstVaapiDisplayWaylandPrivate *const priv_display =
+      GST_VAAPI_DISPLAY_WAYLAND_GET_PRIVATE (display);
+  GArray *formats = priv_display->dmabuf_formats;
+  GstDRMFormat fmt = g_array_index (formats, GstDRMFormat, index);
+  GstVideoFormat format = gst_vaapi_video_format_from_drm_format (fmt.format);
+  GstVaapiSurface *surface;
+
+  /* unkown formats should be filtered out in the display */
+  g_assert (format != GST_VIDEO_FORMAT_UNKNOWN);
+
+  if ((expect != GST_VIDEO_FORMAT_UNKNOWN) && (format != expect))
+    return GST_VIDEO_FORMAT_UNKNOWN;
+
+  surface = gst_vaapi_surface_new_with_format (display, format, 64, 64,
+      fmt.modifier == 0 ? GST_VAAPI_SURFACE_ALLOC_FLAG_LINEAR_STORAGE : 0);
+  if (!surface)
+    return GST_VIDEO_FORMAT_UNKNOWN;
+
+  gst_vaapi_surface_unref (surface);
+
+  return format;
+}
+
+GstVideoFormat
+choose_next_format (GstVaapiDisplay * const display, gint * next_index)
+{
+  GstVaapiDisplayWaylandPrivate *const priv_display =
+      GST_VAAPI_DISPLAY_WAYLAND_GET_PRIVATE (display);
+  GArray *formats = priv_display->dmabuf_formats;
+  GstVideoFormat format;
+  gint i;
+
+  if (*next_index < 0) {
+    *next_index = 0;
+    /* try GST_VIDEO_FORMAT_RGBA first */
+    for (i = 0; i < formats->len; i++) {
+      format = check_format (display, i, GST_VIDEO_FORMAT_RGBA);
+      if (format != GST_VIDEO_FORMAT_UNKNOWN)
+        return format;
+    }
+  }
+
+  for (i = *next_index; i < formats->len; i++) {
+    format = check_format (display, i, GST_VIDEO_FORMAT_UNKNOWN);
+    if (format != GST_VIDEO_FORMAT_UNKNOWN) {
+      *next_index = i + 1;
+      return format;
+    }
+  }
+  *next_index = formats->len;
+  return GST_VIDEO_FORMAT_UNKNOWN;
+}
+
+static GstVaapiDmabufStatus
+dmabuf_buffer_from_surface (GstVaapiWindow * window, GstVaapiSurface * surface,
+    guint va_flags, struct wl_buffer **out_buffer)
+{
+  GstVaapiDisplay *const display = GST_VAAPI_WINDOW_DISPLAY (window);
+  GstVaapiDisplayWaylandPrivate *const priv_display =
+      GST_VAAPI_DISPLAY_WAYLAND_GET_PRIVATE (display);
+  struct zwp_linux_buffer_params_v1 *params;
+  struct wl_buffer *buffer = NULL;
+  VADRMPRIMESurfaceDescriptor desc;
+  VAStatus status;
+  GstVaapiDmabufStatus ret;
+  guint format, i, j, plane = 0;
+
+  if (!priv_display->dmabuf)
+    return GST_VAAPI_DMABUF_NOT_SUPPORTED;
+
+  if ((va_flags & (VA_TOP_FIELD | VA_BOTTOM_FIELD)) != VA_FRAME_PICTURE)
+    return GST_VAAPI_DMABUF_BAD_FLAGS;
+
+  GST_VAAPI_WINDOW_LOCK_DISPLAY (window);
+  status = vaExportSurfaceHandle (GST_VAAPI_DISPLAY_VADISPLAY (display),
+      GST_VAAPI_SURFACE_ID (surface), VA_SURFACE_ATTRIB_MEM_TYPE_DRM_PRIME_2,
+      VA_EXPORT_SURFACE_SEPARATE_LAYERS | VA_EXPORT_SURFACE_READ_ONLY, &desc);
+  /* Try again with composed layers, in case the format is supported there */
+  if (status == VA_STATUS_ERROR_INVALID_SURFACE)
+    status = vaExportSurfaceHandle (GST_VAAPI_DISPLAY_VADISPLAY (display),
+        GST_VAAPI_SURFACE_ID (surface), VA_SURFACE_ATTRIB_MEM_TYPE_DRM_PRIME_2,
+        VA_EXPORT_SURFACE_COMPOSED_LAYERS | VA_EXPORT_SURFACE_READ_ONLY, &desc);
+  GST_VAAPI_WINDOW_UNLOCK_DISPLAY (window);
+
+  if (!vaapi_check_status (status, "vaExportSurfaceHandle()")) {
+    if (status == VA_STATUS_ERROR_UNIMPLEMENTED)
+      return GST_VAAPI_DMABUF_NOT_SUPPORTED;
+    else
+      return GST_VAAPI_DMABUF_BAD_FORMAT;
+  }
+
+  format = gst_vaapi_drm_format_from_va_fourcc (desc.fourcc);
+  params = zwp_linux_dmabuf_v1_create_params (priv_display->dmabuf);
+  for (i = 0; i < desc.num_layers; i++) {
+    for (j = 0; j < desc.layers[i].num_planes; ++j) {
+      gint object = desc.layers[i].object_index[j];
+      guint64 modifier = desc.objects[object].drm_format_modifier;
+
+      ret = dmabuf_format_supported (priv_display, format, modifier);
+      if (ret != GST_VAAPI_DMABUF_SUCCESS) {
+        GST_DEBUG ("skipping unsupported format/modifier %s/0x%"
+            G_GINT64_MODIFIER "x", gst_video_format_to_string
+            (gst_vaapi_video_format_from_drm_format (format)), modifier);
+        goto out;
+      }
+
+      zwp_linux_buffer_params_v1_add (params,
+          desc.objects[object].fd, plane, desc.layers[i].offset[j],
+          desc.layers[i].pitch[j], modifier >> 32,
+          modifier & G_GUINT64_CONSTANT (0xffffffff));
+      plane++;
+    }
+  }
+
+  buffer = zwp_linux_buffer_params_v1_create_immed (params, window->width,
+      window->height, format, 0);
+
+  if (!buffer)
+    ret = GST_VAAPI_DMABUF_NOT_SUPPORTED;
+
+out:
+  zwp_linux_buffer_params_v1_destroy (params);
+
+  for (i = 0; i < desc.num_objects; i++)
+    close (desc.objects[i].fd);
+
+  *out_buffer = buffer;
+  return ret;
+}
+
+static gboolean
+buffer_from_surface (GstVaapiWindow * window, GstVaapiSurface ** surf,
+    const GstVaapiRectangle * src_rect, const GstVaapiRectangle * dst_rect,
+    guint flags, struct wl_buffer **buffer)
+{
+  GstVaapiDisplay *const display = GST_VAAPI_WINDOW_DISPLAY (window);
+  GstVaapiWindowWaylandPrivate *const priv =
+      GST_VAAPI_WINDOW_WAYLAND_GET_PRIVATE (window);
+  GstVaapiSurface *surface;
+  GstVaapiDmabufStatus ret;
+  guint va_flags;
+  VAStatus status;
+  gint format_index = -1;
+
+  va_flags = from_GstVaapiSurfaceRenderFlags (flags);
+
+again:
+  surface = *surf;
+  if (priv->need_vpp) {
+    GstVaapiSurface *vpp_surface = NULL;
+    if (window->has_vpp) {
+      GST_LOG ("VPP: %s <%d, %d, %d, %d> -> %s <%d, %d, %d, %d>",
+          gst_video_format_to_string (gst_vaapi_surface_get_format (surface)),
+          src_rect->x, src_rect->y, src_rect->width, src_rect->height,
+          gst_video_format_to_string (window->surface_pool_format),
+          dst_rect->x, dst_rect->y, dst_rect->width, dst_rect->height);
+      vpp_surface = gst_vaapi_window_vpp_convert_internal (window, surface,
+          src_rect, dst_rect, flags);
+    }
+    if (G_UNLIKELY (!vpp_surface)) {
+      /* Not all formats are supported as destination format during VPP.
+         So try again with the next format if VPP fails. */
+      GstVideoFormat format = choose_next_format (display, &format_index);
+      if ((format != GST_VIDEO_FORMAT_UNKNOWN) && window->has_vpp) {
+        GST_DEBUG ("VPP failed. Try again with format %s",
+            gst_video_format_to_string (format));
+        gst_vaapi_window_set_vpp_format_internal (window, format, 0);
+        goto again;
+      } else {
+        GST_WARNING ("VPP failed. No supported format found.");
+        priv->dmabuf_broken = TRUE;
+      }
+    } else {
+      surface = vpp_surface;
+      va_flags = VA_FRAME_PICTURE;
+    }
+  }
+  if (!priv->dmabuf_broken) {
+    ret = dmabuf_buffer_from_surface (window, surface, va_flags, buffer);
+    switch (ret) {
+      case GST_VAAPI_DMABUF_SUCCESS:
+        goto out;
+      case GST_VAAPI_DMABUF_BAD_FLAGS:
+        /* FIXME: how should this be handed? */
+        break;
+      case GST_VAAPI_DMABUF_BAD_FORMAT:{
+        /* The Wayland server does not accept the current format or
+           vaExportSurfaceHandle() failed. Try again with a different format */
+        GstVideoFormat format = choose_next_format (display, &format_index);
+        if ((format != GST_VIDEO_FORMAT_UNKNOWN) && window->has_vpp) {
+          GST_DEBUG ("Failed to export buffer. Try again with format %s",
+              gst_video_format_to_string (format));
+          priv->need_vpp = TRUE;
+          gst_vaapi_window_set_vpp_format_internal (window, format, 0);
+          goto again;
+        }
+        if (window->has_vpp)
+          GST_WARNING ("Failed to export buffer and VPP not supported.");
+        else
+          GST_WARNING ("Failed to export buffer. No supported format found.");
+        priv->dmabuf_broken = TRUE;
+        break;
+      }
+      case GST_VAAPI_DMABUF_BAD_MODIFIER:
+        /* The format is supported by the Wayland server but not with the
+           current modifier. Try linear instead. */
+        if (window->has_vpp) {
+          GST_DEBUG ("Modifier rejected by the server. Try linear instead.");
+          priv->need_vpp = TRUE;
+          gst_vaapi_window_set_vpp_format_internal (window,
+              gst_vaapi_surface_get_format (surface),
+              GST_VAAPI_SURFACE_ALLOC_FLAG_LINEAR_STORAGE);
+          goto again;
+        }
+        GST_WARNING ("Modifier rejected by the server and VPP not supported.");
+        priv->dmabuf_broken = TRUE;
+        break;
+      case GST_VAAPI_DMABUF_NOT_SUPPORTED:
+        GST_DEBUG ("DMABuf protocol not supported");
+        priv->dmabuf_broken = TRUE;
+        break;
+      case GST_VAAPI_DMABUF_FLUSH:
+        return FALSE;
+    }
+  }
+
+  /* DMABuf is not available or does not work. Fall back to the old API.
+     There is no format negotiation so stick with NV12  */
+  gst_vaapi_window_set_vpp_format_internal (window, GST_VIDEO_FORMAT_NV12, 0);
+
+  GST_VAAPI_WINDOW_LOCK_DISPLAY (window);
+  status = vaGetSurfaceBufferWl (GST_VAAPI_DISPLAY_VADISPLAY (display),
+      GST_VAAPI_SURFACE_ID (surface),
+      va_flags & (VA_TOP_FIELD | VA_BOTTOM_FIELD), buffer);
+  GST_VAAPI_WINDOW_UNLOCK_DISPLAY (window);
+
+  if (window->has_vpp && !priv->need_vpp &&
+      (status == VA_STATUS_ERROR_FLAG_NOT_SUPPORTED ||
+          status == VA_STATUS_ERROR_UNIMPLEMENTED ||
+          status == VA_STATUS_ERROR_INVALID_IMAGE_FORMAT)) {
+    priv->need_vpp = TRUE;
+    goto again;
+  }
+  if (!vaapi_check_status (status, "vaGetSurfaceBufferWl()"))
+    return FALSE;
+
+out:
+  *surf = surface;
+  return TRUE;
+}
+
 static gboolean
 gst_vaapi_window_wayland_render (GstVaapiWindow * window,
     GstVaapiSurface * surface,
@@ -529,19 +898,25 @@ gst_vaapi_window_wayland_render (GstVaapiWindow * window,
 {
   GstVaapiWindowWaylandPrivate *const priv =
       GST_VAAPI_WINDOW_WAYLAND_GET_PRIVATE (window);
-  GstVaapiDisplay *const display = GST_VAAPI_WINDOW_DISPLAY (window);
+  GstVaapiDisplayWaylandPrivate *const priv_display =
+      GST_VAAPI_DISPLAY_WAYLAND_GET_PRIVATE (GST_VAAPI_WINDOW_DISPLAY (window));
   struct wl_display *const wl_display =
       GST_VAAPI_WINDOW_NATIVE_DISPLAY (window);
   struct wl_buffer *buffer;
   FrameState *frame;
-  guint width, height, va_flags;
-  VAStatus status;
+  guint width, height;
+  gboolean ret;
+
+  /* Skip rendering without valid window size. This can happen with a foreign
+     window if the render rectangle is not yet set. */
+  if (window->width == 0 || window->height == 0)
+    return TRUE;
 
   /* Check that we don't need to crop source VA surface */
   gst_vaapi_surface_get_size (surface, &width, &height);
   if (src_rect->x != 0 || src_rect->y != 0)
     priv->need_vpp = TRUE;
-  if (src_rect->width != width || src_rect->height != height)
+  if (src_rect->width != width)
     priv->need_vpp = TRUE;
 
   /* Check that we don't render to a subregion of this window */
@@ -550,43 +925,20 @@ gst_vaapi_window_wayland_render (GstVaapiWindow * window,
   if (dst_rect->width != window->width || dst_rect->height != window->height)
     priv->need_vpp = TRUE;
 
-  /* Try to construct a Wayland buffer from VA surface as is (without VPP) */
-  if (!priv->need_vpp) {
-    GST_VAAPI_WINDOW_LOCK_DISPLAY (window);
-    va_flags = from_GstVaapiSurfaceRenderFlags (flags);
-    status = vaGetSurfaceBufferWl (GST_VAAPI_DISPLAY_VADISPLAY (display),
-        GST_VAAPI_OBJECT_ID (surface),
-        va_flags & (VA_TOP_FIELD | VA_BOTTOM_FIELD), &buffer);
-    GST_VAAPI_WINDOW_UNLOCK_DISPLAY (window);
-    if (status == VA_STATUS_ERROR_FLAG_NOT_SUPPORTED ||
-        status == VA_STATUS_ERROR_UNIMPLEMENTED ||
-        status == VA_STATUS_ERROR_INVALID_IMAGE_FORMAT)
-      priv->need_vpp = TRUE;
-    else if (!vaapi_check_status (status, "vaGetSurfaceBufferWl()"))
-      return FALSE;
-  }
+  /* Check that the surface has the correct size for the window */
+  if (dst_rect->width != src_rect->width ||
+      dst_rect->height != src_rect->height)
+    priv->need_vpp = TRUE;
 
-  /* Try to construct a Wayland buffer with VPP */
+  ret = buffer_from_surface (window, &surface, src_rect, dst_rect, flags,
+      &buffer);
+  if (!ret)
+    return FALSE;
+
+  /* if need_vpp is set then the vpp happend */
   if (priv->need_vpp) {
-    if (window->has_vpp) {
-      GstVaapiSurface *const vpp_surface =
-          gst_vaapi_window_vpp_convert_internal (window, surface, src_rect,
-          dst_rect, flags);
-      if (G_UNLIKELY (!vpp_surface))
-        priv->need_vpp = FALSE;
-      else {
-        surface = vpp_surface;
-        width = window->width;
-        height = window->height;
-      }
-    }
-
-    GST_VAAPI_WINDOW_LOCK_DISPLAY (window);
-    status = vaGetSurfaceBufferWl (GST_VAAPI_DISPLAY_VADISPLAY (display),
-        GST_VAAPI_OBJECT_ID (surface), VA_FRAME_PICTURE, &buffer);
-    GST_VAAPI_WINDOW_UNLOCK_DISPLAY (window);
-    if (!vaapi_check_status (status, "vaGetSurfaceBufferWl()"))
-      return FALSE;
+    width = window->width;
+    height = window->height;
   }
 
   /* Wait for the previous frame to complete redraw */
@@ -614,17 +966,25 @@ gst_vaapi_window_wayland_render (GstVaapiWindow * window,
   wl_surface_attach (priv->surface, buffer, 0, 0);
   wl_surface_damage (priv->surface, 0, 0, width, height);
 
-  if (priv->opaque_region) {
-    wl_surface_set_opaque_region (priv->surface, priv->opaque_region);
-    wl_region_destroy (priv->opaque_region);
-    priv->opaque_region = NULL;
+  g_mutex_lock (&priv->opaque_mutex);
+  if (priv->opaque_width > 0) {
+    struct wl_region *opaque_region;
+    opaque_region = wl_compositor_create_region (priv_display->compositor);
+    wl_region_add (opaque_region, 0, 0, width, height);
+    wl_surface_set_opaque_region (priv->surface, opaque_region);
+    wl_region_destroy (opaque_region);
+    priv->opaque_width = 0;
+    priv->opaque_height = 0;
   }
+  g_mutex_unlock (&priv->opaque_mutex);
 
   wl_proxy_set_queue ((struct wl_proxy *) buffer, priv->event_queue);
   wl_buffer_add_listener (buffer, &frame_buffer_listener, frame);
 
+  frame->buffer = buffer;
   frame->callback = wl_surface_frame (priv->surface);
   wl_callback_add_listener (frame->callback, &frame_callback_listener, frame);
+  priv->frames = g_list_append (priv->frames, frame);
 
   wl_surface_commit (priv->surface);
   wl_display_flush (wl_display);
@@ -670,6 +1030,7 @@ gst_vaapi_window_wayland_class_init (GstVaapiWindowWaylandClass * klass)
   window_class->set_fullscreen = gst_vaapi_window_wayland_set_fullscreen;
   window_class->unblock = gst_vaapi_window_wayland_unblock;
   window_class->unblock_cancel = gst_vaapi_window_wayland_unblock_cancel;
+  window_class->set_render_rect = gst_vaapi_window_wayland_set_render_rect;
 
   signals[SIZE_CHANGED] = g_signal_new ("size-changed",
       G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST, 0, NULL, NULL, NULL,
@@ -691,7 +1052,7 @@ gst_vaapi_window_wayland_init (GstVaapiWindowWayland * window)
  * will be attached to the @display and remains invisible to the user
  * until gst_vaapi_window_show() is called.
  *
- * Return value: the newly allocated #GstVaapiWindow object
+ * Return value (transfer full): the newly allocated #GstVaapiWindow object
  */
 GstVaapiWindow *
 gst_vaapi_window_wayland_new (GstVaapiDisplay * display,
@@ -702,3 +1063,30 @@ gst_vaapi_window_wayland_new (GstVaapiDisplay * display,
   return gst_vaapi_window_new_internal (GST_TYPE_VAAPI_WINDOW_WAYLAND, display,
       GST_VAAPI_ID_INVALID, width, height);
 }
+
+/**
+ * gst_vaapi_window_wayland_new_with_surface:
+ * @display: a #GstVaapiDisplay
+ * @wl_surface: a Wayland surface pointer
+ *
+ * Creates a window with the specified @wl_surface. The window
+ * will be attached to the @display and remains invisible to the user
+ * until gst_vaapi_window_show() is called.
+ *
+ * Return value (transfer full): the newly allocated #GstVaapiWindow object
+ *
+ * Since: 1.18
+ */
+GstVaapiWindow *
+gst_vaapi_window_wayland_new_with_surface (GstVaapiDisplay * display,
+    guintptr wl_surface)
+{
+  g_return_val_if_fail (GST_VAAPI_IS_DISPLAY_WAYLAND (display), NULL);
+  g_return_val_if_fail (wl_surface, NULL);
+
+  GST_CAT_DEBUG (gst_debug_vaapi, "new window from surface 0x%"
+      G_GINTPTR_MODIFIER "x", wl_surface);
+
+  return gst_vaapi_window_new_internal (GST_TYPE_VAAPI_WINDOW_WAYLAND, display,
+      wl_surface, 0, 0);
+}
