@@ -29,6 +29,11 @@
 
 G_BEGIN_DECLS
 
+#define GST_VAAPI_FORMATS_ALL "{ ENCODED, " \
+  "NV12, YV12, I420, YUY2, UYVY, Y444, GRAY8, P010_10LE, P012_LE, VUYA, Y210, Y410, " \
+  "ARGB, xRGB, RGBA, RGBx, ABGR, xBGR, BGRA, BGRx, RGB16, RGB, BGR10A2_LE "  \
+  "}"
+
 const gchar *
 gst_vaapi_video_format_to_string (GstVideoFormat format);
 
@@ -59,6 +64,18 @@ gst_vaapi_video_format_from_chroma (guint chroma);
 GstVideoFormat
 gst_vaapi_video_format_get_best_native (GstVideoFormat format);
 
+GArray *
+gst_vaapi_video_format_get_formats_by_chroma (guint chroma);
+
+gboolean
+gst_vaapi_video_format_create_map (VAImageFormat * formats, guint n);
+
+guint
+gst_vaapi_drm_format_from_va_fourcc (guint32 fourcc);
+
+GstVideoFormat
+gst_vaapi_video_format_from_drm_format (guint drm_format);
+
 G_END_DECLS
 
 #endif /* GST_VAAPI_VIDEO_FORMAT_H */
