@@ -30,15 +30,14 @@
  * the unregistered vaapidecode, a #GstQueue, and the
  * #GstVaapiPostproc, if it is available and functional in the setup.
  *
- * It offers the functionality of #GstVaapiDecode and the many options
+ * It offers the functionality of GstVaapiDecoder and the many options
  * of #GstVaapiPostproc.
  *
- * <refsect2>
- * <title>Example launch line</title>
+ * ## Example launch line
+ *
  * |[
  * gst-launch-1.0 filesrc location=~/big_buck_bunny.mov ! qtdemux ! h264parse ! vaapidecodebin ! vaapisink
  * ]|
- * </refsect2>
  */
 
 #include "gstcompat.h"
@@ -58,7 +57,7 @@
 GST_DEBUG_CATEGORY_STATIC (gst_debug_vaapi_decode_bin);
 #define GST_CAT_DEFAULT gst_debug_vaapi_decode_bin
 
-#define DEFAULT_QUEUE_MAX_SIZE_BUFFERS 0
+#define DEFAULT_QUEUE_MAX_SIZE_BUFFERS 1
 #define DEFAULT_QUEUE_MAX_SIZE_BYTES   0
 #define DEFAULT_QUEUE_MAX_SIZE_TIME    0
 #define DEFAULT_DEINTERLACE_METHOD     GST_VAAPI_DEINTERLACE_METHOD_BOB
@@ -314,9 +313,8 @@ gst_vaapi_decode_bin_configure (GstVaapiDecodeBin * vaapidecbin)
 
   GST_INFO_OBJECT (vaapidecbin, "enabling VPP");
 
-  /* capsfilter to avoid negotiation with vaapidecode */
-  caps = gst_caps_from_string
-      ("video/x-raw(memory:VASurface), format=(string)NV12");
+  /* capsfilter to force memory:VASurface */
+  caps = gst_caps_from_string ("video/x-raw(memory:VASurface)");
   if (!caps)
     goto error_cannot_set_caps;
   capsfilter = gst_element_factory_make ("capsfilter", NULL);
@@ -400,7 +398,9 @@ gst_vaapi_decode_bin_init (GstVaapiDecodeBin * vaapidecbin)
 {
   GstPad *pad, *ghostpad;
 
-  vaapidecbin->deinterlace_method = DEFAULT_DEINTERLACE_METHOD;
+  vaapidecbin->max_size_bytes = DEFAULT_QUEUE_MAX_SIZE_BYTES;
+  vaapidecbin->max_size_buffers = DEFAULT_QUEUE_MAX_SIZE_BUFFERS;
+  vaapidecbin->max_size_time = DEFAULT_QUEUE_MAX_SIZE_TIME;
   vaapidecbin->disable_vpp = (g_getenv ("GST_VAAPI_DISABLE_VPP") != NULL);
 
   /* create the decoder */
