@@ -24,6 +24,7 @@
 #include "gstcompat.h"
 #include <gst/vaapi/gstvaapivalue.h>
 #include <gst/vaapi/gstvaapidisplay.h>
+#include <gst/vaapi/gstvaapiprofilecaps.h>
 #include "gstvaapiencode.h"
 #include "gstvaapipluginutil.h"
 #include "gstvaapivideometa.h"
@@ -103,33 +104,11 @@ gst_vaapiencode_src_query (GstVideoEncoder * encoder, GstQuery * query)
 
 typedef struct
 {
-  GstVaapiEncoderProp id;
+  guint id;
   GParamSpec *pspec;
   GValue value;
 } PropValue;
 
-static PropValue *
-prop_value_new (GstVaapiEncoderPropInfo * prop)
-{
-  static const GValue default_value = G_VALUE_INIT;
-  PropValue *prop_value;
-
-  if (!prop || !prop->pspec)
-    return NULL;
-
-  prop_value = g_slice_new (PropValue);
-  if (!prop_value)
-    return NULL;
-
-  prop_value->id = prop->prop;
-  prop_value->pspec = g_param_spec_ref (prop->pspec);
-
-  memcpy (&prop_value->value, &default_value, sizeof (prop_value->value));
-  g_value_init (&prop_value->value, prop->pspec->value_type);
-  g_param_value_set_default (prop->pspec, &prop_value->value);
-  return prop_value;
-}
-
 static void
 prop_value_free (PropValue * prop_value)
 {
@@ -146,50 +125,46 @@ prop_value_free (PropValue * prop_value)
   g_slice_free (PropValue, prop_value);
 }
 
+static PropValue *
+prop_value_new_entry (guint id, GParamSpec * pspec, const GValue * value)
+{
+  PropValue *prop_value;
+
+  if (!pspec)
+    return NULL;
+
+  prop_value = g_slice_new0 (PropValue);
+  if (!prop_value)
+    return NULL;
+
+  prop_value->id = id;
+  prop_value->pspec = g_param_spec_ref (pspec);
+  g_value_init (&prop_value->value, pspec->value_type);
+
+  g_assert (g_value_type_compatible (pspec->value_type, G_VALUE_TYPE (value)));
+  g_value_copy (value, &prop_value->value);
+
+  return prop_value;
+}
+
 static inline PropValue *
-prop_value_lookup (GstVaapiEncode * encode, guint prop_id)
+prop_value_lookup_entry (GPtrArray * prop_values, guint prop_id)
 {
-  GPtrArray *const prop_values = encode->prop_values;
+  guint i;
+  PropValue *prop_value;
+
+  if (prop_values == NULL)
+    return NULL;
+
+  for (i = 0; i < prop_values->len; i++) {
+    prop_value = g_ptr_array_index (prop_values, i);
+    if (prop_value->id == prop_id)
+      return prop_value;
+  }
 
-  if (prop_values &&
-      (prop_id >= PROP_BASE && prop_id < PROP_BASE + prop_values->len))
-    return g_ptr_array_index (prop_values, prop_id - PROP_BASE);
   return NULL;
 }
 
-static gboolean
-gst_vaapiencode_default_get_property (GstVaapiEncode * encode, guint prop_id,
-    GValue * value)
-{
-  PropValue *const prop_value = prop_value_lookup (encode, prop_id);
-
-  if (prop_value) {
-    g_value_copy (&prop_value->value, value);
-    return TRUE;
-  }
-  return FALSE;
-}
-
-static gboolean
-gst_vaapiencode_default_set_property (GstVaapiEncode * encode, guint prop_id,
-    const GValue * value)
-{
-  PropValue *const prop_value = prop_value_lookup (encode, prop_id);
-  GstVaapiEncoder *encoder = encode->encoder;
-
-  if (prop_value) {
-    g_value_copy (value, &prop_value->value);
-
-    if (encoder)
-      return (gst_vaapi_encoder_set_property (encoder, prop_value->id,
-              value) == GST_VAAPI_ENCODER_STATUS_SUCCESS);
-
-    return TRUE;
-  }
-
-  return FALSE;
-}
-
 static GstFlowReturn
 gst_vaapiencode_default_alloc_buffer (GstVaapiEncode * encode,
     GstVaapiCodedBuffer * coded_buf, GstBuffer ** outbuf_ptr)
@@ -278,9 +253,6 @@ gst_vaapiencode_push_frame (GstVaapiEncode * encode, gint64 timeout)
   GstVaapiEncoderStatus status;
   GstBuffer *out_buffer;
   GstFlowReturn ret;
-#if USE_H264_FEI_ENCODER
-  GstVaapiFeiVideoMeta *feimeta = NULL;
-#endif
 
   status = gst_vaapi_encoder_get_buffer_with_timeout (encode->encoder,
       &codedbuf_proxy, timeout);
@@ -306,14 +278,6 @@ gst_vaapiencode_push_frame (GstVaapiEncode * encode, gint64 timeout)
   ret = klass->alloc_buffer (encode,
       GST_VAAPI_CODED_BUFFER_PROXY_BUFFER (codedbuf_proxy), &out_buffer);
 
-#if USE_H264_FEI_ENCODER
-  if (klass->save_stats_to_meta) {
-    feimeta = klass->save_stats_to_meta (encode, codedbuf_proxy);
-    if (feimeta != NULL)
-      gst_buffer_set_vaapi_fei_video_meta (out_buffer, feimeta);
-  }
-#endif
-
   gst_vaapi_coded_buffer_proxy_replace (&codedbuf_proxy, NULL);
   if (ret != GST_FLOW_OK)
     goto error_allocate_buffer;
@@ -365,44 +329,63 @@ gst_vaapiencode_buffer_loop (GstVaapiEncode * encode)
   gst_pad_pause_task (GST_VAAPI_PLUGIN_BASE_SRC_PAD (encode));
 }
 
+static GArray *
+get_profiles (GstVaapiEncode * encode)
+{
+  GstVaapiEncodeClass *klass = GST_VAAPIENCODE_GET_CLASS (encode);
+  GArray *profiles = NULL;
+
+  if (klass->get_allowed_profiles) {
+    GstCaps *allowed =
+        gst_pad_get_allowed_caps (GST_VAAPI_PLUGIN_BASE_SRC_PAD (encode));
+    GST_LOG_OBJECT (encode,
+        "Get allowed sink caps from downstream %" GST_PTR_FORMAT, allowed);
+    if (allowed && !gst_caps_is_empty (allowed) && !gst_caps_is_any (allowed))
+      profiles = klass->get_allowed_profiles (encode, allowed);
+
+    if (allowed)
+      gst_caps_unref (allowed);
+
+    if (profiles)
+      return profiles;
+  }
+
+  profiles = gst_vaapi_encoder_get_available_profiles (encode->encoder);
+  return profiles;
+}
+
 static gboolean
 ensure_allowed_sinkpad_caps (GstVaapiEncode * encode)
 {
-  GstVaapiEncodeClass *klass = GST_VAAPIENCODE_GET_CLASS (encode);
-  GstCaps *out_caps, *raw_caps = NULL;
+  GstCaps *out_caps = NULL;
   GArray *formats = NULL;
   gboolean ret = FALSE;
-  GstVaapiProfile profile = GST_VAAPI_PROFILE_UNKNOWN;
+  GArray *profiles = NULL;
+  gint min_width, min_height, max_width, max_height;
+  guint mem_types;
 
   if (encode->allowed_sinkpad_caps)
     return TRUE;
   if (!encode->encoder)
     return TRUE;
 
-  out_caps = gst_caps_from_string (GST_VAAPI_MAKE_SURFACE_CAPS);
-  if (!out_caps)
-    goto failed_create_va_caps;
+  /* First, get all possible profiles. */
+  profiles = get_profiles (encode);
+  if (profiles == NULL)
+    goto failed_get_profiles;
 
-  if (klass->get_profile) {
-    GstCaps *allowed =
-        gst_pad_get_allowed_caps (GST_VAAPI_PLUGIN_BASE_SRC_PAD (encode));
-    if (allowed) {
-      if (!gst_caps_is_empty (allowed) && !gst_caps_is_any (allowed))
-        profile = klass->get_profile (allowed);
-      gst_caps_unref (allowed);
-    }
-  }
-
-  formats = gst_vaapi_encoder_get_surface_formats (encode->encoder, profile);
+  /* Then get all supported formats, all these formats should be recognized
+     in video-format map. */
+  formats = gst_vaapi_encoder_get_surface_attributes (encode->encoder, profiles,
+      &min_width, &min_height, &max_width, &max_height, &mem_types);
   if (!formats)
-    goto failed_get_formats;
+    goto failed_get_attributes;
 
-  raw_caps = gst_vaapi_video_format_new_template_caps_from_list (formats);
-  if (!raw_caps)
-    goto failed_create_raw_caps;
+  out_caps = gst_vaapi_build_caps_from_formats (formats, min_width, min_height,
+      max_width, max_height, mem_types);
+  if (!out_caps)
+    goto failed_create_caps;
 
-  out_caps = gst_caps_make_writable (out_caps);
-  gst_caps_append (out_caps, gst_caps_copy (raw_caps));
   gst_caps_replace (&encode->allowed_sinkpad_caps, out_caps);
   GST_INFO_OBJECT (encode, "Allowed sink caps %" GST_PTR_FORMAT,
       encode->allowed_sinkpad_caps);
@@ -410,27 +393,30 @@ ensure_allowed_sinkpad_caps (GstVaapiEncode * encode)
   ret = TRUE;
 
 bail:
+  if (!encode->allowed_sinkpad_caps)
+    encode->allowed_sinkpad_caps = gst_caps_new_empty ();
+
+  if (profiles)
+    g_array_unref (profiles);
   if (out_caps)
     gst_caps_unref (out_caps);
-  if (raw_caps)
-    gst_caps_unref (raw_caps);
   if (formats)
     g_array_unref (formats);
   return ret;
 
-failed_create_va_caps:
+failed_get_attributes:
   {
-    GST_WARNING_OBJECT (encode, "failed to create VA/GL sink caps");
-    return FALSE;
-  }
-failed_get_formats:
-  {
-    GST_WARNING_OBJECT (encode, "failed to get allowed surface formats");
+    GST_WARNING_OBJECT (encode, "failed to get surface attributes");
     goto bail;
   }
-failed_create_raw_caps:
+failed_create_caps:
   {
-    GST_WARNING_OBJECT (encode, "failed to create raw sink caps");
+    GST_WARNING_OBJECT (encode, "failed to create sink caps");
+    goto bail;
+  }
+failed_get_profiles:
+  {
+    GST_WARNING_OBJECT (encode, "failed to get supported profiles");
     goto bail;
   }
 }
@@ -491,8 +477,6 @@ static gboolean
 ensure_encoder (GstVaapiEncode * encode)
 {
   GstVaapiEncodeClass *klass = GST_VAAPIENCODE_GET_CLASS (encode);
-  GstVaapiEncoderStatus status;
-  GPtrArray *const prop_values = encode->prop_values;
   guint i;
 
   g_return_val_if_fail (klass->alloc_encoder, FALSE);
@@ -505,15 +489,17 @@ ensure_encoder (GstVaapiEncode * encode)
   if (!encode->encoder)
     return FALSE;
 
-  if (prop_values) {
-    for (i = 0; i < prop_values->len; i++) {
-      PropValue *const prop_value = g_ptr_array_index (prop_values, i);
-      status = gst_vaapi_encoder_set_property (encode->encoder, prop_value->id,
-          &prop_value->value);
-      if (status != GST_VAAPI_ENCODER_STATUS_SUCCESS)
-        return FALSE;
+  if (encode->prop_values && encode->prop_values->len) {
+    for (i = 0; i < encode->prop_values->len; i++) {
+      PropValue *const prop_value = g_ptr_array_index (encode->prop_values, i);
+      g_object_set_property ((GObject *) encode->encoder,
+          g_param_spec_get_name (prop_value->pspec), &prop_value->value);
     }
+    /* clear alll the cache */
+    g_ptr_array_unref (encode->prop_values);
+    encode->prop_values = NULL;
   }
+
   return TRUE;
 }
 
@@ -625,9 +611,8 @@ gst_vaapiencode_set_format (GstVideoEncoder * venc, GstVideoCodecState * state)
       gst_tag_list_add (tags, GST_TAG_MERGE_REPLACE, GST_TAG_ENCODER, encoder,
           NULL);
 
-    if ((codec =
-            gst_vaapi_codec_get_name (gst_vaapi_profile_get_codec
-                (gst_vaapi_profile_from_caps (state->caps)))))
+    if ((codec = gst_vaapi_codec_get_name
+            (gst_vaapi_get_codec_from_caps (state->caps))))
       gst_tag_list_add (tags, GST_TAG_MERGE_REPLACE, GST_TAG_CODEC, codec,
           NULL);
 
@@ -650,10 +635,6 @@ gst_vaapiencode_handle_frame (GstVideoEncoder * venc,
   GstFlowReturn ret;
   GstBuffer *buf;
   GstTaskState task_state;
-#if USE_H264_FEI_ENCODER
-  GstVaapiFeiVideoMeta *feimeta = NULL;
-  GstVaapiEncodeClass *const klass = GST_VAAPIENCODE_GET_CLASS (venc);
-#endif
 
   task_state = gst_pad_get_task_state (srcpad);
   if (task_state == GST_TASK_STOPPED || task_state == GST_TASK_PAUSED)
@@ -678,12 +659,6 @@ gst_vaapiencode_handle_frame (GstVideoEncoder * venc,
   if (!proxy)
     goto error_buffer_no_surface_proxy;
 
-#if USE_H264_FEI_ENCODER
-  feimeta = gst_buffer_get_vaapi_fei_video_meta (buf);
-  if (feimeta && klass->load_control_data)
-    klass->load_control_data (encode, feimeta, proxy);
-#endif
-
   gst_video_codec_frame_set_user_data (frame,
       gst_vaapi_surface_proxy_ref (proxy),
       (GDestroyNotify) gst_vaapi_surface_proxy_unref);
@@ -868,7 +843,7 @@ gst_vaapiencode_init (GstVaapiEncode * encode)
   GstVaapiPluginBase *const plugin = GST_VAAPI_PLUGIN_BASE (encode);
 
   gst_vaapi_plugin_base_init (GST_VAAPI_PLUGIN_BASE (encode), GST_CAT_DEFAULT);
-  gst_pad_use_fixed_caps (plugin->srcpad);
+  gst_pad_use_fixed_caps (GST_VAAPI_PLUGIN_BASE_SRC_PAD (plugin));
 }
 
 static void
@@ -902,62 +877,164 @@ gst_vaapiencode_class_init (GstVaapiEncodeClass * klass)
   venc_class->flush = GST_DEBUG_FUNCPTR (gst_vaapiencode_flush);
   venc_class->sink_event = GST_DEBUG_FUNCPTR (gst_vaapiencode_sink_event);
 
-  klass->get_property = gst_vaapiencode_default_get_property;
-  klass->set_property = gst_vaapiencode_default_set_property;
   klass->alloc_buffer = gst_vaapiencode_default_alloc_buffer;
 
   venc_class->src_query = GST_DEBUG_FUNCPTR (gst_vaapiencode_src_query);
   venc_class->sink_query = GST_DEBUG_FUNCPTR (gst_vaapiencode_sink_query);
+
+  gst_type_mark_as_plugin_api (GST_TYPE_VAAPIENCODE, 0);
 }
 
-static inline GPtrArray *
-get_properties (GstVaapiEncodeClass * klass)
+/* Only used by the drived class */
+void
+gst_vaapiencode_set_property_subclass (GObject * object,
+    guint prop_id, const GValue * value, GParamSpec * pspec)
 {
-  return klass->get_properties ? klass->get_properties () : NULL;
-}
+  GstVaapiEncodeClass *const encode_class = GST_VAAPIENCODE_GET_CLASS (object);
+  GstVaapiEncode *const encode = GST_VAAPIENCODE_CAST (object);
+  PropValue *prop_value;
 
-gboolean
-gst_vaapiencode_init_properties (GstVaapiEncode * encode)
-{
-  GPtrArray *const props = get_properties (GST_VAAPIENCODE_GET_CLASS (encode));
-  guint i;
-
-  /* XXX: use base_init()/base_finalize() to avoid multiple initializations */
-  if (!props)
-    return FALSE;
-
-  encode->prop_values =
-      g_ptr_array_new_full (props->len, (GDestroyNotify) prop_value_free);
-  if (!encode->prop_values) {
-    g_ptr_array_unref (props);
-    return FALSE;
+  if (prop_id <= PROP_BASE || prop_id >= encode_class->prop_num) {
+    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+    return;
   }
 
-  for (i = 0; i < props->len; i++) {
-    PropValue *const prop_value = prop_value_new ((GstVaapiEncoderPropInfo *)
-        g_ptr_array_index (props, i));
-    if (!prop_value)
-      return FALSE;
-    g_ptr_array_add (encode->prop_values, prop_value);
+  /* direct forward the property to encoder */
+  if (encode->encoder) {
+    g_object_set_property ((GObject *) encode->encoder,
+        g_param_spec_get_name (pspec), value);
+    return;
   }
-  g_ptr_array_unref (props);
-  return TRUE;
+
+  if (encode->prop_values) {
+    /* Delete the same prop already in cache */
+    prop_value = prop_value_lookup_entry (encode->prop_values, prop_id);
+    if (prop_value)
+      g_ptr_array_remove (encode->prop_values, prop_value);
+  } else {
+    encode->prop_values =
+        g_ptr_array_new_with_free_func ((GDestroyNotify) prop_value_free);
+  }
+
+  /* The encoder is delay created, we need to cache the property setting */
+  prop_value = prop_value_new_entry (prop_id, pspec, value);
+  g_ptr_array_add (encode->prop_values, prop_value);
 }
 
+/* Only used by the drived class */
+void
+gst_vaapiencode_get_property_subclass (GObject * object, guint prop_id,
+    GValue * value, GParamSpec * pspec)
+{
+  GstVaapiEncodeClass *const encode_class = GST_VAAPIENCODE_GET_CLASS (object);
+  GstVaapiEncode *const encode = GST_VAAPIENCODE_CAST (object);
+  PropValue *prop_value = NULL;
+
+  if (prop_id <= PROP_BASE || prop_id >= encode_class->prop_num) {
+    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+    return;
+  }
+
+  /* direct forward the property to encoder */
+  if (encode->encoder) {
+    g_object_get_property ((GObject *) encode->encoder,
+        g_param_spec_get_name (pspec), value);
+    return;
+  }
+
+  if (encode->prop_values)
+    prop_value = prop_value_lookup_entry (encode->prop_values, prop_id);
+
+  if (prop_value) {
+    /* In the cache */
+    g_value_copy (&prop_value->value, value);
+  } else {
+    /* set the default value */
+    g_param_value_set_default (pspec, value);
+  }
+}
+
+/* Called by drived class to install all properties. The encode base class
+   does not have any property, all the properties of the according encoderXXX
+   class are installed to encodeXXX class. */
 gboolean
-gst_vaapiencode_class_init_properties (GstVaapiEncodeClass * klass)
+gst_vaapiencode_class_install_properties (GstVaapiEncodeClass * klass,
+    GObjectClass * encoder_class)
 {
   GObjectClass *const object_class = G_OBJECT_CLASS (klass);
-  GPtrArray *const props = get_properties (klass);
-  guint i;
+  guint i, n_props, installed;
+  GParamSpec **specs = NULL;
+  GParamSpec *pspec;
+  GParamSpec *new_spec;
+  GParamFlags flags;
 
-  if (!props)
+  if (encoder_class)
+    specs = g_object_class_list_properties (encoder_class, &n_props);
+  if (!specs)
     return FALSE;
 
-  for (i = 0; i < props->len; i++) {
-    GstVaapiEncoderPropInfo *const prop = g_ptr_array_index (props, i);
-    g_object_class_install_property (object_class, PROP_BASE + i, prop->pspec);
+  installed = 0;
+  for (i = 0; i < n_props; i++) {
+    pspec = specs[i];
+
+    /* Encoder do not want to expose */
+    if (!(pspec->flags & GST_VAAPI_PARAM_ENCODER_EXPOSURE))
+      continue;
+    /* Can only set on encoder init time */
+    if (pspec->flags & G_PARAM_CONSTRUCT_ONLY)
+      continue;
+
+    /* filter out the G_PARAM_CONSTRUCT, the encoder created later, no need
+       to set the init value in encode.
+       Also no GST_VAAPI_PARAM_ENCODER_EXPOSURE */
+    flags = pspec->flags & (~(GST_VAAPI_PARAM_ENCODER_EXPOSURE |
+            G_PARAM_CONSTRUCT));
+
+    if (G_PARAM_SPEC_TYPE (pspec) == G_TYPE_PARAM_UINT) {
+      GParamSpecUInt *pspecuint = G_PARAM_SPEC_UINT (pspec);
+      new_spec = g_param_spec_uint (g_param_spec_get_name (pspec),
+          g_param_spec_get_nick (pspec), g_param_spec_get_blurb (pspec),
+          pspecuint->minimum, pspecuint->maximum,
+          pspecuint->default_value, flags);
+    } else if (G_PARAM_SPEC_TYPE (pspec) == G_TYPE_PARAM_INT) {
+      GParamSpecInt *pspecint = G_PARAM_SPEC_INT (pspec);
+      new_spec = g_param_spec_int (g_param_spec_get_name (pspec),
+          g_param_spec_get_nick (pspec), g_param_spec_get_blurb (pspec),
+          pspecint->minimum, pspecint->maximum, pspecint->default_value, flags);
+    } else if (G_IS_PARAM_SPEC_ENUM (pspec)) {
+      GParamSpecEnum *pspecenum = G_PARAM_SPEC_ENUM (pspec);
+      new_spec = g_param_spec_enum (g_param_spec_get_name (pspec),
+          g_param_spec_get_nick (pspec),
+          g_param_spec_get_blurb (pspec),
+          pspec->value_type, pspecenum->default_value, flags);
+    } else if (G_IS_PARAM_SPEC_BOOLEAN (pspec)) {
+      GParamSpecBoolean *pspecbool = G_PARAM_SPEC_BOOLEAN (pspec);
+      new_spec = g_param_spec_boolean (g_param_spec_get_name (pspec),
+          g_param_spec_get_nick (pspec), g_param_spec_get_blurb (pspec),
+          pspecbool->default_value, flags);
+    } else if (G_IS_PARAM_SPEC_FLAGS (pspec)) {
+      GParamSpecFlags *pspecflags = G_PARAM_SPEC_FLAGS (pspec);
+      new_spec = g_param_spec_flags (g_param_spec_get_name (pspec),
+          g_param_spec_get_nick (pspec), g_param_spec_get_blurb (pspec),
+          pspec->value_type, pspecflags->default_value, flags);
+    } else if (GST_IS_PARAM_SPEC_ARRAY_LIST (pspec)) {
+      GstParamSpecArray *pspecarray = GST_PARAM_SPEC_ARRAY_LIST (pspec);
+      new_spec = gst_param_spec_array (g_param_spec_get_name (pspec),
+          g_param_spec_get_nick (pspec), g_param_spec_get_blurb (pspec),
+          pspecarray->element_spec, flags);
+    } else {
+      GST_WARNING ("encoder's %s property has a unimplemented"
+          " type to expose to encode, the encode may lose the %s property",
+          g_param_spec_get_name (pspec), g_param_spec_get_name (pspec));
+      continue;
+    }
+
+    g_object_class_install_property (object_class, PROP_BASE + 1 + installed,
+        new_spec);
+    installed++;
   }
-  g_ptr_array_unref (props);
+
+  g_free (specs);
+  klass->prop_num = PROP_BASE + 1 + installed;
   return TRUE;
 }
