@@ -24,6 +24,8 @@
 
 #include "gstcompat.h"
 #include "gstvaapivideocontext.h"
+#include <gst/vaapi/gstvaapiprofilecaps.h>
+#include <gst/vaapi/gstvaapiutils.h>
 #if USE_DRM
 # include <gst/vaapi/gstvaapidisplay_drm.h>
 #endif
@@ -156,6 +158,11 @@ gst_vaapi_get_display_type_from_gl (GstGLDisplayType gl_display_type,
     case GST_GL_DISPLAY_TYPE_EGL:{
       return GST_VAAPI_DISPLAY_TYPE_EGL;
     }
+#endif
+#if USE_DRM
+    case GST_GL_DISPLAY_TYPE_GBM:{
+      return GST_VAAPI_DISPLAY_TYPE_DRM;
+    }
 #endif
     default:
       /* unsupported display. Still DRM may work. */
@@ -482,7 +489,7 @@ gst_vaapi_apply_composition (GstVaapiSurface * surface, GstBuffer * buffer)
   if (cmeta)
     composition = cmeta->overlay;
   return gst_vaapi_surface_set_subpictures_from_composition (surface,
-      composition, TRUE);
+      composition);
 }
 
 gboolean
@@ -676,7 +683,7 @@ gst_vaapi_find_preferred_caps_feature (GstPad * pad, GstCaps * allowed_caps,
     caps = gst_caps_new_full (gst_structure_copy (structure), NULL);
     if (!caps)
       continue;
-    gst_caps_set_features (caps, 0, gst_caps_features_copy (features));
+    gst_caps_set_features_simple (caps, gst_caps_features_copy (features));
 
     for (j = 0; j < G_N_ELEMENTS (feature_list); j++) {
       if (gst_vaapi_caps_feature_contains (caps, feature_list[j])
@@ -692,33 +699,23 @@ gst_vaapi_find_preferred_caps_feature (GstPad * pad, GstCaps * allowed_caps,
       break;
   }
 
-  if (!caps)
-    goto cleanup;
-
 find_format:
+  if (!caps)
+    caps = gst_caps_ref (out_caps);
+
   if (out_format_ptr) {
     GstVideoFormat out_format;
     GstStructure *structure = NULL;
     const GValue *format_list;
-    GstCapsFeatures *features;
 
-    /* if the best feature is SystemMemory, we should choose the
-     * vidoe/x-raw caps in the filtered peer caps set. If not, use
-     * the first caps, which is the preferred by downstream. */
-    if (feature == GST_VAAPI_CAPS_FEATURE_SYSTEM_MEMORY) {
-      gst_caps_replace (&caps, out_caps);
-      num_structures = gst_caps_get_size (caps);
-      for (i = 0; i < num_structures; i++) {
+    num_structures = gst_caps_get_size (caps);
+    for (i = 0; i < num_structures; i++) {
+      GstCapsFeatures *const features = gst_caps_get_features (caps, i);
+      if (gst_caps_features_contains (features,
+              gst_vaapi_caps_feature_to_string (feature))) {
         structure = gst_caps_get_structure (caps, i);
-        features = gst_caps_get_features (caps, i);
-        if (!gst_caps_features_is_any (features)
-            && gst_caps_features_contains (features,
-                gst_vaapi_caps_feature_to_string
-                (GST_VAAPI_CAPS_FEATURE_SYSTEM_MEMORY)))
-          break;
+        break;
       }
-    } else {
-      structure = gst_caps_get_structure (caps, 0);
     }
     if (!structure)
       goto cleanup;
@@ -954,6 +951,8 @@ gst_vaapi_driver_is_whitelisted (GstVaapiDisplay * display)
   guint i;
   static const gchar *whitelist[] = {
     "Intel i965 driver",
+    "Intel iHD driver",
+    "Mesa Gallium driver",
     NULL
   };
 
@@ -1007,3 +1006,343 @@ gst_vaapi_codecs_has_codec (GArray * codecs, GstVaapiCodec codec)
   }
   return FALSE;
 }
+
+/**
+ * gst_vaapi_encoder_get_profiles_from_caps:
+ * @caps: a #GstCaps to detect
+ * @func: a #GstVaapiStrToProfileFunc
+ *
+ * This function will detect all profile strings in @caps and
+ * return the according GstVaapiProfile in array.
+ *
+ * Return: A #GArray of @GstVaapiProfile if succeed, %NULL if fail.
+ **/
+GArray *
+gst_vaapi_encoder_get_profiles_from_caps (GstCaps * caps,
+    GstVaapiStrToProfileFunc func)
+{
+  guint i, j;
+  GstVaapiProfile profile;
+  GArray *profiles = NULL;
+
+  if (!caps)
+    return NULL;
+
+  profiles = g_array_new (FALSE, FALSE, sizeof (GstVaapiProfile));
+  if (!profiles)
+    return NULL;
+
+  for (i = 0; i < gst_caps_get_size (caps); i++) {
+    GstStructure *const structure = gst_caps_get_structure (caps, i);
+    const GValue *const value = gst_structure_get_value (structure, "profile");
+
+    if (value && G_VALUE_HOLDS_STRING (value)) {
+      const gchar *str = g_value_get_string (value);
+      if (str) {
+        profile = func (str);
+        if (profile == GST_VAAPI_PROFILE_H264_BASELINE)
+          profile = GST_VAAPI_PROFILE_H264_CONSTRAINED_BASELINE;
+        if (profile != GST_VAAPI_PROFILE_UNKNOWN)
+          g_array_append_val (profiles, profile);
+      }
+    } else if (value && GST_VALUE_HOLDS_LIST (value)) {
+      const GValue *v;
+      const gchar *str;
+      for (j = 0; j < gst_value_list_get_size (value); j++) {
+        v = gst_value_list_get_value (value, j);
+        if (!v || !G_VALUE_HOLDS_STRING (v))
+          continue;
+
+        str = g_value_get_string (v);
+        if (str) {
+          profile = func (str);
+          if (profile != GST_VAAPI_PROFILE_UNKNOWN)
+            g_array_append_val (profiles, profile);
+        }
+      }
+    }
+  }
+
+  if (profiles->len == 0) {
+    g_array_unref (profiles);
+    profiles = NULL;
+  }
+
+  return profiles;
+}
+
+void
+gst_vaapi_caps_set_width_and_height_range (GstCaps * caps, gint min_width,
+    gint min_height, gint max_width, gint max_height)
+{
+  guint size, i;
+  GstStructure *structure;
+
+  /* Set the width/height info to caps */
+  size = gst_caps_get_size (caps);
+  for (i = 0; i < size; i++) {
+    structure = gst_caps_get_structure (caps, i);
+    if (!structure)
+      continue;
+    gst_structure_set (structure, "width", GST_TYPE_INT_RANGE, min_width,
+        max_width, "height", GST_TYPE_INT_RANGE, min_height, max_height,
+        "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1, NULL);
+  }
+}
+
+/**
+ * gst_vaapi_build_caps_from_formats:
+ * @formats: an array of supported #GstVideoFormat
+ * @min_width: the min supported width
+ * @min_height: the min supported height
+ * @max_width: the max supported width
+ * @max_height: the max supported height
+ * @mem_types: the supported VA mem types
+ *
+ * This function generates a #GstCaps based on the information such as
+ * formats, width and height.
+ *
+ * Return: A #GstCaps.
+ **/
+GstCaps *
+gst_vaapi_build_caps_from_formats (GArray * formats, gint min_width,
+    gint min_height, gint max_width, gint max_height, guint mem_types)
+{
+  GstCaps *out_caps, *raw_caps, *va_caps, *dma_caps;
+
+  dma_caps = NULL;
+
+  raw_caps = gst_vaapi_video_format_new_template_caps_from_list (formats);
+  if (!raw_caps)
+    return NULL;
+  gst_vaapi_caps_set_width_and_height_range (raw_caps, min_width, min_height,
+      max_width, max_height);
+
+  va_caps = gst_caps_copy (raw_caps);
+  gst_caps_set_features_simple (va_caps,
+      gst_caps_features_from_string (GST_CAPS_FEATURE_MEMORY_VAAPI_SURFACE));
+
+  if (gst_vaapi_mem_type_supports (mem_types,
+          GST_VAAPI_BUFFER_MEMORY_TYPE_DMA_BUF)) {
+    dma_caps = gst_caps_copy (raw_caps);
+    gst_caps_set_features_simple (dma_caps,
+        gst_caps_features_from_string (GST_CAPS_FEATURE_MEMORY_DMABUF));
+  }
+
+  out_caps = va_caps;
+  if (dma_caps)
+    gst_caps_append (out_caps, dma_caps);
+  gst_caps_append (out_caps, raw_caps);
+
+  return out_caps;
+}
+
+/**
+ * gst_vaapi_build_template_raw_caps_by_codec:
+ * @display: a #GstVaapiDisplay
+ * @usage: used for encode, decode or postproc
+ * @codec: a #GstVaapiCodec specify the codec to detect
+ * @extra_fmts: a #GArray of extra #GstVideoFormat
+ *
+ * Called by vaapi elements to detect the all possible video formats belong to
+ * the specified codec and build the caps. Only YUV kinds of formats are detected
+ * because so far almost all codecs use YUV kinds of formats as input/output.
+ * extra_fmts can specified more formats to be included.
+ *
+ * Returns: a built #GstCaps if succeeds, or %NULL if error.
+ **/
+GstCaps *
+gst_vaapi_build_template_raw_caps_by_codec (GstVaapiDisplay * display,
+    GstVaapiContextUsage usage, GstVaapiCodec codec, GArray * extra_fmts)
+{
+  GArray *profiles = NULL;
+  GArray *supported_fmts = NULL;
+  GstCaps *out_caps = NULL;
+  guint i, e;
+  GstVaapiProfile profile;
+  guint value;
+  guint chroma;
+  GstVaapiChromaType gst_chroma;
+  GstVaapiEntrypoint entrypoint_start, entrypoint_end;
+
+  if (usage == GST_VAAPI_CONTEXT_USAGE_ENCODE) {
+    profiles = gst_vaapi_display_get_encode_profiles (display);
+    entrypoint_start = GST_VAAPI_ENTRYPOINT_SLICE_ENCODE;
+    entrypoint_end = GST_VAAPI_ENTRYPOINT_SLICE_ENCODE_LP;
+  } else if (usage == GST_VAAPI_CONTEXT_USAGE_DECODE) {
+    profiles = gst_vaapi_display_get_decode_profiles (display);
+    entrypoint_start = GST_VAAPI_ENTRYPOINT_VLD;
+    entrypoint_end = GST_VAAPI_ENTRYPOINT_MOCO;
+  }
+  /* TODO: VPP */
+
+  if (!profiles)
+    goto out;
+
+  chroma = 0;
+  for (i = 0; i < profiles->len; i++) {
+    profile = g_array_index (profiles, GstVaapiProfile, i);
+    if (gst_vaapi_profile_get_codec (profile) != codec)
+      continue;
+
+    for (e = entrypoint_start; e <= entrypoint_end; e++) {
+      if (!gst_vaapi_get_config_attribute (display,
+              gst_vaapi_profile_get_va_profile (profile),
+              gst_vaapi_entrypoint_get_va_entrypoint (e),
+              VAConfigAttribRTFormat, &value))
+        continue;
+
+      chroma |= value;
+    }
+  }
+
+  if (!chroma)
+    goto out;
+
+  for (gst_chroma = GST_VAAPI_CHROMA_TYPE_YUV420;
+      gst_chroma <= GST_VAAPI_CHROMA_TYPE_YUV444_12BPP; gst_chroma++) {
+    GArray *fmts;
+    if (!(chroma & from_GstVaapiChromaType (gst_chroma)))
+      continue;
+
+    fmts = gst_vaapi_video_format_get_formats_by_chroma (gst_chroma);
+    if (!fmts)
+      continue;
+
+    /* One format can not belong to different chroma, no need to merge */
+    if (supported_fmts == NULL) {
+      supported_fmts = fmts;
+    } else {
+      for (i = 0; i < fmts->len; i++)
+        g_array_append_val (supported_fmts,
+            g_array_index (fmts, GstVideoFormat, i));
+      g_array_unref (fmts);
+    }
+  }
+
+  if (!supported_fmts)
+    goto out;
+
+  if (extra_fmts) {
+    for (i = 0; i < extra_fmts->len; i++)
+      g_array_append_val (supported_fmts,
+          g_array_index (extra_fmts, GstVideoFormat, i));
+  }
+
+  out_caps = gst_vaapi_build_caps_from_formats (supported_fmts, 1, 1,
+      G_MAXINT, G_MAXINT,
+      from_GstVaapiBufferMemoryType (GST_VAAPI_BUFFER_MEMORY_TYPE_DMA_BUF));
+
+out:
+  if (profiles)
+    g_array_unref (profiles);
+  if (supported_fmts)
+    g_array_unref (supported_fmts);
+
+  return out_caps;
+}
+
+/**
+ * gst_vaapi_structure_set_profiles:
+ * @st: a #GstStructure
+ * @list: a %NULL-terminated array of strings
+ *
+ * The @list of profiles are set in @st
+ **/
+void
+gst_vaapi_structure_set_profiles (GstStructure * st, gchar ** list)
+{
+  guint i;
+  GValue vlist = G_VALUE_INIT;
+  GValue value = G_VALUE_INIT;
+
+  g_value_init (&vlist, GST_TYPE_LIST);
+  g_value_init (&value, G_TYPE_STRING);
+
+  for (i = 0; list[i]; i++) {
+    g_value_set_string (&value, list[i]);
+    gst_value_list_append_value (&vlist, &value);
+  }
+
+  if (i == 1)
+    gst_structure_set_value (st, "profile", &value);
+  else if (i > 1)
+    gst_structure_set_value (st, "profile", &vlist);
+
+  g_value_unset (&value);
+  g_value_unset (&vlist);
+}
+
+/**
+ * gst_vaapi_build_template_coded_caps_by_codec:
+ * @display: a #GstVaapiDisplay
+ * @usage: used for encode or decode
+ * @codec: a #GstVaapiCodec specify the codec to detect
+ * @caps_str: a string of basic caps
+ *
+ * Called by vaapi elements to detect the all possible profiles belong to the
+ * specified codec and build the caps based on the basic caps description.
+ *
+ * Returns: a built #GstCaps if succeeds, or %NULL if error.
+ **/
+GstCaps *
+gst_vaapi_build_template_coded_caps_by_codec (GstVaapiDisplay * display,
+    GstVaapiContextUsage usage, GstVaapiCodec codec, const char *caps_str,
+    GstVaapiProfileToStrFunc func)
+{
+  GValue v_profiles = G_VALUE_INIT;
+  GValue v_profile = G_VALUE_INIT;
+  GstCaps *caps = NULL;
+  guint i, num;
+  GArray *profiles = NULL;
+  GstVaapiProfile profile;
+  const gchar *str;
+
+  caps = gst_caps_from_string (caps_str);
+  if (!caps)
+    goto out;
+
+  if (!func)
+    goto out;
+
+  /* If no profiles, just ignore the profile field. */
+  if (usage == GST_VAAPI_CONTEXT_USAGE_ENCODE) {
+    profiles = gst_vaapi_display_get_encode_profiles (display);
+  } else if (usage == GST_VAAPI_CONTEXT_USAGE_DECODE) {
+    profiles = gst_vaapi_display_get_decode_profiles (display);
+  }
+  if (!profiles || profiles->len == 0)
+    goto out;
+
+  num = 0;
+  g_value_init (&v_profiles, GST_TYPE_LIST);
+  g_value_init (&v_profile, G_TYPE_STRING);
+
+  for (i = 0; i < profiles->len; i++) {
+    profile = g_array_index (profiles, GstVaapiProfile, i);
+    if (gst_vaapi_profile_get_codec (profile) != codec)
+      continue;
+
+    str = func (profile);
+    if (!str)
+      continue;
+
+    g_value_set_string (&v_profile, str);
+    num++;
+    gst_value_list_append_value (&v_profiles, &v_profile);
+  }
+
+  if (num == 1) {
+    gst_caps_set_value (caps, "profile", &v_profile);
+  } else if (num > 1) {
+    gst_caps_set_value (caps, "profile", &v_profiles);
+  }
+
+out:
+  g_value_unset (&v_profile);
+  g_value_unset (&v_profiles);
+  if (profiles)
+    g_array_unref (profiles);
+
+  return caps;
+}
