@@ -26,16 +26,15 @@
  * SECTION:element-vaapisink
  * @short_description: A VA-API based video sink
  *
- * vaapisink renders video frames to a drawable (X #Window) on a local
+ * vaapisink renders video frames to a drawable (X Window) on a local
  * display using the Video Acceleration (VA) API. The element will
  * create its own internal window and render into it.
  *
- * <refsect2>
- * <title>Example launch line</title>
+ * ## Example launch line
+ *
  * |[
  * gst-launch-1.0 videotestsrc ! vaapisink
  * ]|
- * </refsect2>
  */
 
 #include "gstcompat.h"
@@ -562,11 +561,31 @@ gst_vaapisink_wayland_create_window (GstVaapiSink * sink, guint width,
   return TRUE;
 }
 
+static gboolean
+gst_vaapisink_wayland_create_window_from_handle (GstVaapiSink * sink,
+    guintptr window)
+{
+  GstVaapiDisplay *display;
+
+  if (!gst_vaapisink_ensure_display (sink))
+    return FALSE;
+  display = GST_VAAPI_PLUGIN_BASE_DISPLAY (sink);
+
+  if (sink->window == NULL || (guintptr) sink->window != window) {
+    gst_vaapi_window_replace (&sink->window, NULL);
+    sink->window = gst_vaapi_window_wayland_new_with_surface (display, window);
+  }
+
+  return sink->window != NULL;
+}
+
 static const inline GstVaapiSinkBackend *
 gst_vaapisink_backend_wayland (void)
 {
   static const GstVaapiSinkBackend GstVaapiSinkBackendWayland = {
     .create_window = gst_vaapisink_wayland_create_window,
+    .create_window_from_handle =
+        gst_vaapisink_wayland_create_window_from_handle,
     .render_surface = gst_vaapisink_render_surface,
   };
   return &GstVaapiSinkBackendWayland;
@@ -614,6 +633,12 @@ gst_vaapisink_video_overlay_set_render_rectangle (GstVideoOverlay * overlay,
   display_rect->width = width;
   display_rect->height = height;
 
+  if (gst_vaapisink_ensure_render_rect (sink, width, height) && sink->window) {
+    gst_vaapi_window_set_render_rectangle (sink->window, x, y, width, height);
+    gst_vaapi_window_set_size (sink->window, width, height);
+    gst_vaapisink_reconfigure_window (sink);
+  }
+
   GST_DEBUG ("render rect (%d,%d):%ux%u",
       display_rect->x, display_rect->y,
       display_rect->width, display_rect->height);
