@@ -47,6 +47,8 @@ struct _GstVaapiVideoBufferPoolPrivate
   guint options;
   guint use_dmabuf_memory:1;
   guint forced_video_meta:1;
+  /* Map between surface and GstMemory, only DMA */
+  GHashTable *dma_mem_map;
 };
 
 G_DEFINE_TYPE_WITH_PRIVATE (GstVaapiVideoBufferPool,
@@ -60,6 +62,8 @@ gst_vaapi_video_buffer_pool_finalize (GObject * object)
 
   gst_vaapi_display_replace (&priv->display, NULL);
   g_clear_object (&priv->allocator);
+  if (priv->dma_mem_map)
+    g_hash_table_destroy (priv->dma_mem_map);
 
   G_OBJECT_CLASS (gst_vaapi_video_buffer_pool_parent_class)->finalize (object);
 }
@@ -131,8 +135,8 @@ static gboolean
 gst_vaapi_video_buffer_pool_set_config (GstBufferPool * pool,
     GstStructure * config)
 {
-  GstVaapiVideoBufferPoolPrivate *const priv =
-      GST_VAAPI_VIDEO_BUFFER_POOL (pool)->priv;
+  GstVaapiVideoBufferPool *const base_pool = GST_VAAPI_VIDEO_BUFFER_POOL (pool);
+  GstVaapiVideoBufferPoolPrivate *const priv = base_pool->priv;
   GstCaps *caps;
   GstVideoInfo new_allocation_vinfo;
   const GstVideoInfo *allocator_vinfo;
@@ -143,7 +147,7 @@ gst_vaapi_video_buffer_pool_set_config (GstBufferPool * pool,
   guint size, min_buffers, max_buffers;
   guint surface_alloc_flags;
 
-  GST_DEBUG_OBJECT (pool, "config %" GST_PTR_FORMAT, config);
+  GST_DEBUG_OBJECT (base_pool, "config %" GST_PTR_FORMAT, config);
 
   caps = NULL;
   if (!gst_buffer_pool_config_get_params (config, &caps, &size, &min_buffers,
@@ -154,6 +158,10 @@ gst_vaapi_video_buffer_pool_set_config (GstBufferPool * pool,
   if (!gst_video_info_from_caps (&new_allocation_vinfo, caps))
     goto error_invalid_caps;
 
+  if (!gst_buffer_pool_config_has_option (config,
+          GST_BUFFER_POOL_OPTION_VAAPI_VIDEO_META))
+    goto error_no_vaapi_video_meta_option;
+
   allocator = NULL;
   if (!gst_buffer_pool_config_get_allocator (config, &allocator, NULL))
     goto error_invalid_allocator;
@@ -163,7 +171,7 @@ gst_vaapi_video_buffer_pool_set_config (GstBufferPool * pool,
       && (g_strcmp0 (allocator->mem_type, GST_VAAPI_VIDEO_MEMORY_NAME) != 0
           && g_strcmp0 (allocator->mem_type,
               GST_VAAPI_DMABUF_ALLOCATOR_NAME) != 0))
-    allocator = NULL;
+    goto error_invalid_allocator;
 
   /* get the allocator properties */
   if (allocator) {
@@ -184,7 +192,9 @@ gst_vaapi_video_buffer_pool_set_config (GstBufferPool * pool,
       && gst_video_info_changed (allocator_vinfo, &new_allocation_vinfo)) {
     gst_object_replace ((GstObject **) & priv->allocator, NULL);
 
-    if (allocator && priv->use_dmabuf_memory) {
+    /* dmabuf allocator can change its parameters: no need to create a
+     * new one */
+    if (priv->use_dmabuf_memory) {
       gst_allocator_set_vaapi_video_info (allocator, &new_allocation_vinfo,
           surface_alloc_flags);
     } else {
@@ -192,20 +202,11 @@ gst_vaapi_video_buffer_pool_set_config (GstBufferPool * pool,
     }
   }
 
-  if (!gst_buffer_pool_config_has_option (config,
-          GST_BUFFER_POOL_OPTION_VAAPI_VIDEO_META))
-    goto error_no_vaapi_video_meta_option;
-
   /* create a new allocator if needed */
   if (!allocator) {
-    if (priv->use_dmabuf_memory) {
-      allocator = gst_vaapi_dmabuf_allocator_new (priv->display,
-          &new_allocation_vinfo, /* FIXME: */ 0, GST_PAD_SRC);
-    } else {
-      allocator = gst_vaapi_video_allocator_new (priv->display,
-          &new_allocation_vinfo, surface_alloc_flags, 0);
-    }
-
+    /* if no allocator set, let's create a VAAPI one */
+    allocator = gst_vaapi_video_allocator_new (priv->display,
+        &new_allocation_vinfo, surface_alloc_flags, 0);
     if (!allocator)
       goto error_no_allocator;
 
@@ -214,7 +215,8 @@ gst_vaapi_video_buffer_pool_set_config (GstBufferPool * pool,
           negotiated_vinfo);
     }
 
-    GST_INFO_OBJECT (pool, "created new allocator %" GST_PTR_FORMAT, allocator);
+    GST_INFO_OBJECT (base_pool, "created new allocator %" GST_PTR_FORMAT,
+        allocator);
     gst_buffer_pool_config_set_allocator (config, allocator, NULL);
     gst_object_unref (allocator);
   }
@@ -252,10 +254,12 @@ gst_vaapi_video_buffer_pool_set_config (GstBufferPool * pool,
       if (GST_VIDEO_INFO_PLANE_OFFSET (&new_allocation_vinfo, i) !=
           GST_VIDEO_INFO_PLANE_OFFSET (&priv->vmeta_vinfo, i) ||
           GST_VIDEO_INFO_PLANE_STRIDE (&new_allocation_vinfo, i) !=
-          GST_VIDEO_INFO_PLANE_STRIDE (&priv->vmeta_vinfo, i)) {
+          GST_VIDEO_INFO_PLANE_STRIDE (&priv->vmeta_vinfo, i) ||
+          GST_VIDEO_INFO_SIZE (&new_allocation_vinfo) !=
+          GST_VIDEO_INFO_SIZE (&priv->vmeta_vinfo)) {
         priv->options |= GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_VIDEO_META;
         priv->forced_video_meta = TRUE;
-        GST_INFO_OBJECT (pool, "adding unrequested video meta");
+        GST_INFO_OBJECT (base_pool, "adding unrequested video meta");
         break;
       }
     }
@@ -279,74 +283,132 @@ gst_vaapi_video_buffer_pool_set_config (GstBufferPool * pool,
   /* ERRORS */
 error_invalid_config:
   {
-    GST_ERROR_OBJECT (pool, "invalid config");
+    GST_WARNING_OBJECT (base_pool, "invalid config");
     return FALSE;
   }
 error_no_caps:
   {
-    GST_ERROR_OBJECT (pool, "no caps in config");
+    GST_WARNING_OBJECT (base_pool, "no caps in config");
     return FALSE;
   }
 error_invalid_caps:
   {
-    GST_ERROR_OBJECT (pool, "invalid caps %" GST_PTR_FORMAT, caps);
+    GST_WARNING_OBJECT (base_pool, "invalid caps %" GST_PTR_FORMAT, caps);
     return FALSE;
   }
 error_invalid_allocator:
   {
-    GST_ERROR_OBJECT (pool, "no allocator in config");
+    GST_INFO_OBJECT (base_pool, "no allocator in config");
     return FALSE;
   }
 error_no_vaapi_video_meta_option:
   {
-    GST_ERROR_OBJECT (pool, "no GstVaapiVideoMeta option in config");
+    GST_WARNING_OBJECT (base_pool, "no GstVaapiVideoMeta option in config");
     return FALSE;
   }
 error_no_allocator:
   {
-    GST_ERROR_OBJECT (pool, "no allocator defined");
+    GST_WARNING_OBJECT (base_pool, "no allocator defined");
     return FALSE;
   }
 }
 
+static void
+vaapi_buffer_pool_cache_dma_mem (GstVaapiVideoBufferPool * pool,
+    GstVaapiSurfaceProxy * proxy, GstMemory * mem)
+{
+  GstVaapiVideoBufferPoolPrivate *const priv = pool->priv;
+  GstVaapiSurface *surface;
+
+  surface = GST_VAAPI_SURFACE_PROXY_SURFACE (proxy);
+  g_assert (surface);
+  g_assert (gst_vaapi_surface_peek_buffer_proxy (surface));
+
+  if (!priv->dma_mem_map)
+    priv->dma_mem_map = g_hash_table_new_full (g_direct_hash,
+        g_direct_equal, NULL, (GDestroyNotify) gst_memory_unref);
+
+  if (!g_hash_table_contains (priv->dma_mem_map, surface)) {
+    g_hash_table_insert (priv->dma_mem_map, surface, gst_memory_ref (mem));
+  } else {
+    g_assert (g_hash_table_lookup (priv->dma_mem_map, surface) == mem);
+  }
+}
+
+static GstMemory *
+vaapi_buffer_pool_lookup_dma_mem (GstVaapiVideoBufferPool * pool,
+    GstVaapiSurfaceProxy * proxy)
+{
+  GstVaapiSurface *surface;
+  GstVaapiVideoBufferPoolPrivate *const priv = pool->priv;
+  GstVaapiBufferProxy *buf_proxy;
+  GstMemory *mem;
+
+  g_assert (priv->use_dmabuf_memory);
+
+  if (!priv->dma_mem_map)
+    return NULL;
+
+  surface = GST_VAAPI_SURFACE_PROXY_SURFACE (proxy);
+  g_assert (surface);
+
+  buf_proxy = gst_vaapi_surface_peek_buffer_proxy (surface);
+  /* Have not exported yet */
+  if (!buf_proxy) {
+    g_assert (!g_hash_table_contains (priv->dma_mem_map, surface));
+    return NULL;
+  }
+
+  mem = g_hash_table_lookup (priv->dma_mem_map, surface);
+  g_assert (mem);
+
+  return gst_memory_ref (mem);
+}
+
 static GstFlowReturn
 gst_vaapi_video_buffer_pool_alloc_buffer (GstBufferPool * pool,
     GstBuffer ** out_buffer_ptr, GstBufferPoolAcquireParams * params)
 {
-  GstVaapiVideoBufferPoolPrivate *const priv =
-      GST_VAAPI_VIDEO_BUFFER_POOL (pool)->priv;
+  GstVaapiVideoBufferPool *const base_pool = GST_VAAPI_VIDEO_BUFFER_POOL (pool);
+  GstVaapiVideoBufferPoolPrivate *const priv = base_pool->priv;
   GstVaapiVideoBufferPoolAcquireParams *const priv_params =
       (GstVaapiVideoBufferPoolAcquireParams *) params;
   GstVaapiVideoMeta *meta;
   GstMemory *mem;
   GstBuffer *buffer;
 
-  const gboolean alloc_vaapi_video_meta = !params ||
-      !(params->flags & GST_VAAPI_VIDEO_BUFFER_POOL_ACQUIRE_FLAG_NO_ALLOC);
-
   if (!priv->allocator)
     goto error_no_allocator;
 
-  if (alloc_vaapi_video_meta) {
-    meta = gst_vaapi_video_meta_new (priv->display);
-    if (!meta)
-      goto error_create_meta;
+  meta = gst_vaapi_video_meta_new (priv->display);
+  if (!meta)
+    goto error_create_meta;
+
+  buffer = gst_vaapi_video_buffer_new (meta);
 
-    buffer = gst_vaapi_video_buffer_new (meta);
-  } else {
-    meta = NULL;
-    buffer = gst_vaapi_video_buffer_new_empty ();
-  }
   if (!buffer)
     goto error_create_buffer;
 
   if (priv_params && priv_params->proxy)
     gst_vaapi_video_meta_set_surface_proxy (meta, priv_params->proxy);
 
-  if (priv->use_dmabuf_memory)
-    mem = gst_vaapi_dmabuf_memory_new (priv->allocator, meta);
-  else
+  if (priv->use_dmabuf_memory) {
+    mem = NULL;
+    if (priv_params && priv_params->proxy) {
+      mem = vaapi_buffer_pool_lookup_dma_mem (base_pool, priv_params->proxy);
+      if (!mem) {
+        mem = gst_vaapi_dmabuf_memory_new (priv->allocator, meta);
+        if (!mem)
+          goto error_create_memory;
+
+        vaapi_buffer_pool_cache_dma_mem (base_pool, priv_params->proxy, mem);
+      }
+    } else {
+      mem = gst_vaapi_dmabuf_memory_new (priv->allocator, meta);
+    }
+  } else {
     mem = gst_vaapi_video_memory_new (priv->allocator, meta);
+  }
   if (!mem)
     goto error_create_memory;
   gst_vaapi_video_meta_replace (&meta, NULL);
@@ -366,10 +428,15 @@ gst_vaapi_video_buffer_pool_alloc_buffer (GstBufferPool * pool,
       vmeta->map = gst_video_meta_map_vaapi_memory;
       vmeta->unmap = gst_video_meta_unmap_vaapi_memory;
     }
+
+    GST_META_FLAG_SET (vmeta, GST_META_FLAG_POOLED);
   }
 #if (USE_GLX || USE_EGL)
-  if (priv->options & GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_GL_TEXTURE_UPLOAD)
-    gst_buffer_add_texture_upload_meta (buffer);
+  if (priv->options & GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_GL_TEXTURE_UPLOAD) {
+    GstMeta *tex_meta = gst_buffer_add_texture_upload_meta (buffer);
+    if (tex_meta)
+      GST_META_FLAG_SET (tex_meta, GST_META_FLAG_POOLED);
+  }
 #endif
 
   *out_buffer_ptr = buffer;
@@ -378,23 +445,23 @@ gst_vaapi_video_buffer_pool_alloc_buffer (GstBufferPool * pool,
   /* ERRORS */
 error_no_allocator:
   {
-    GST_ERROR_OBJECT (pool, "no GstAllocator in buffer pool");
+    GST_ERROR_OBJECT (base_pool, "no GstAllocator in buffer pool");
     return GST_FLOW_ERROR;
   }
 error_create_meta:
   {
-    GST_ERROR_OBJECT (pool, "failed to allocate vaapi video meta");
+    GST_ERROR_OBJECT (base_pool, "failed to allocate vaapi video meta");
     return GST_FLOW_ERROR;
   }
 error_create_buffer:
   {
-    GST_ERROR_OBJECT (pool, "failed to create video buffer");
+    GST_ERROR_OBJECT (base_pool, "failed to create video buffer");
     gst_vaapi_video_meta_unref (meta);
     return GST_FLOW_ERROR;
   }
 error_create_memory:
   {
-    GST_ERROR_OBJECT (pool, "failed to create video memory");
+    GST_ERROR_OBJECT (base_pool, "failed to create video memory");
     gst_buffer_unref (buffer);
     gst_vaapi_video_meta_unref (meta);
     return GST_FLOW_ERROR;
@@ -405,8 +472,8 @@ static GstFlowReturn
 gst_vaapi_video_buffer_pool_acquire_buffer (GstBufferPool * pool,
     GstBuffer ** out_buffer_ptr, GstBufferPoolAcquireParams * params)
 {
-  GstVaapiVideoBufferPoolPrivate *const priv =
-      GST_VAAPI_VIDEO_BUFFER_POOL (pool)->priv;
+  GstVaapiVideoBufferPool *const base_pool = GST_VAAPI_VIDEO_BUFFER_POOL (pool);
+  GstVaapiVideoBufferPoolPrivate *const priv = base_pool->priv;
   GstVaapiVideoBufferPoolAcquireParams *const priv_params =
       (GstVaapiVideoBufferPoolAcquireParams *) params;
   GstFlowReturn ret;
@@ -414,7 +481,6 @@ gst_vaapi_video_buffer_pool_acquire_buffer (GstBufferPool * pool,
   GstMemory *mem;
   GstVaapiVideoMeta *meta;
   GstVaapiSurface *surface;
-  GstVaapiBufferProxy *dmabuf_proxy;
 
   ret =
       GST_BUFFER_POOL_CLASS
@@ -427,47 +493,58 @@ gst_vaapi_video_buffer_pool_acquire_buffer (GstBufferPool * pool,
     return ret;
   }
 
-  /* The point of the following dance is to attach the right GstMemory to the
-   * current acquired buffer. Indeed this buffer can contain any of the
-   * GstFdmemory since this buffer have been popped out from the buffer pool's
-   * FIFO. So there is no garantee that this matches the current surface. The
-   * va decoder driver might not even use a FIFO. So there is no way to guess
-   * on the ordering. In short acquire_current_buffer on the va driver and on
-   * the buffer pool return none matching data. So we have to manually attach
-   * the right GstFdMemory to the acquired GstBuffer. The right GstMemory is
-   * the one associated with the current surface. */
+  /* Some pool users, such as decode, needs to acquire a buffer for a
+   * specified surface (via surface proxy). If not it is a DMABuf, we
+   * just replace the underlying surface proxy of buffer's
+   * GstVaapiVideoMeta. But in DMABuf case, the thing is a little bit
+   * more complicated:
+   *
+   * For DMABuf, GstMemory is-a GstFdMemory, which doesn't provide a
+   * way to change its FD, thus once created it's bound to a
+   * surface. On the other side, for performace reason, when the
+   * buffer is released, the buffer and its memory are cached in the
+   * buffer pool, and at next acquire_buffer() may still reuse a
+   * buffer and its memory. But the pushed surface by the decoder may
+   * be different from the one popped by the pool, so we need to
+   * replace the buffer's memory with the correct one. */
   g_assert (gst_buffer_n_memory (buffer) == 1);
 
-  /* Find the cached memory associated with the given surface. */
-  surface = GST_VAAPI_SURFACE_PROXY_SURFACE (priv_params->proxy);
-  dmabuf_proxy = gst_vaapi_surface_peek_buffer_proxy (surface);
-  if (dmabuf_proxy) {
-    mem = gst_vaapi_buffer_proxy_peek_mem (dmabuf_proxy);
-    if (mem == gst_buffer_peek_memory (buffer, 0))
-      mem = NULL;
-    else
-      mem = gst_memory_ref (mem);
-  } else {
-    /* The given surface has not been exported yet. */
-    meta = gst_buffer_get_vaapi_video_meta (buffer);
-    if (gst_vaapi_video_meta_get_surface_proxy (meta))
-      gst_vaapi_video_meta_set_surface_proxy (meta, priv_params->proxy);
+  /* Update the underlying surface proxy */
+  meta = gst_buffer_get_vaapi_video_meta (buffer);
+  if (!meta) {
+    *out_buffer_ptr = buffer;
+    return GST_FLOW_ERROR;
+  }
+  gst_vaapi_video_meta_set_surface_proxy (meta, priv_params->proxy);
 
-    mem =
-        gst_vaapi_dmabuf_memory_new (priv->allocator,
-        gst_buffer_get_vaapi_video_meta (buffer));
+  mem = vaapi_buffer_pool_lookup_dma_mem (base_pool, priv_params->proxy);
+  if (mem) {
+    if (mem == gst_buffer_peek_memory (buffer, 0)) {
+      gst_memory_unref (mem);
+      *out_buffer_ptr = buffer;
+      return GST_FLOW_OK;
+    }
+  } else {
+    /* Should be an unexported surface */
+    surface = GST_VAAPI_SURFACE_PROXY_SURFACE (priv_params->proxy);
+    g_assert (surface);
+    g_assert (gst_vaapi_surface_peek_buffer_proxy (surface) == NULL);
+    gst_vaapi_video_meta_set_surface_proxy (meta, priv_params->proxy);
+    mem = gst_vaapi_dmabuf_memory_new (priv->allocator, meta);
+    if (mem)
+      vaapi_buffer_pool_cache_dma_mem (base_pool, priv_params->proxy, mem);
   }
 
-  /* Attach the GstFdMemory to the output buffer. */
   if (mem) {
-    GST_DEBUG_OBJECT (pool, "assigning memory %p to acquired buffer %p", mem,
-        buffer);
     gst_buffer_replace_memory (buffer, 0, mem);
     gst_buffer_unset_flags (buffer, GST_BUFFER_FLAG_TAG_MEMORY);
+    *out_buffer_ptr = buffer;
+    return GST_FLOW_OK;
+  } else {
+    gst_buffer_unref (buffer);
+    *out_buffer_ptr = NULL;
+    return GST_FLOW_ERROR;
   }
-
-  *out_buffer_ptr = buffer;
-  return GST_FLOW_OK;
 }
 
 static void
@@ -475,10 +552,19 @@ gst_vaapi_video_buffer_pool_reset_buffer (GstBufferPool * pool,
     GstBuffer * buffer)
 {
   GstMemory *const mem = gst_buffer_peek_memory (buffer, 0);
+  GstVaapiVideoMeta *meta;
 
   /* Release the underlying surface proxy */
-  if (GST_VAAPI_IS_VIDEO_MEMORY (mem))
+  if (GST_VAAPI_IS_VIDEO_MEMORY (mem)) {
     gst_vaapi_video_memory_reset_surface (GST_VAAPI_VIDEO_MEMORY_CAST (mem));
+  } else if (!gst_vaapi_dmabuf_memory_holds_surface (mem)) {
+    /* If mem holds an internally created surface, don't reset it!
+     * While surface is passed, we should clear it to avoid wrong
+     * reference. */
+    meta = gst_buffer_get_vaapi_video_meta (buffer);
+    g_assert (meta);
+    gst_vaapi_video_meta_set_surface_proxy (meta, NULL);
+  }
 
   GST_BUFFER_POOL_CLASS (gst_vaapi_video_buffer_pool_parent_class)->reset_buffer
       (pool, buffer);
@@ -519,6 +605,7 @@ static void
 gst_vaapi_video_buffer_pool_init (GstVaapiVideoBufferPool * pool)
 {
   pool->priv = gst_vaapi_video_buffer_pool_get_instance_private (pool);
+  pool->priv->dma_mem_map = NULL;
 }
 
 GstBufferPool *
