@@ -128,7 +128,7 @@ ensure_image (GstVaapiVideoMemory * mem)
       reset_image_usage (&mem->usage_flag);
     } else if (gst_vaapi_surface_get_format (mem->surface) !=
         GST_VIDEO_INFO_FORMAT (mem->image_info)) {
-      gst_vaapi_object_replace (&mem->image, NULL);
+      gst_mini_object_replace ((GstMiniObject **) & mem->image, NULL);
       reset_image_usage (&mem->usage_flag);
     }
   }
@@ -367,7 +367,7 @@ gst_vaapi_video_memory_new (GstAllocator * base_allocator,
 
   vip = &allocator->image_info;
   gst_memory_init (&mem->parent_instance, GST_MEMORY_FLAG_NO_SHARE,
-      gst_object_ref (allocator), NULL, GST_VIDEO_INFO_SIZE (vip), 0,
+      base_allocator, NULL, GST_VIDEO_INFO_SIZE (vip), 0,
       0, GST_VIDEO_INFO_SIZE (vip));
 
   mem->proxy = NULL;
@@ -393,7 +393,7 @@ gst_vaapi_video_memory_reset_image (GstVaapiVideoMemory * mem)
       GST_VAAPI_VIDEO_ALLOCATOR_CAST (GST_MEMORY_CAST (mem)->allocator);
 
   if (!use_native_formats (mem->usage_flag))
-    gst_vaapi_object_replace (&mem->image, NULL);
+    gst_mini_object_replace ((GstMiniObject **) & mem->image, NULL);
   else if (mem->image) {
     gst_vaapi_video_pool_put_object (allocator->image_pool, mem->image);
     mem->image = NULL;
@@ -607,7 +607,6 @@ gst_vaapi_video_allocator_free (GstAllocator * allocator, GstMemory * base_mem)
   gst_vaapi_video_memory_reset_image (mem);
   gst_vaapi_surface_proxy_replace (&mem->proxy, NULL);
   gst_vaapi_video_meta_replace (&mem->meta, NULL);
-  gst_object_unref (GST_MEMORY_CAST (mem)->allocator);
   g_mutex_clear (&mem->lock);
   g_slice_free (GstVaapiVideoMemory, mem);
 }
@@ -703,7 +702,7 @@ gst_video_info_update_from_surface (GstVideoInfo * vip,
   gst_vaapi_image_unmap (image);
 
 bail:
-  gst_vaapi_object_unref (image);
+  gst_vaapi_image_unref (image);
   return ret;
 
   /* ERRORS */
@@ -739,14 +738,16 @@ gst_vaapi_image_usage_flags_to_string (GstVaapiImageUsageFlags usage_flag)
 static inline gboolean
 allocator_configure_surface_try_specified_format (GstVaapiDisplay * display,
     const GstVideoInfo * allocation_info, GstVaapiImageUsageFlags usage_flag,
-    GstVideoInfo * ret_surface_info, GstVaapiImageUsageFlags * ret_usage_flag)
+    guint surface_alloc_flag, GstVideoInfo * ret_surface_info,
+    GstVaapiImageUsageFlags * ret_usage_flag)
 {
   GstVaapiImageUsageFlags rflag;
   GstVaapiSurface *surface;
   GstVideoInfo sinfo, rinfo;
 
   /* Try to create a surface with the given allocation info */
-  surface = gst_vaapi_surface_new_full (display, allocation_info, 0);
+  surface =
+      gst_vaapi_surface_new_full (display, allocation_info, surface_alloc_flag);
   if (!surface)
     return FALSE;
 
@@ -789,7 +790,7 @@ allocator_configure_surface_try_specified_format (GstVaapiDisplay * display,
   rinfo = *allocation_info;
 
 out:
-  gst_vaapi_object_unref (surface);
+  gst_vaapi_surface_unref (surface);
 
   *ret_surface_info = rinfo;
   *ret_usage_flag = rflag;
@@ -820,7 +821,7 @@ allocator_configure_surface_try_other_format (GstVaapiDisplay * display,
   surface = gst_vaapi_surface_new_full (display, &sinfo, 0);
   if (!surface)
     goto error_no_surface;
-  gst_vaapi_object_unref (surface);
+  gst_vaapi_surface_unref (surface);
 
   *ret_surface_info = sinfo;
   return TRUE;
@@ -841,7 +842,8 @@ error_no_surface:
 
 static inline gboolean
 allocator_configure_surface_info (GstVaapiDisplay * display,
-    GstVaapiVideoAllocator * allocator, GstVaapiImageUsageFlags req_usage_flag)
+    GstVaapiVideoAllocator * allocator, GstVaapiImageUsageFlags req_usage_flag,
+    guint surface_alloc_flags)
 {
   GstVaapiImageUsageFlags usage_flag;
   GstVideoInfo allocation_info, surface_info;
@@ -853,7 +855,8 @@ allocator_configure_surface_info (GstVaapiDisplay * display,
   /* Step1: Try the specified format and flag. May fallback to native if
      direct upload/rendering is unavailable. */
   if (allocator_configure_surface_try_specified_format (display,
-          &allocation_info, req_usage_flag, &surface_info, &usage_flag)) {
+          &allocation_info, req_usage_flag, surface_alloc_flags,
+          &surface_info, &usage_flag)) {
     allocator->usage_flag = usage_flag;
     allocator->surface_info = surface_info;
     goto success;
@@ -912,7 +915,7 @@ allocator_configure_image_info (GstVaapiDisplay * display,
 
 bail:
   if (image)
-    gst_vaapi_object_unref (image);
+    gst_vaapi_image_unref (image);
   return ret;
 
   /* ERRORS */
@@ -935,7 +938,8 @@ allocator_params_init (GstVaapiVideoAllocator * allocator,
 {
   allocator->allocation_info = *alloc_info;
 
-  if (!allocator_configure_surface_info (display, allocator, req_usage_flag))
+  if (!allocator_configure_surface_info (display, allocator, req_usage_flag,
+          surface_alloc_flags))
     return FALSE;
   allocator->surface_pool = gst_vaapi_surface_pool_new_full (display,
       &allocator->surface_info, surface_alloc_flags);
@@ -1007,6 +1011,18 @@ gst_vaapi_buffer_proxy_quark_get (void)
   return g_quark;
 }
 
+/* Whether @mem holds an internal VA surface proxy created at
+ * gst_vaapi_dmabuf_memory_new(). */
+gboolean
+gst_vaapi_dmabuf_memory_holds_surface (GstMemory * mem)
+{
+  g_return_val_if_fail (mem != NULL, FALSE);
+
+  return
+      GPOINTER_TO_INT (gst_mini_object_get_qdata (GST_MINI_OBJECT_CAST (mem),
+          GST_VAAPI_BUFFER_PROXY_QUARK)) == TRUE;
+}
+
 GstMemory *
 gst_vaapi_dmabuf_memory_new (GstAllocator * base_allocator,
     GstVaapiVideoMeta * meta)
@@ -1049,20 +1065,21 @@ gst_vaapi_dmabuf_memory_new (GstAllocator * base_allocator,
     proxy = gst_vaapi_surface_proxy_new (surface);
     if (!proxy)
       goto error_create_surface_proxy;
+    /* The proxy has incremented the surface ref count. */
+    gst_vaapi_surface_unref (surface);
   } else {
     /* When exporting existing surfaces that come from decoder's
      * context. */
     surface = GST_VAAPI_SURFACE_PROXY_SURFACE (proxy);
   }
 
-  dmabuf_proxy = gst_vaapi_surface_get_dma_buf_handle (surface);
+  dmabuf_proxy = gst_vaapi_surface_peek_dma_buf_handle (surface);
   if (!dmabuf_proxy)
     goto error_create_dmabuf_proxy;
 
   if (needs_surface) {
-    /* The proxy has incremented the surface ref count.  */
-    gst_vaapi_object_unref (surface);
     gst_vaapi_video_meta_set_surface_proxy (meta, proxy);
+    /* meta holds the proxy's reference */
     gst_vaapi_surface_proxy_unref (proxy);
   }
 
@@ -1079,17 +1096,9 @@ gst_vaapi_dmabuf_memory_new (GstAllocator * base_allocator,
     goto error_create_dmabuf_memory;
 
   if (needs_surface) {
-    /* Just set the GstVaapiBufferProxy (dmabuf_proxy) as qdata and
-     * forget about it. */
+    /* qdata express that memory has an associated surface. */
     gst_mini_object_set_qdata (GST_MINI_OBJECT_CAST (mem),
-        GST_VAAPI_BUFFER_PROXY_QUARK, dmabuf_proxy,
-        (GDestroyNotify) gst_vaapi_buffer_proxy_unref);
-  } else {
-    /* When not allocating the surface from this pool, so when
-     * exporting from the decoder's VA context, we need to know which
-     * GstMemory belongs to a provided surface. */
-    gst_vaapi_buffer_proxy_set_mem (dmabuf_proxy, mem);
-    gst_vaapi_surface_set_buffer_proxy (surface, dmabuf_proxy);
+        GST_VAAPI_BUFFER_PROXY_QUARK, GINT_TO_POINTER (TRUE), NULL);
   }
 
   /* When a VA surface is going to be filled by a VAAPI element
@@ -1113,14 +1122,14 @@ error_create_surface:
 error_create_surface_proxy:
   {
     GST_ERROR ("failed to create VA surface proxy");
-    gst_vaapi_object_unref (surface);
+    gst_vaapi_surface_unref (surface);
     return NULL;
   }
 error_create_dmabuf_proxy:
   {
     GST_ERROR ("failed to export VA surface to DMABUF");
     if (surface)
-      gst_vaapi_object_unref (surface);
+      gst_vaapi_surface_unref (surface);
     if (proxy)
       gst_vaapi_surface_proxy_unref (proxy);
     return NULL;
@@ -1189,7 +1198,7 @@ gst_vaapi_dmabuf_allocator_new (GstVaapiDisplay * display,
     goto error_no_surface;
   if (!gst_video_info_update_from_surface (&surface_info, surface))
     goto fail;
-  gst_vaapi_object_replace (&surface, NULL);
+  gst_mini_object_replace ((GstMiniObject **) & surface, NULL);
 
   gst_allocator_set_vaapi_video_info (base_allocator, &surface_info,
       surface_alloc_flags);
@@ -1201,7 +1210,7 @@ gst_vaapi_dmabuf_allocator_new (GstVaapiDisplay * display,
   /* ERRORS */
 fail:
   {
-    gst_vaapi_object_replace (&surface, NULL);
+    gst_mini_object_replace ((GstMiniObject **) & surface, NULL);
     gst_object_replace ((GstObject **) & base_allocator, NULL);
     return NULL;
   }
@@ -1345,7 +1354,8 @@ gst_allocator_set_vaapi_video_info (GstAllocator * allocator,
 /**
  * gst_allocator_set_vaapi_negotiated_video_info:
  * @allocator: a #GstAllocator
- * @negotiated_vinfo: the negotiated #GstVideoInfo to store
+ * @negotiated_vinfo: the negotiated #GstVideoInfo to store.  If NULL, then
+ * removes any previously set value.
  *
  * Stores as GObject's qdata the @negotiated_vinfo in the allocator
  * instance.
@@ -1358,11 +1368,13 @@ gst_allocator_set_vaapi_negotiated_video_info (GstAllocator * allocator,
     const GstVideoInfo * negotiated_vinfo)
 {
   g_return_if_fail (allocator && GST_IS_ALLOCATOR (allocator));
-  g_return_if_fail (negotiated_vinfo);
 
-  g_object_set_qdata_full (G_OBJECT (allocator), NEGOTIATED_VINFO_QUARK,
-      gst_video_info_copy (negotiated_vinfo),
-      (GDestroyNotify) gst_video_info_free);
+  if (negotiated_vinfo)
+    g_object_set_qdata_full (G_OBJECT (allocator), NEGOTIATED_VINFO_QUARK,
+        gst_video_info_copy (negotiated_vinfo),
+        (GDestroyNotify) gst_video_info_free);
+  else
+    g_object_set_qdata (G_OBJECT (allocator), NEGOTIATED_VINFO_QUARK, NULL);
 }
 
 /**
