@@ -125,7 +125,7 @@ meta_texture_free (GstVaapiVideoMetaTexture * meta)
   if (G_UNLIKELY (!meta))
     return;
 
-  gst_vaapi_texture_replace (&meta->texture, NULL);
+  gst_mini_object_replace ((GstMiniObject **) & meta->texture, NULL);
   g_slice_free (GstVaapiVideoMetaTexture, meta);
 }
 
@@ -164,7 +164,9 @@ meta_texture_copy (GstVaapiVideoMetaTexture * meta)
   copy->gl_format = meta->gl_format;
   copy->width = meta->width;
   copy->height = meta->height;
-  gst_vaapi_texture_replace (&copy->texture, meta->texture);
+
+  gst_mini_object_replace ((GstMiniObject **) & copy->texture,
+      (GstMiniObject *) meta->texture);
   return copy;
 }
 
@@ -178,8 +180,7 @@ gst_vaapi_texture_upload (GstVideoGLTextureUploadMeta * meta,
   GstVaapiSurfaceProxy *const proxy =
       gst_vaapi_video_meta_get_surface_proxy (vmeta);
   GstVaapiSurface *const surface = gst_vaapi_surface_proxy_get_surface (proxy);
-  GstVaapiDisplay *const dpy =
-      gst_vaapi_object_get_display (GST_VAAPI_OBJECT (surface));
+  GstVaapiDisplay *const dpy = gst_vaapi_surface_get_display (surface);
   GstVaapiTexture *texture = NULL;
 
   if (!gst_vaapi_display_has_opengl (dpy))
@@ -187,8 +188,7 @@ gst_vaapi_texture_upload (GstVideoGLTextureUploadMeta * meta,
 
   if (meta_texture->texture
       /* Check whether VA display changed */
-      && gst_vaapi_object_get_display
-      (GST_VAAPI_OBJECT (meta_texture->texture)) == dpy
+      && GST_VAAPI_TEXTURE_DISPLAY (meta_texture->texture) == dpy
       /* Check whether texture id changed */
       && (gst_vaapi_texture_get_id (meta_texture->texture) == texture_id[0])) {
     texture = meta_texture->texture;
@@ -203,7 +203,8 @@ gst_vaapi_texture_upload (GstVideoGLTextureUploadMeta * meta,
   }
 
   if (meta_texture->texture != texture) {
-    gst_vaapi_texture_replace (&meta_texture->texture, texture);
+    gst_mini_object_replace ((GstMiniObject **) & meta_texture->texture,
+        (GstMiniObject *) texture);
   }
 
   if (!texture)
@@ -217,10 +218,9 @@ gst_vaapi_texture_upload (GstVideoGLTextureUploadMeta * meta,
       gst_vaapi_video_meta_get_render_flags (vmeta));
 }
 
-gboolean
+GstMeta *
 gst_buffer_add_texture_upload_meta (GstBuffer * buffer)
 {
-  GstVideoGLTextureUploadMeta *meta = NULL;
   GstVaapiVideoMetaTexture *meta_texture;
 
   if (!buffer)
@@ -233,20 +233,16 @@ gst_buffer_add_texture_upload_meta (GstBuffer * buffer)
   if (!meta_texture_ensure_info_from_buffer (meta_texture, buffer))
     goto error;
 
-  meta = gst_buffer_add_video_gl_texture_upload_meta (buffer,
-      GST_VIDEO_GL_TEXTURE_ORIENTATION_X_NORMAL_Y_NORMAL,
-      1, meta_texture->texture_type, gst_vaapi_texture_upload,
-      meta_texture, (GBoxedCopyFunc) meta_texture_copy,
-      (GBoxedFreeFunc) meta_texture_free);
-  if (!meta)
-    goto error;
-  return TRUE;
+  return (GstMeta *) gst_buffer_add_video_gl_texture_upload_meta (buffer,
+      GST_VIDEO_GL_TEXTURE_ORIENTATION_X_NORMAL_Y_NORMAL, 1,
+      meta_texture->texture_type, gst_vaapi_texture_upload, meta_texture,
+      (GBoxedCopyFunc) meta_texture_copy, (GBoxedFreeFunc) meta_texture_free);
 
   /* ERRORS */
 error:
   {
     meta_texture_free (meta_texture);
-    return FALSE;
+    return NULL;
   }
 }
 
@@ -258,6 +254,6 @@ gst_buffer_ensure_texture_upload_meta (GstBuffer * buffer)
 
   return meta ?
       meta_texture_ensure_info_from_buffer (meta->user_data, buffer) :
-      gst_buffer_add_texture_upload_meta (buffer);
+      (gst_buffer_add_texture_upload_meta (buffer) != NULL);
 }
 #endif
